import rivet
import types
import string

def associate_scrollbar(client, sbar):
	"""
	Utility function to set up callbacks for the tk objects that
	have support for scrollbars (canvas, entry, listbox, and text).
	"""
	if sbar.cget('orient') == 'vertical':
		sbar.configure('command', client.yview)
		client.configure('yscrollcommand', sbar.set)
	else:
		sbar.configure('command', client.xview)
		client.configure('xscrollcommand', sbar.set)

def option_menu(parent, var, *args):
	if var == None:
		var = stringvar()

	if len(args) < 1:
		return None

	if type(args[0]) == types.TupleType or type(args[0]) == types.ListType:
		# flatten, so that you can pass in a list or tuple of names.
		args = args[0]

	var.set(args[0])
	mb = rivet.menubutton(parent, '-textvariable', var,
		'-indicatoron', 1, '-relief', 'raised', '-bd', 2, \
		'-padx', '4p', '-pady', '4p', '-highlightthickness', 2, \
		'-anchor', 'c')
	menu = mb.menu('-tearoff', 0)
	mb.configure('-menu', menu)

	for i in args:
		menu.add('command', '-label', i,
			'-command', (var.set, i))

	mb.var = var	# so caller can get it later
	return mb

class stringvar:
	varnamecount = 0
	def __init__(self):
		stringvar.varnamecount = stringvar.varnamecount + 1
		self.__var = str(stringvar.varnamecount)

	def __del__(self):
		try:
			rivet.unsetvar(self.__var)
		except:
			pass # it may have already been deleted

	def set(self, value):
		rivet.setvar(self.__var, str(value))

	def get(self):
		return rivet.getvar(self.__var)

	def unset(self):
		rivet.unsetvar(self.__var)

	def trace(self, mode, callback):
		rivet.tracevar(self.__var, mode, callback)

	def untrace(self, mode, callback):
		rivet.untracevar(self.__var, mode, callback)

class intvar(stringvar):
	def __init__(self):
		stringvar.__init__(self)
	def get(self):
		return string.atoi(rivet.getvar(self.__var))

class floatvar(stringvar):
	def __init__(self):
		stringvar.__init__(self)
	def get(self):
		return string.atof(rivet.getvar(self.__var))


