
#include "rivet.h"

int
callback(Rivetobj rivetobj, XEvent *unused1, ClientData unused2, ClientData unused3, int unused4)
{
	static int color;
	rivet_set(rivetobj, "-bg", (color?"yellow":"purple"));
	color ^= 1;
	return TCL_CONTINUE;
}
int
canvas_cb(Rivetobj canvas, XEvent *xevent, void *closure, int unused1)
{
	static int color;
	rivet_set(canvas, "-bg", (color?"red":"orange"));
	color ^= 1;
	return TCL_CONTINUE;
}

int
goodbye(Rivetobj rivetobj, XEvent *xevent, ClientData assoc, ClientData atom, int unused1)
{
	char *atomname = (char*)atom;
	printf("goodbye\n");
	rivet_destroy_window(rivetobj);
}

Rivetobj
do_canvas(Rivetobj mainrobj, Rivetobj parent)
{
	Rivetobj	obj;

	obj = rivet_create(CanvasClass, parent, 0, 0);
	if (!obj)
		 /* error */ ;
	rivet_va_cmd(obj, "config",
		"-width", "50",
		"-height", "50",
		"-bg", "lightblue",
		"-bd", "2",
		"-relief", "raised",
		0);
	rivet_va_cmd(obj, "create", "oval", "10", "15", "40", "35",
			"-fill", "green",
			0);
	rivet_create_binding(obj, "1", "<1>", canvas_cb, 0, 0, 0);
	return obj;
}

Rivetobj listboxw;
Rivetobj listsb;
int
scroll_command(Rivetobj scrollbar, XEvent *unused,
		ClientData unused2, ClientData data, int unused1)
{
	rivet_scrollbar_set_client_view(listboxw,
				(Scrollbar_notify_data *) data);
}

int
setlistboxview(Rivetobj listbox, XEvent *unused1,
		ClientData unused2, ClientData data, int unused3)
{
	double *where = (double*) data;

	rivet_scrollbar_set(listsb, where[0], where[1]);
}

Rivetobj
build_listbox(Rivetobj parent)
{
	Rivetobj fr;
	Rivetobj sb;

	fr = rivet_create(FrameClass, parent, 0, 0);

	listboxw = rivet_create(ListboxClass, fr, 0, 0);
	listsb = rivet_create(ScrollbarClass, fr, 0, 0);
	rivet_set(listsb, "-command",
		rivet_new_callback(scroll_command, 0, 0));
	rivet_set(listboxw, "-yscrollcommand",
		rivet_new_callback(setlistboxview, 0, 0));

	rivet_va_func(parent, Tk_PackCmd,
		rivet_path(listboxw), "-side", "left", 0);
	rivet_va_func(parent, Tk_PackCmd,
		rivet_path(listsb), "-side", "right", "-fill", "y", 0);

	rivet_va_cmd(listboxw, "insert", "0", 
		"barend", "headset", "stem", "fork", "crank",
		"tire", "bottom braket", "shifter", "handlebar",
		"stoppers", "seatpost", "frame", "chain", "cog",
		0);

	return fr;
}
 
int
postcallback(Rivetobj menu, XEvent *unused1, ClientData ignored, ClientData unused2)
{
	printf("posting menu\n");
	return TCL_OK;
}

int
menuentry_callback1(Rivetobj menu, XEvent *unused, ClientData ignored, ClientData unused2, int unused3)
{
	printf("menu item 1\n");
	rivet_set(listboxw, "-selectmode", "single");
	return TCL_OK;
}

int
menuentry_callback2(Rivetobj menu, XEvent *unused, ClientData ignored, ClientData unused2, int unused3)
{
	printf("menu item 2\n");
	rivet_set(listboxw, "-selectmode", "browse");
	return TCL_OK;
}

int
menuentry_callback3(Rivetobj menu, XEvent *unused, ClientData ignored, ClientData unused2, int unused3)
{
	printf("menu item 3\n");
	rivet_set(listboxw, "-selectmode", "extended");
	return TCL_OK;
}

int
menuentry_callback4(Rivetobj menu, XEvent *unused, ClientData ignored, ClientData unused2, int unused3)
{
	printf("menu item 4\n");
	rivet_set(listboxw, "-selectmode", "multiple");
	return TCL_OK;
}

Rivetobj build_menu(Rivetobj parent)
{ 
        Rivetobj mb, menu;

        mb = rivet_create(MenubuttonClass, parent, 0, 0);
        rivet_set(mb, "-text", "Listbox modes");

	menu = rivet_create(MenuClass, mb, 0, 0);
	rivet_set(menu,
		"-postcommand", rivet_new_callback(postcallback, 0, 0));

	rivet_va_cmd(menu, "add", "command", "-label", "single", 
		"-command", rivet_new_callback(menuentry_callback1, 0, 0),
		0);

	rivet_va_cmd(menu, "add", "command", "-label", "browse",
		"-command", rivet_new_callback(menuentry_callback2, 0, 0),
		0);

	rivet_va_cmd(menu, "add", "command", "-label", "extended",
		"-command", rivet_new_callback(menuentry_callback3, 0, 0),
		0);

	rivet_va_cmd(menu, "add", "command", "-label", "multiple",
		"-command", rivet_new_callback(menuentry_callback4, 0, 0),
		0);

	rivet_set(mb, "-menu", rivet_path(menu));

	return mb;
} 

int
buttoncallback(Rivetobj button, XEvent *unused1, ClientData assoc, ClientData unused2)
{
	printf("button invoked\n");
	return TCL_OK;
}


int
scalecallback(Rivetobj scale, XEvent *unused, ClientData assoc, ClientData calldata)
{
	double *value;
	value = (double*)calldata;
	printf("scale set to %g\n", *value);
	return TCL_OK;
}
/*----------------------------------------------------------------------*/
void
dummy5(Rivetobj mainrobj)
{
	int             status;
	int             argc;
	char           *argv[20];
	int		numtopack;

	unsigned long   mask;
	TkWindow       *winPtr;

	Rivetobj        fr1;
	Rivetobj        r1, r2, r3, r4, r5, r6, r7, r8, r9;
	Rivetobj        r10, r11, r12;

	argv[0] = argv[1] = 0;
	argv[2] = "-bd";
	argv[3] = "4";
	argv[4] = "-relief";
	argv[5] = "raised";
	argc=6;
	fr1 = rivet_create(FrameClass, mainrobj, argc, argv);
	if (!fr1)
		 /* error */ ;

	argv[0] = argv[1] = 0;
	argv[2] = "-text";
	argv[3] = "Button1";
	argv[4] = "-command";
	argv[5] = rivet_new_callback(buttoncallback, 0, 0);
	argc=6;
	r1 = rivet_create(ButtonClass, fr1, argc, argv);
	if (!r1)
		 /* error */ ;

	argv[0] = 0;
	argv[1] = ".childlabelname";
	argv[2] = "-text";
	argv[3] = "Label1";
	argc=4;
	r2 = rivet_create(LabelClass, fr1, argc, argv);
	if (!r2)
		 /* error */ ;

	argv[0] = argv[1] = 0;
	argv[2] = "-text";
	argv[3] = "RadioButton1";
	argc=4;
	r3 = rivet_create(RadioButtonClass, fr1, argc, argv);
	if (!r3)
		 /* error */ ;

	argv[0] = argv[1] = 0;
	argv[2] = "-text";
	argv[3] = "CheckButton1";
	argc=4;
	r4 = rivet_create(CheckButtonClass, fr1, argc, argv);
	if (!r4)
		 /* error */ ;

	r5 = rivet_create(ScaleClass, fr1, 0, 0);
	if (!r5)
		 /* error */ ;
	rivet_set(r5, "-command", rivet_new_callback(scalecallback, 0, 0));

	argv[0] = argv[1] = 0;
	argv[2] = "-orient";
	argv[3] = "horizontal";
	argc=4;
	r6 = rivet_create(ScaleClass, fr1, argc, argv);
	if (!r6)
		 /* error */ ;

	r7 = do_canvas(mainrobj, fr1);

	r8 = rivet_create(EntryClass, fr1, 0, 0);
	r9 = build_listbox(fr1);
	r10 = build_menu(fr1);
	r11 = rivet_create(MessageClass, fr1, 0, 0);
	r12 = rivet_create(TextClass, fr1, 0, 0);
	rivet_va_cmd(r12, "config", "-width", "20", "-height", "4", 0);

	rivet_va_func(mainrobj, Tk_PackCmd,
		rivet_path(fr1),
		rivet_path(r1),
		rivet_path(r2),
		rivet_path(r3),
		rivet_path(r4),
		rivet_path(r5),
		rivet_path(r6),
		rivet_path(r7),
		rivet_path(r8),
		rivet_path(r9),
		rivet_path(r10),
		rivet_path(r11),
		rivet_path(r12),
		0);

	if (status != TCL_OK)
		 /* error */ ;

        if( rivet_create_binding(r1, 0, "<1>", callback, 0, 1, 0) == TCL_ERROR)
                ; /* error */

	rivet_va_func(mainrobj, Tk_WmCmd,
		"protocol", rivet_path(mainrobj), "WM_DELETE_WINDOW", 
		rivet_new_callback(goodbye, 0, 0), 0);
}
/*----------------------------------------------------------------------*/


int
main(argc, argv)
	int             argc;
	char          **argv;
{
	Rivetobj top;
	Tcl_Interp *interp;

	top = rivet_init(argc, argv);
	dummy5(top);
	Tk_MainLoop();
}

