
#include "rivet.h"

/*
 *--------------------------------------------------------------
 * Following code from button.tcl.
 *--------------------------------------------------------------
 */

/*
 * tkButtonEnter --
 * The procedure below is invoked when the mouse pointer enters a
 * button widget.  It records the button we're in and changes the
 * state of the button to active unless the button is disabled.
 */
static int
rivet_button_enter(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	if (!rivet_state_p(rivetobj, "disabled")) {
		if(rivet_set(rivetobj, "-state", "active")==TCL_ERROR)
			return TCL_BREAK;
 		if (tkPriv->buttonWindow == rivetobj) {
			if(rivet_set(rivetobj, "-state", "active")==TCL_ERROR)
				return TCL_BREAK;
			if(rivet_set(rivetobj, "-relief", "sunken")==TCL_ERROR)
				return TCL_BREAK;
 		}
	}
	tkPriv->window = rivetobj;
        return TCL_CONTINUE;
}
 
 
/*
 * tkButtonLeave --
 * The procedure below is invoked when the mouse pointer leaves a
 * button widget.  It changes the state of the button back to
 * inactive.  If we're leaving the button window with a mouse button
 * pressed (tkPriv(buttonWindow) == $w), restore the relief of the
 * button too.
 *
 */
static int
rivet_button_leave(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	if (!rivet_state_p(rivetobj, "disabled")) {
		rivet_set(rivetobj, "-state", "normal");
	}
	if (tkPriv->buttonWindow == rivetobj) {
		rivet_set(rivetobj, "-relief", tkPriv->relief);
	}
	tkPriv->window = NULL;
        return TCL_CONTINUE;
}
 
/*
 * tkButtonDown --
 * The procedure below is invoked when the mouse button is pressed in
 * a button widget.  It records the fact that the mouse is in the button,
 * saves the button's relief so it can be restored later, and changes
 * the relief to sunken.
 *
 */
static int
rivet_button_down(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	strcpy(tkPriv->relief, rivet_get(rivetobj, "-relief"));
	if (!rivet_state_p(rivetobj, "disabled")) {
		tkPriv->buttonWindow = rivetobj;
		if(rivet_set(rivetobj, "-relief", "sunken")==TCL_ERROR)
			return TCL_BREAK;
	}
        return TCL_CONTINUE;
}

XEvent *rivet_button_invoke_event;
int
rivet_invoke(widget)
	Rivetobj widget;
{
	char *argv[2];
	int argc = 2;
	int rv;

	Tcl_SetResult(widget->interp, 0, 0);
	argv[0] = rivet_path(widget);
	argv[1] = "invoke";
	rv = (widget->rivet_class->widgetcmd) (widget, widget->interp,
					       argc, argv);
	return rv;
}

 
/*
 * tkButtonUp --
 * The procedure below is invoked when the mouse button is released
 * in a button widget.  It restores the button's relief and invokes
 * the command as long as the mouse hasn't left the button.
 */
static int
rivet_button_up(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	if(rivetobj == tkPriv->buttonWindow) {
		tkPriv->buttonWindow = NULL;
		if(rivet_set(rivetobj, "-relief", tkPriv->relief)==TCL_ERROR)
			return TCL_BREAK;
		if (rivetobj == tkPriv->window &&
			!rivet_state_p(rivetobj, "disabled")) {
				rivet_button_invoke_event = xevent;
				if(rivet_invoke(rivetobj) == TCL_ERROR)
					return TCL_BREAK;
		}
	}
        return TCL_CONTINUE;
}
 
/*
 * tkButtonInvoke --
 * The procedure below is called when a button is invoked through
 * the keyboard.  It simulate a press of the button via the mouse.
 *
 */
static int
rivet_button_invoke(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	char oldRelief[20];
	char oldState[20];

	if(rivetobj->rivet_class != ButtonClass) {
		/* tkCheckRadioInvoke */
		if(!rivet_state_p(rivetobj, "disabled"))
			rivet_invoke(rivetobj);
	} else
	if(!rivet_state_p(rivetobj, "disabled")) {
		strcpy(oldRelief, rivet_get(rivetobj, "-relief"));
		strcpy(oldState, rivet_get(rivetobj, "-state"));

		if(rivet_set(rivetobj, "-state", "active")==TCL_ERROR)
			return TCL_BREAK;
		if(rivet_set(rivetobj, "-relief", "sunken")==TCL_ERROR)
			return TCL_BREAK;

	 	rivet_update(rivetobj, 1);
		Tk_Sleep(100);

		if(rivet_set(rivetobj, "-state", oldState)==TCL_ERROR)
			return TCL_BREAK;
		if(rivet_set(rivetobj, "-relief", oldRelief)==TCL_ERROR)
			return TCL_BREAK;
		rivet_button_invoke_event = xevent;
		if(rivet_invoke(rivetobj) == TCL_ERROR)
			return TCL_BREAK;
	}
        return TCL_CONTINUE;
}


void
rivet_button_bind(widget)
	Rivetobj	widget;
{
	if(rivet_is_class_initialized(widget->rivet_class))
		return;
	rivet_class_init_done(widget->rivet_class);

	rivet_create_binding(0, widget->rivet_class->name,
				"<Enter>", rivet_button_enter, 0, 0, 0);
	rivet_create_binding(0, widget->rivet_class->name,
				"<Leave>", rivet_button_leave, 0, 0, 0);
	if(widget->rivet_class == ButtonClass)
		rivet_create_binding(0, widget->rivet_class->name,
					"<1>", rivet_button_down, 0, 0, 0);
	else
		rivet_create_binding(0, widget->rivet_class->name,
					"<1>", rivet_button_invoke, 0, 0, 0);
	rivet_create_binding(0, widget->rivet_class->name,
				"<ButtonRelease-1>", rivet_button_up, 0, 0, 0);
	rivet_create_binding(0, widget->rivet_class->name,
				"<space>", rivet_button_invoke, 0, 0, 0);
	if(!tk_strictMotif)
		rivet_create_binding(0, widget->rivet_class->name,
				"<Return>", rivet_button_invoke, 0, 0, 0);
}

