#include "rivet.h"


/*
 *--------------------------------------------------------------
 * Following code from entry.tcl and tcl/tclCmdMZ.c
 *--------------------------------------------------------------
 */
#define KEY_SELECT_END -9999

static int
rivet_wordend(str1, index)
	char           *str1;
	int             index;
{
	int             length, cur;
	char		c;

	length = strlen(str1);
	if (index < 0) {
		index = 0;
	}
	if (index >= length) {
		cur = length;
		goto wordendDone;
	}
	for (cur = index; cur < length; cur++) {
		c = UCHAR(str1[cur]);
		if (!isalnum(c) && (c != '_')) {
			break;
		}
	}
	if (cur == index) {
		cur = index + 1;
	}
wordendDone:
	return cur;
}


static int
rivet_wordstart(str1, index)
	char           *str1;
	int             index;
{
	int             length, cur;
	char		c;

	length = strlen(str1);
	if (index >= length) {
		index = length - 1;
	}
	if (index <= 0) {
		cur = 0;
		goto wordstartDone;
	}
	for (cur = index; cur >= 0; cur--) {
		c = UCHAR(str1[cur]);
		if (!isalnum(c) && (c != '_')) {
			break;
		}
	}
	if (cur != index) {
		cur += 1;
	}
wordstartDone:
	return cur;
}



 
/*
 * tkEntrySeeInsert --
 * Make sure that the insertion cursor is visible in the entry window.
 * If not, adjust the view so that it is.
 *
 * Arguments:
 * w -		The entry window.
 */
static void
rivet_entry_see_insert(rivetobj)
	Rivetobj rivetobj;
{
#if 0
 proc tkEntrySeeInsert w {
     set c [$w index insert]
     set left [$w index @0]
     if {$left > $c} {
 	$w xview $c
 	return
     }
     set x [winfo width $w]
     while {([$w index @$x] <= $c) && ($left < $c)} {
 	incr left
 	$w xview $left
     }
 }
#endif
	int c;
	int left;
	char xbuf[20];

	c = rivet_va_cmd_int(rivetobj, "index", "insert", 0);
	left = rivet_va_cmd_int(rivetobj, "index", "@0", 0);
	if (left > c) {
		rivet_va_cmd(rivetobj, "xview", rivet_itoa(c), 0);
		return;
	}
	sprintf(xbuf, "@%d", Tk_Width(rivetobj->tkwin));
	while((rivet_va_cmd_int(rivetobj, "index", xbuf, 0) <= c) &&
		(left < c)) {
		left++;
		rivet_va_cmd(rivetobj, "xview", rivet_itoa(left), 0);
	}
}
 
/*
 * tkEntryButton1 --
 * This procedure is invoked to handle button-1 presses in entry
 * widgets.  It moves the insertion cursor, sets the selection anchor,
 * and claims the input focus.
 *
 * Arguments:
 * w -		The entry window in which the button was pressed.
 * x -		The x-coordinate of the button press.
 */
static int
rivet_entry_button1(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
#if 0
 
 proc tkEntryButton1 {w x} {
     global tkPriv
 
     set tkPriv(selectMode) char
     set tkPriv(mouseMoved) 0
     set tkPriv(pressX) $x
     $w icursor @$x
     $w selection from @$x
     if {[lindex [$w configure -state] 4] == "normal"} {focus $w}
 }
#endif
	char buf[20];
	tkPriv->selectMode = "char";
	tkPriv->mouseMoved = 0;
	tkPriv->pressX = ((XButtonEvent*)xevent)->x;
	sprintf(buf, "@%d", tkPriv->pressX);
	rivet_va_cmd(rivetobj, "icursor", buf, 0);
	rivet_va_cmd(rivetobj, "selection", "from", buf, 0);
	if(rivet_state_p(rivetobj, "normal"))
		rivet_set_focus(rivetobj);
}
 
/*
 * 
 * Call directly to avoid copying and length limits of rivet_va_cmd.
 * Warning, results are very volatile.
 *
 */
char *
rivet_entry_get(rivetobj)
	Rivetobj rivetobj;
{
	char *argv[2];

	Tcl_SetResult(rivetobj->interp, 0, 0);
	argv[0] = rivet_path(rivetobj);
	argv[1] = "get";
	if((rivetobj->rivet_class->widgetcmd)(rivetobj, rivetobj->interp,
			2, argv) == TCL_OK)
		return rivetobj->interp->result;
	else
		return 0;
}
	
/*
 * tkEntryMouseSelect --
 * This procedure is invoked when dragging out a selection with
 * the mouse.  Depending on the selection mode (character, word,
 * line) it selects in different-sized units.  This procedure
 * ignores mouse motions initially until the mouse has moved from
 * one character to another or until there have been multiple clicks.
 *
 * Arguments:
 * w -		The entry window in which the button was pressed.
 * x -		The x-coordinate of the mouse.
 */
static void
rivet_entry_mouse_select(rivetobj, x)
	Rivetobj rivetobj;
	int x;
{
 
#if 0
 proc tkEntryMouseSelect {w x} {
     global tkPriv
 
     set cur [$w index @$x]
     set anchor [$w index anchor]
     if {($cur != $anchor) || (abs($tkPriv(pressX) - $x) >= 3)} {
 	set tkPriv(mouseMoved) 1
     }
     switch $tkPriv(selectMode) {
 	char {
 	    if $tkPriv(mouseMoved) {
 		if {$cur < [$w index anchor]} {
 		    $w selection to $cur
 		} else {
 		    $w selection to [expr $cur+1]
 		}
 	    }
 	}
 	word {
 	    if {$cur < [$w index anchor]} {
 		$w selection range [string wordstart [$w get] $cur] \
 			[string wordend [$w get] [expr $anchor-1]]
 	    } else {
 		$w selection range [string wordstart [$w get] $anchor] \
 			[string wordend [$w get] $cur]
 	    }
 	}
 	line {
 	    $w selection range 0 end
 	}
     }
     update idletasks
 }
#endif
	char buf[20];
	int cur;
	int anchor;
	char *get;

	sprintf(buf, "@%d", x);

	cur = rivet_va_cmd_int(rivetobj, "index", buf, 0);
	anchor = rivet_va_cmd_int(rivetobj, "index", "anchor", 0);
	if((cur != anchor) || abs(tkPriv->pressX - x) >=3)
		tkPriv->mouseMoved = 1;

	if(streq(tkPriv->selectMode, "char")) {
	  if(tkPriv->mouseMoved)
		rivet_va_cmd(rivetobj, "selection", "to",
			((cur < anchor) ? rivet_itoa(cur) : rivet_itoa(cur+1)),
			0);
	} else
	if(streq(tkPriv->selectMode, "word")) {
 	    if (cur < anchor) {
		if(!(get = rivet_entry_get(rivetobj)))
			return;
		rivet_va_cmd(rivetobj, "selection", "range",
			rivet_itoa( rivet_wordstart(get, cur)),
			rivet_itoa( rivet_wordend(get, anchor-1)), 0);
 	    } else {
		if(!(get = rivet_entry_get(rivetobj)))
			return;
		rivet_va_cmd(rivetobj, "selection", "range",
			rivet_itoa( rivet_wordstart(get, anchor)),
			rivet_itoa( rivet_wordend(get, cur)), 0);
 	    }
	} else
	if(streq(tkPriv->selectMode, "line")) {
		rivet_va_cmd(rivetobj, "selection", "range", "0", "end", 0);
	}
	rivet_update(rivetobj, 1);
}

/* 
 * tkEntryAutoScan --
 * This procedure is invoked when the mouse leaves an entry window
 * with button 1 down.  It scrolls the window left or right,
 * depending on where the mouse is, and reschedules itself as an
 * "after" command so that the window continues to scroll until the
 * mouse moves back into the window or the mouse button is released.
 *
 * Arguments:
 * w -		The entry window.
 */
static void
rivet_entry_auto_scan(ro)
	ClientData ro;
{
#if 0
 proc tkEntryAutoScan {w} {
     global tkPriv
     set x $tkPriv(x)
     if {$x >= [winfo width $w]} {
 	$w xview scroll 2 units
 	tkEntryMouseSelect $w $x
     } elseif {$x < 0} {
 	$w xview scroll -2 units
 	tkEntryMouseSelect $w $x
     }
     set tkPriv(afterId) [after 50 tkEntryAutoScan $w]
 }
#endif
	Rivetobj rivetobj = (Rivetobj) ro;
	int x;

	x = tkPriv->x;
	if(x >= Tk_Width(rivetobj->tkwin)) {
		rivet_va_cmd(rivetobj, "xview", "scroll", "2", "units", 0);
		rivet_entry_mouse_select(rivetobj, x);
	} else 
	if(x < 0) {
		rivet_va_cmd(rivetobj, "xview", "scroll", "-2", "units", 0);
		rivet_entry_mouse_select(rivetobj, x);
	}
	tkPriv->afterId = Tk_CreateTimerHandler(50,
			rivet_entry_auto_scan, rivetobj);
}
 
/*
 * tkEntryKeySelect --
 * This procedure is invoked when stroking out selections using the
 * keyboard.  It moves the cursor to a new position, then extends
 * the selection to that position.
 *
 * Arguments:
 * w -		The entry window.
 * new -		A new position for the insertion cursor (the cursor hasn't
 *		actually been moved to this position yet).
 */
static int
rivet_entry_key_select(rivetobj, where)
	Rivetobj rivetobj;
	int where;
{
#if 0
 proc tkEntryKeySelect {w new} {
     if ![$w selection present] {
 	$w selection from insert
 	$w selection to $new
     } else {
 	$w selection adjust $new
     }
     $w icursor $new
 }
#endif
	char *new;
	if(where == KEY_SELECT_END)
		new = "end";
	else
		new = rivet_itoa(where);
	if(strfalse(rivet_va_cmd(rivetobj, "selection", "present", 0))) {
		rivet_va_cmd(rivetobj, "selection", "from", "insert", 0);
		rivet_va_cmd(rivetobj, "selection", "to", new, 0);
	} else {
		rivet_va_cmd(rivetobj, "selection", "adjust", new, 0);
	}
	rivet_va_cmd(rivetobj, "icursor", new, 0);
}
 
/*
 * tkEntryInsert --
 * Insert a string into an entry at the point of the insertion cursor.
 * If there is a selection in the entry, and it covers the point of the
 * insertion cursor, then delete the selection before inserting.
 *
 * Arguments:
 * w -		The entry window in which to insert the string
 * s -		The string to insert (usually just a single character)
 */
static void
rivet_entry_insert(rivetobj, s)
	Rivetobj rivetobj;
	char *s;
{
#if 0
 proc tkEntryInsert {w s} {
     if {$s == ""} {
 	return
     }
     catch {
 	set insert [$w index insert]
 	if {([$w index sel.first] <= $insert)
 		&& ([$w index sel.last] >= $insert)} {
 	    $w delete sel.first sel.last
 	}
     }
     $w insert insert $s
     tkEntrySeeInsert $w
 }
#endif
	int insert;
	if(streq(s, ""))
		return;
	insert = rivet_va_cmd_int(rivetobj, "index", "insert", 0);
	if((rivet_va_cmd_int(rivetobj, "index", "sel.first", 0) <= insert) &&
	   (rivet_va_cmd_int(rivetobj, "index", "sel.last", 0) >= insert)) {
		rivet_va_cmd(rivetobj, "delete", "sel.first", "sel.last", 0);
	}
	rivet_va_cmd(rivetobj, "insert", "insert", s, 0);
	rivet_entry_see_insert(rivetobj);
}
 
/*
 * tkEntryBackspace --
 * Backspace over the character just before the insertion cursor.
 * If backspacing would move the cursor off the left edge of the
 * window, reposition the cursor at about the middle of the window.
 *
 * Arguments:
 * w -		The entry window in which to backspace.
 */
static int
rivet_entry_backspace(rivetobj)
	Rivetobj rivetobj;
{
#if 0
 proc tkEntryBackspace w {
     if [$w selection present] {
 	$w delete sel.first sel.last
     } else {
 	set x [expr {[$w index insert] - 1}]
 	if {$x >= 0} {$w delete $x}
	if {[$w index @0] >= [$w index insert]} {
	    set range [$w xview]
	    set left [lindex $range 0]
	    set right [lindex $range 1]
	    $w xview moveto [expr $left - ($right - $left)/2.0]
	}
     }
 }
#endif
	int x;
	char *range;
	double left, right;
	char buf[80];

	if(strtrue(rivet_va_cmd(rivetobj, "selection", "present", 0))) {
		rivet_va_cmd(rivetobj, "delete", "sel.first", "sel.last", 0);
	} else {
		x = rivet_va_cmd_int(rivetobj, "index", "insert", 0) - 1;
		if(x >=0)
			rivet_va_cmd(rivetobj, "delete", rivet_itoa(x), 0);
	}

	if(rivet_va_cmd_int(rivetobj, "index", "@0", 0) >=
		rivet_va_cmd_int(rivetobj, "index", "insert", 0)) {
		range = rivet_va_cmd(rivetobj, "xview", 0);
		left = strtod(range, &range);
		right = strtod(range, 0);
		sprintf(buf, "%g", left - (right - left)/2.0);
		rivet_va_cmd(rivetobj, "xview", "moveto", buf, 0);
	}
}
/*
 * tkEntrySetCursor -
 * Move the insertion cursor to a given position in an entry.  Also
 * clears the selection, if there is one in the entry, and makes sure
 * that the insertion cursor is visible.
 *
 * Arguments:
 * w -		The entry window.
 * pos -		The desired new position for the cursor in the window.
 */
static int
rivet_entry_set_cursor(rivetobj, pos)
	Rivetobj rivetobj;
	char *pos;
{
#if 0
 proc tkEntrySetCursor {w pos} {
     $w icursor $pos
     $w selection clear
     tkEntrySeeInsert $w
 }
#endif
	rivet_va_cmd(rivetobj, "icursor", pos, 0);
	rivet_va_cmd(rivetobj, "selection", "clear", 0);
	rivet_entry_see_insert(rivetobj);
}
 
/*
 * tkEntryTranspose -
 * This procedure implements the "transpose" function for entry widgets.
 * It tranposes the characters on either side of the insertion cursor,
 * unless the cursor is at the end of the line.  In this case it
 * transposes the two characters to the left of the cursor.  In either
 * case, the cursor ends up to the right of the transposed characters.
 *
 * Arguments:
 * w -		The entry window.
 */
static int
rivet_entry_transpose(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
#if 0
 proc tkEntryTranspose w {
     set i [$w index insert]
     if {$i < [$w index end]} {
 	incr i
     }
     set first [expr $i-2]
     if {$first < 0} {
 	return
     }
     set new [string index [$w get] [expr $i-1]][string index [$w get] $first]
     $w delete $first $i
     $w insert insert $new
     tkEntrySeeInsert $w
 }
#endif

	char new[3];
	int i;
	int first;
	char *str1;

	i = rivet_va_cmd_int(rivetobj, "index", "insert", 0);
	if(i < rivet_va_cmd_int(rivetobj, "index", "end", 0))
		i++;
	first = i-2;
	if (first < 0)
		return;

	str1 = rivet_entry_get(rivetobj);
	new[0] = str1[i-1];
	new[1] = str1[first];
	new[2] = 0;

	rivet_va_cmd(rivetobj, "delete", rivet_itoa(first), rivet_itoa(i), 0);
	rivet_va_cmd(rivetobj, "insert", "insert", new, 0);
	rivet_entry_see_insert(rivetobj);
}
 
typedef enum {
	ecf_ONE, ecf_B1Motion, ecf_Double1, ecf_Triple1, ecf_Shift1,
	ecf_DoubleShift1, ecf_TripleShift1, ecf_B1Leave, ecf_B1Enter,
	ecf_ButtonRelease1, ecf_Control1, ecf_Left, ecf_Right,
	ecf_ShiftLeft, ecf_ShiftRight, ecf_ControlLeft, ecf_ControlRight,
	ecf_ShiftControlLeft, ecf_ShiftControlRight, ecf_Home,
	ecf_ShiftHome, ecf_End, ecf_ShiftEnd, ecf_Delete, ecf_BackSpace,
	ecf_Controlspace, ecf_Select, ecf_ControlShiftspace, ecf_ShiftSelect,
	ecf_Controlslash, ecf_Controlbackslash, ecf_KeyPress, ecf_Insert,
	ecf_Controla, ecf_Controlb, ecf_Controld, ecf_Controle, ecf_Controlf,
	ecf_Controlh, ecf_Controlk, ecf_Controlt, ecf_Metab, ecf_Metad,
	ecf_Metaf, ecf_MetaBackSpace,
	ecf_TWO, ecf_B2Motion, ecf_ButtonRelease2, ecf_Nothing,
} Entry_code_frag;
 
static int
rivet_entry_do_code_frag(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	Entry_code_frag ecf = (Entry_code_frag) assoc;
	int x, y;

	/* Don't use these if it ain't the right event! */
	/* They won't crash, but they won't be any good either */
	x = ((XMotionEvent*)xevent)->x;
	y = ((XMotionEvent*)xevent)->y;
	
	switch(ecf) {

	  case ecf_ONE: /*
	        tkEntryButton1 %W %x
		%W selection clear
		*/
		rivet_entry_button1(rivetobj, xevent, (ClientData) "clear");
		rivet_va_cmd(rivetobj, "selection", "clear", 0);
		break;

	  case ecf_B1Motion: /*
       		set tkPriv(x) %x
		tkEntryMouseSelect %W %x
		*/
		tkPriv->x = x;
		rivet_entry_mouse_select(rivetobj, x);
		break;

	  case ecf_Double1: /*
		set tkPriv(selectMode) word
		tkEntryMouseSelect %W %x
		catch {%W icursor sel.first}
		*/
		tkPriv->selectMode = "word";
		rivet_entry_mouse_select(rivetobj, x);
		(void) rivet_va_cmd(rivetobj, "icursor", "sel.first", 0);
		break;

	  case ecf_Triple1: /*
		set tkPriv(selectMode) line
		tkEntryMouseSelect %W %x
		%W icursor 0
		*/
		tkPriv->selectMode = "word";
		rivet_entry_mouse_select(rivetobj, x);
		rivet_va_cmd(rivetobj, "icursor", "0", 0);
		break;

	  case ecf_Shift1: /*
		set tkPriv(selectMode) char
		%W selection adjust @%x
		*/
		{
		char str[40];
		sprintf(str, "@%d", x);
		tkPriv->selectMode = "char";
		rivet_va_cmd(rivetobj, "selection", "adjust", str, 0);
		}
		break;

	  case ecf_DoubleShift1: /*
		set tkPriv(selectMode) word
		tkEntryMouseSelect %W %x
		*/
		tkPriv->selectMode = "word";
		rivet_entry_mouse_select(rivetobj, x);
		break;

	  case ecf_TripleShift1: /*
		set tkPriv(selectMode) line
		tkEntryMouseSelect %W %x
		*/
		tkPriv->selectMode = "line";
		rivet_entry_mouse_select(rivetobj, x);
		break;

	  case ecf_B1Leave: /*
		set tkPriv(x) %x
		tkEntryAutoScan %W
		*/
		tkPriv->x = x;
		rivet_entry_auto_scan(rivetobj);
		break;

	  case ecf_B1Enter: /*
		tkCancelRepeat
		*/
		rivet_cancel_repeat();
		break;

	  case ecf_ButtonRelease1: /*
	        tkCancelRepeat
		*/
		rivet_cancel_repeat();
		break;

	  case ecf_Control1: /*
	        %W icursor @%x
		*/
		{
		char str[40];
		sprintf(str, "@%d", x);
		rivet_va_cmd(rivetobj, "icursor", str, 0);
		}
		break;

	  case ecf_Left: /*
	        tkEntrySetCursor %W [expr [%W index insert] - 1]
		*/
		rivet_entry_set_cursor(rivetobj,
		    rivet_itoa(
			rivet_va_cmd_int(rivetobj, "index", "insert", 0) - 1));
		break;

	  case ecf_Right: /*
	        tkEntrySetCursor %W [expr [%W index insert] + 1]
		*/
		rivet_entry_set_cursor(rivetobj,
		    rivet_itoa(
			rivet_va_cmd_int(rivetobj, "index", "insert", 0) + 1));
		break;

	  case ecf_ShiftLeft: /*
	        tkEntryKeySelect %W [expr [%W index insert] - 1]
	        tkEntrySeeInsert %W
		*/
	        rivet_entry_key_select(rivetobj,
			rivet_va_cmd_int(rivetobj, "index", "insert", 0) - 1);
	        rivet_entry_see_insert(rivetobj);
		break;

	  case ecf_ShiftRight: /*
	        tkEntryKeySelect %W [expr [%W index insert] + 1]
	        tkEntrySeeInsert %W
		*/
		rivet_entry_key_select(rivetobj,
			rivet_va_cmd_int(rivetobj, "index", "insert", 0) + 1);
	        rivet_entry_see_insert(rivetobj);
		break;

	  case ecf_ControlLeft: /*
	        tkEntrySetCursor %W \
	                [string wordstart [%W get] [expr [%W index insert] - 1]]
		*/
		rivet_entry_set_cursor(rivetobj,
		    rivet_itoa(rivet_wordstart(rivet_entry_get(rivetobj),
			rivet_va_cmd_int(rivetobj, "index", "insert", 0) - 1)));
		break;

	  case ecf_ControlRight: /*
		tkEntrySetCursor %W [string wordend [%W get] [%W index insert]]
		*/
		rivet_entry_set_cursor(rivetobj,
		    rivet_itoa(rivet_wordstart(rivet_entry_get(rivetobj),
			rivet_va_cmd_int(rivetobj, "index", "insert", 0))));
		break;

	  case ecf_ShiftControlLeft: /*
		tkEntryKeySelect %W \
			[string wordstart [%W get] [expr [%W index insert] - 1]]
		tkEntrySeeInsert %W
		*/
		rivet_entry_key_select(rivetobj,
		    rivet_wordstart(rivet_entry_get(rivetobj),
			rivet_va_cmd_int(rivetobj, "index", "insert", 0) -1));
		rivet_entry_see_insert(rivetobj);
		break;

	  case ecf_ShiftControlRight: /*
		tkEntryKeySelect %W [string wordend [%W get] [%W index insert]]
		tkEntrySeeInsert %W
		*/
		rivet_entry_key_select(rivetobj,
		    rivet_wordend(rivet_entry_get(rivetobj),
			rivet_va_cmd_int(rivetobj, "index", "insert", 0)));
		rivet_entry_see_insert(rivetobj);
		break;

	  case ecf_Home: /*
		tkEntrySetCursor %W 0
		*/
		rivet_entry_set_cursor(rivetobj, "0");
		break;

	  case ecf_ShiftHome: /*
		tkEntryKeySelect %W 0
		tkEntrySeeInsert %W
		*/
		rivet_entry_key_select(rivetobj, 0);
		rivet_entry_see_insert(rivetobj);
		break;

	  case ecf_End: /*
		tkEntrySetCursor %W end
		*/
		rivet_entry_set_cursor(rivetobj, "end");
		break;

	  case ecf_ShiftEnd: /*
		tkEntryKeySelect %W end
		tkEntrySeeInsert %W
		*/
		rivet_entry_key_select(rivetobj, KEY_SELECT_END);
		rivet_entry_see_insert(rivetobj);
		break;


	  case ecf_Delete: /*
	        if [%W selection present] {
			%W delete sel.first sel.last
		} else {
			%W delete insert
	        }
		*/
		if(strtrue(rivet_va_cmd(rivetobj, "selection", "present", 0)))
			rivet_va_cmd(rivetobj,
				"delete", "sel.first", "sel.last", 0);
		else
			rivet_va_cmd(rivetobj, "delete", "insert", 0);
		break;

	  case ecf_BackSpace: /*
		tkEntryBackspace %W
		*/
		rivet_entry_backspace(rivetobj);
		break;

	  case ecf_Controlspace: /*
		%W selection from insert
		*/
		rivet_va_cmd(rivetobj, "selection", "from", "insert", 0);
		break;

	  case ecf_Select: /*
		%W selection from insert
		*/
		rivet_va_cmd(rivetobj, "selection", "from", "insert", 0);
		break;

	  case ecf_ControlShiftspace: /*
		%W selection adjust insert
		rivet_va_cmd(rivetobj, "selection", "adjust", "insert", 0);
		*/
		break;

	  case ecf_ShiftSelect: /*
		%W selection adjust insert
		*/
		rivet_va_cmd(rivetobj, "selection", "adjust", "insert", 0);
		break;

	  case ecf_Controlslash: /*
		%W selection range 0 end
		*/
		rivet_va_cmd(rivetobj, "selection", "range", "0", "end", 0);
		break;

	  case ecf_Controlbackslash: /*
		%W selection clear
		*/
		rivet_va_cmd(rivetobj, "selection", "clear", 0);
		break;

	  case ecf_KeyPress: /*
		tkEntryInsert %W %A
		*/
		rivet_entry_insert(rivetobj,
			rivet_input_str(rivetobj, xevent));
		break;

	  case ecf_Insert: /*
		catch {tkEntryInsert %W [selection get -displayof %W]}
		*/
		rivet_entry_insert(rivetobj, rivet_sel_displayof(rivetobj));
		break;

	  case ecf_Controla: /*
		tkEntrySetCursor %W 0
		*/
		rivet_entry_set_cursor(rivetobj, "0");
		break;

	  case ecf_Controlb: /*
		tkEntrySetCursor %W [expr [%W index insert] - 1]
		*/
		rivet_entry_set_cursor(rivetobj,
		    rivet_itoa(
			rivet_va_cmd_int(rivetobj, "index", "insert", 0) - 1));
		break;

	  case ecf_Controld: /*
		%W delete insert
		*/
		rivet_va_cmd(rivetobj, "delete", "insert", 0);
		break;

	  case ecf_Controle: /*
		tkEntrySetCursor %W end
		*/
		rivet_entry_set_cursor(rivetobj, "end");
		break;

	  case ecf_Controlf: /*
		tkEntrySetCursor %W [expr [%W index insert] + 1]
		*/
		rivet_entry_set_cursor(rivetobj,
		    rivet_itoa(
			rivet_va_cmd_int(rivetobj, "index", "insert", 0) + 1));
		break;

	  case ecf_Controlh: /*
		tkEntryBackspace %W
		*/
		rivet_entry_backspace(rivetobj);
		break;

	  case ecf_Controlk: /*
		%W delete insert end
		*/
		rivet_va_cmd(rivetobj, "delete", "insert", "end", 0);
		break;

	  case ecf_Controlt: /*
		tkEntryTranspose %W
		*/
		rivet_entry_transpose(rivetobj);
		break;

	  case ecf_Metab: /*
		tkEntrySetCursor %W \
			[string wordstart [%W get] [expr [%W index insert] - 1]]
		*/
		rivet_entry_set_cursor(rivetobj,
		    rivet_itoa(rivet_wordstart(rivet_entry_get(rivetobj),
			rivet_va_cmd_int(rivetobj, "index", "insert", 0) - 1)));
		break;

	  case ecf_Metad: /*
		%W delete insert [string wordend [%W get] [%W index insert]]
		*/
		rivet_va_cmd(rivetobj, "delete", "insert",
			rivet_itoa(rivet_wordend(rivet_entry_get(rivetobj),
			rivet_va_cmd_int(rivetobj, "index", "insert", 0))),0);
		break;

	  case ecf_Metaf: /*
		tkEntrySetCursor %W [string wordend [%W get] [%W index insert]]
		*/
		rivet_entry_set_cursor(rivetobj,
		    rivet_itoa(rivet_wordstart(rivet_entry_get(rivetobj),
			rivet_va_cmd_int(rivetobj, "index", "insert", 0))));
		break;

	  case ecf_MetaBackSpace: /*
		%W delete [string wordstart [%W get] \
			[expr [%W index insert] - 1]] insert
		*/
		rivet_va_cmd(rivetobj, "delete", rivet_itoa(
		  rivet_wordstart(rivet_entry_get(rivetobj),
		    rivet_va_cmd_int(rivetobj, "index", "insert", 0) -1)),
				"insert", 0);
		break;

	  case ecf_TWO: /*
		%W scan mark %x
		set tkPriv(x) %x
		set tkPriv(y) %y
		set tkPriv(mouseMoved) 0
		*/
		rivet_va_cmd(rivetobj, "scan", "mark", rivet_itoa(x), 0);
		tkPriv->x = x;
		tkPriv->y = y;
		tkPriv->mouseMoved = 0;
		break;

	  case ecf_B2Motion: /*
		if {abs(%x-$tkPriv(x)) > 2} {
			set tkPriv(mouseMoved) 1
		}
		%W scan dragto %x
		*/
		if(abs(x - tkPriv->x) > 2)
			tkPriv->mouseMoved = 1;
		rivet_va_cmd(rivetobj, "scan", "dragto", rivet_itoa(x), 0);
		break;

	  case ecf_ButtonRelease2: /*
		if !$tkPriv(mouseMoved) {
			catch {
				%W insert insert [selection get -displayof %W]
				tkEntrySeeInsert %W
			}
		}   
		*/
		if(tkPriv->mouseMoved) {
			char buf[40];
			sprintf(buf, "@%d", xevent->xkey.x);
			rivet_va_cmd(rivetobj, "insert", buf,
				rivet_sel_displayof(rivetobj), 0);
		}
		break;

	  case ecf_Nothing:
		break;
	}

	return TCL_CONTINUE;
}


#define frag_bind(_pat_, _frag_)				\
	if(rivet_create_binding(0, widget->rivet_class->name,	\
		_pat_, rivet_entry_do_code_frag,		\
		(ClientData) _frag_, 0, 0) == TCL_ERROR)	\
			goto error;

void
rivet_entry_bind(widget)
	Rivetobj	widget;
{
	if(rivet_is_class_initialized(widget->rivet_class))
		return;
	rivet_class_init_done(widget->rivet_class);

	frag_bind("<1>", ecf_ONE);
	frag_bind("<B1-Motion>", ecf_B1Motion);
	frag_bind("<Double-1>", ecf_Double1);
	frag_bind("<Triple-1>", ecf_Triple1);
	frag_bind("<Shift-1>", ecf_Shift1);
	frag_bind("<Double-Shift-1>", ecf_DoubleShift1);
	frag_bind("<Triple-Shift-1>", ecf_TripleShift1);
	frag_bind("<B1-Leave>", ecf_B1Leave);
	frag_bind("<B1-Enter>", ecf_B1Enter);
	frag_bind("<ButtonRelease-1>", ecf_ButtonRelease1);
	frag_bind("<Control-1>", ecf_Control1);
	frag_bind("<Left>", ecf_Left);
	frag_bind("<Right>", ecf_Right);
	frag_bind("<Shift-Left>", ecf_ShiftLeft);
	frag_bind("<Shift-Right>", ecf_ShiftRight);
	frag_bind("<Control-Left>", ecf_ControlLeft);
	frag_bind("<Control-Right>", ecf_ControlRight);
	frag_bind("<Shift-Control-Left>", ecf_ShiftControlLeft);
	frag_bind("<Shift-Control-Right>", ecf_ShiftControlRight);
	frag_bind("<Home>", ecf_Home);
	frag_bind("<Shift-Home>", ecf_ShiftHome);
	frag_bind("<End>", ecf_End);
	frag_bind("<Shift-End>", ecf_ShiftEnd);
	frag_bind("<Delete>", ecf_Delete);
	frag_bind("<BackSpace>", ecf_BackSpace);
	frag_bind("<Control-space>", ecf_Controlspace);
	frag_bind("<Select>", ecf_Select);
	frag_bind("<Control-Shift-space>", ecf_ControlShiftspace);
	frag_bind("<Shift-Select>", ecf_ShiftSelect);
	frag_bind("<Control-slash>", ecf_Controlslash);
	frag_bind("<Control-backslash>", ecf_Controlbackslash);
	frag_bind("<KeyPress>", ecf_KeyPress);
	frag_bind("<Insert>", ecf_Insert);

	if(!tk_strictMotif) {
		frag_bind("<Control-a>", ecf_Controla);
		frag_bind("<Control-b>", ecf_Controlb);
		frag_bind("<Control-d>", ecf_Controld);
		frag_bind("<Control-e>", ecf_Controle);
		frag_bind("<Control-f>", ecf_Controlf);
		frag_bind("<Control-h>", ecf_Controlh);
		frag_bind("<Control-k>", ecf_Controlk);
		frag_bind("<Control-t>", ecf_Controlt);
		frag_bind("<Meta-b>", ecf_Metab);
		frag_bind("<Meta-d>", ecf_Metad);
		frag_bind("<Meta-f>", ecf_Metaf);
		frag_bind("<Meta-BackSpace>", ecf_MetaBackSpace);
    
		/* # A few additional bindings of my own. */
    
		frag_bind("<2>", ecf_TWO);
		frag_bind("<B2-Motion>", ecf_B2Motion);
		frag_bind("<ButtonRelease-2>", ecf_ButtonRelease2);
	}

	if(rivet_bind_clipboard(widget, "<F16>", "<F20>", "<F18>") == TCL_ERROR)
		goto error;

	/*
	 * Ignore all Alt, Meta, and Control keypresses unless explicitly bound.
	 * Otherwise, if a widget binding for one of these is defined, the
	 * <KeyPress> class binding will also fire and insert the character,
	 * which is wrong.  Ditto for Escape, Return, and Tab.
	 */

	frag_bind("<Alt-KeyPress>", ecf_Nothing);
	frag_bind("<Meta-KeyPress>", ecf_Nothing);
	frag_bind("<Control-KeyPress>", ecf_Nothing);
	frag_bind("<Escape>", ecf_Nothing);
	frag_bind("<Return>", ecf_Nothing);
	frag_bind("<KP_Enter>", ecf_Nothing);
	frag_bind("<Tab>", ecf_Nothing);

	if(!tk_strictMotif)
		if(rivet_bind_clipboard(widget,
			"<Meta-w>", "<Control-w>", "<Control-y>") == TCL_ERROR)
				goto error;

	error:
		return;

}

