
#include "rivet.h"
/*
 *--------------------------------------------------------------
 * Following code from focus.tcl.
 *--------------------------------------------------------------
 */
static int rivet_focus_ok _ANSI_ARGS_((Rivetobj));

/* utility functions */
static int
index_of_child(list, child)
	TkWindow *list;
	Rivetobj child;
{
	int i;
	for(i=0; list!=child->tkwin; i++)
		list=list->nextPtr;
	return i;
}

static Rivetobj
find_child(list, i)
	TkWindow *list;
	int i;
{
	int j;

	for(j=0; j<i; j++) {
		list = list->nextPtr;
		if(!list)
			return 0;
	}
	return list->rivetobj;
}

static int
len_childlist(list)
	TkWindow *list;
{
	int i;
	for(i=0; list; i++)
		list=list->nextPtr;
	return i;
}
 

/*
 * tk_focusNext --
 * This procedure returns the name of the next window after "w" in
 * "focus order" (the window that should receive the focus next if
 * Tab is typed in w).  "Next" is defined by a pre-order search
 * of a top-level and its non-top-level descendants, with the stacking
 * order determining the order of siblings.  The "-takefocus" options
 * on windows determine whether or not they should be skipped.
 *
 * Arguments:
 * w -		Name of a window.
 */
Rivetobj
rivet_focus_next(rivetobj)
	Rivetobj rivetobj;
{
#if 0
 proc tk_focusNext w {
     set cur $w
     while 1 {
 
 	# Descend to just before the first child of the current widget.
 
 	set parent $cur
 	set children [winfo children $cur]
 	set i -1
 
 	# Look for the next sibling that isn't a top-level.
 
 	while 1 {
 	    incr i
 	    if {$i < [llength $children]} {
 		set cur [lindex $children $i]
 		if {[winfo toplevel $cur] == $cur} {
 		    continue
 		} else {
 		    break
 		}
 	    }
 
 	    # No more siblings, so go to the current widget's parent.
 	    # If it's a top-level, break out of the loop, otherwise
 	    # look for its next sibling.
 
 	    set cur $parent
 	    if {[winfo toplevel $cur] == $cur} {
 		break
 	    }
 	    set parent [winfo parent $parent]
 	    set children [winfo children $parent]
 	    set i [lsearch -exact $children $cur]
 	}
 	if {($cur == $w) || [tkFocusOK $cur]} {
 	    return $cur
 	}
     }
 }
#endif
	int i;
	Rivetobj cur;
	Rivetobj parent;
	TkWindow *children;

	cur = rivetobj;
	while(1) {
		parent = cur;
		children = cur->tkwin->childList;
		i = -1;

		while(1) {
			i++;
			if(i < len_childlist(children)) {
				cur = find_child(children, i);
				if(rivet_to_toplevel(cur) == cur)
					continue;
				else
					break;
			}

			cur = parent;
			if(rivet_to_toplevel(cur) == cur)
				break;
			parent = rivet_parent(parent);
			children = parent->tkwin->childList;
			i = index_of_child(children, cur);
		}
		if((cur == rivetobj) || (rivet_focus_ok(cur)))
			return cur;
	}
}
 
/*
 * tk_focusPrev --
 * This procedure returns the name of the previous window before "w" in
 * "focus order" (the window that should receive the focus next if
 * Shift-Tab is typed in w).  "Next" is defined by a pre-order search
 * of a top-level and its non-top-level descendants, with the stacking
 * order determining the order of siblings.  The "-takefocus" options
 * on windows determine whether or not they should be skipped.
 *
 * Arguments:
 * w -		Name of a window.
 */
Rivetobj
rivet_focus_prev(rivetobj)
	Rivetobj rivetobj;
{
#if 0
 proc tk_focusPrev w {
     set cur $w
     while 1 {
 
 	* Collect information about the current window's position
 	* among its siblings.  Also, if the window is a top-level,
 	* then reposition to just after the last child of the window.
     
 	if {[winfo toplevel $cur] == $cur}  {
 	    set parent $cur
 	    set children [winfo children $cur]
 	    set i [llength $children]
 	} else {
 	    set parent [winfo parent $cur]
 	    set children [winfo children $parent]
 	    set i [lsearch -exact $children $cur]
 	}
 
 	* Go to the previous sibling, then descend to its last descendant
 	* (highest in stacking order.  While doing this, ignore top-levels
 	* and their descendants.  When we run out of descendants, go up
 	* one level to the parent.
 
 	while {$i > 0} {
 	    incr i -1
 	    set cur [lindex $children $i]
 	    if {[winfo toplevel $cur] == $cur} {
 		continue
 	    }
 	    set parent $cur
 	    set children [winfo children $parent]
 	    set i [llength $children]
 	}
 	set cur $parent
 	if {($cur == $w) || [tkFocusOK $cur]} {
 	    return $cur
 	}
     }
 }
#endif

	int i;
	Rivetobj cur;
	Rivetobj parent;
	TkWindow *children;
	TkWindow *tmp;

	cur = rivetobj;
	while(1) {
		if(rivet_to_toplevel(cur) == cur) {
			parent = cur;
			children = cur->tkwin->childList;
			i = len_childlist(children);
		} else {
			parent = rivet_parent(cur);
			children = parent->tkwin->childList;
			i = index_of_child(children, cur);
		}

		while(i > 0) {
			i--;
			cur = find_child(children, i);
			if(cur && (rivet_to_toplevel(cur) == cur))
				continue;
			parent = cur;
			children = parent->tkwin->childList;
			i = len_childlist(children);
		}
		cur = parent;
		if((cur == rivetobj) || rivet_focus_ok(cur))
			return cur;
	}
}
/*
 * tkFocusOK --
 *
 * This procedure is invoked to decide whether or not to focus on
 * a given window.  It returns 1 if it's OK to focus on the window,
 * 0 if it's not OK.  The code first checks whether the window is
 * viewable.  If not, then it never focuses on the window.  Then it
 * checks the -takefocus option for the window and uses it if it's
 * set.  If there's no -takefocus option, the procedure checks to
 * see if (a) the widget isn't disabled, and (b) it has some key
 * bindings.  If all of these are true, then 1 is returned.
 *
 * Arguments:
 * w -		Name of a window.
 */
static int
rivet_focus_ok(rivetobj)
	Rivetobj rivetobj;
{
#if 0 
 proc tkFocusOK w {
     if {![winfo viewable $w]} {
 	return 0
     }
     set code [catch {$w cget -takefocus} value]
     if {($code == 0) && ($value != "")} {
 	if {$value == 0} {
 	    return 0
 	} elseif {$value == 1} {
 	    return 1
 	} else {
 	    set value [uplevel #0 $value $w]
 	    if {$value != ""} {
 		return $value
 	    }
 	}
     }
     set code [catch {$w cget -state} value]
     if {($code == 0) && ($value == "disabled")} {
 	return 0
     }
     regexp Key|Focus "[bind $w] [bind [winfo class $w]]"
 }
#endif

	char *value;
	char *bindings;
	int disabled;

	if(strfalse(rivet_va_func(rivetobj, Tk_WinfoCmd, "viewable",
			rivet_path(rivetobj), 0)))
		return 0;
	value = rivet_get(rivetobj, "-takefocus");
	if(!rivet_error_occured && value && value[0]) {
		if(strfalse(value))
			return 0;
		if(strtrue(value))
			return 1;
		/* gack.
		 * I'm never going to bother with calling outside
		 * functions.  takefocus must be either a 1 or 0,
		 * but never a callback.  It just seems like overkill.
		 */
	}
	disabled = rivet_state_p(rivetobj, "disabled");
	if(!rivet_error_occured && disabled)
		return 0;

	/* gack.
	 * This needs to be cleaned up.  First get the bindings from
	 * rivet_get_all_bindings, then grep for "Key|Focus".
	 */
	bindings = rivet_get_all_bindings(0, (char*)rivetobj->tkwin->classUid);
	if(!bindings)
		return 0;
	if(Tcl_RegExpExec(rivetobj->interp, 
		Tcl_RegExpCompile(rivetobj->interp, "Key|Focus"),
		bindings, bindings) == 1)
		return 1;
	return 0;
}
 
/*
 * tk_focusFollowsMouse --
 *
 * If this procedure is invoked, Tk will enter "focus-follows-mouse"
 * mode, where the focus is always on whatever window contains the
 * mouse.  If this procedure isn't invoked, then the user typically
 * has to click on a window to give it the focus.
 *
 * Arguments:
 * None.
 */
static int
follow_mouse_focus(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent	*xevent;
	ClientData assoc;
{
	int number;

	number = 0;
	if((xevent->type == FocusIn) ||
	   (xevent->type == FocusOut))
		number = xevent->xfocus.detail;
	else
	if((xevent->type == EnterNotify) ||
	   (xevent->type == LeaveNotify))
		number = xevent->xcrossing.detail;

	if((number == NotifyAncestor) ||
	   (number == NotifyNonlinear) ||
	   (number == NotifyInferior))
		rivet_set_focus(rivetobj);

	return TCL_CONTINUE;
}

void
rivet_focus_follows_mouse(ref)
	Rivetobj ref; /* any widget, including the main one. */
{
#if 0
 proc tk_focusFollowsMouse {} {
     set old [bind all <Enter>]
     set script {
 	if {("%d" == "NotifyAncestor") || ("%d" == "NotifyNonlinear")
 		|| ("%d" == "NotifyInferior")} {
 	    focus %W
 	}
     }
     if {$old != ""} {
 	bind all <Enter> "$old; $script"
     } else {
 	bind all <Enter> $script
     }
 }
#endif
	int append;
	int n;

	if(rivet_get_binding(0, "all", "<Enter>", &n))
		append = 1;
	else
		append = 0;


	rivet_create_binding(0, "all",
		"<Enter>", follow_mouse_focus, 0, append, 0);
}

