#include "rivet.h"

/*
 *--------------------------------------------------------------
 * Following code from listbox.tcl.
 *--------------------------------------------------------------
 */
static char active_string[] = "active";
static char anchor_string[] = "anchor";
static char clear_string[] = "clear";
static char index_string[] = "index";
static char xview_string[] = "xview";
static char yview_string[] = "yview";
static char pos1_string[] = "1";
static char neg1_string[] = "-1";
static char pages_string[] = "pages";
static char units_string[] = "units";
static char scroll_string[] = "scroll";
static char selection_string[] = "selection";
#define ACTIVE active_string
#define ANCHOR anchor_string
#define CLEAR clear_string
#define INDEX index_string
#define XVIEW xview_string
#define YVIEW yview_string
#define POS1 pos1_string
#define NEG1 neg1_string
#define PAGES pages_string
#define UNITS units_string
#define SCROLL scroll_string
#define SELECTION selection_string

static void rivet_lb_autoscan _ANSI_ARGS_((Rivetobj));

static char *
index_at(w, xevent)
	Rivetobj w;
	XEvent *xevent;
{
	char buf[20];
	sprintf(buf, "@%d,%d",
		((XMotionEvent*)xevent)->x, ((XMotionEvent*)xevent)->y);
	/* be careful to copy result */
	return rivet_va_cmd(w, INDEX, buf, 0);
}

static char *
index_active(w)
	Rivetobj w;
{
	/* be careful to copy result */
	return rivet_va_cmd(w, INDEX, ACTIVE, 0);
}

static int
includes_p(w, what)
	Rivetobj w;
	char *what;
{
	return strtrue(rivet_va_cmd(w, SELECTION, "includes", what, 0));
}

static void
clear_set(w, where)
	Rivetobj w;
	char *where;
{
	rivet_va_cmd(w, SELECTION, CLEAR, "0", "end", 0);
	rivet_va_cmd(w, SELECTION, "set", where, 0);
}

static int
selectmode_p(w, expected)
	Rivetobj w;
	char *expected;
{
	return streq(rivet_get(w, "-selectmode"), expected);
}

static int
int_in_list(lookfor)
	int lookfor;
{
	/*
	if {[lsearch $tkPriv(listboxSelection) $i] >= 0}
	*/
	int i;
	if(!tkPriv->listboxSelection)
		return 0;
	for(i=0;tkPriv->listboxSelection[i] != -1; i++) {
		if(tkPriv->listboxSelection[i] == lookfor)
			return 1;
	}
	return 0;
}

/*
 * tkListboxBeginSelect --
 *
 * This procedure is typically invoked on button-1 presses.  It begins
 * the process of making a selection in the listbox.  Its exact behavior
 * depends on the selection mode currently in effect for the listbox;
 * see the Motif documentation for details.
 *
 * Arguments:
 * w -		The listbox widget.
 * el -		The element for the selection operation (typically the
 *		one under the pointer).  Must be in numerical form.
 */

static void
rivet_lb_begin_select(w, el_in)
	Rivetobj w;
	char *el_in;
{
#if 0
 proc tkListboxBeginSelect {w el} {
    global tkPriv
    if {[$w cget -selectmode]  == "multiple"} {
	if [$w selection includes $el] {
	    $w selection clear $el
	} else {
	    $w selection set $el
	}
    } else {
	$w selection clear 0 end
	$w selection set $el
	$w selection anchor $el
	set tkPriv(listboxSelection) {}
	set tkPriv(listboxPrev) $el
    }
 }
#endif
	char el[20];

	strncpy(el, el_in, sizeof(el));
	if(selectmode_p(w, "multiple")) {
		if(includes_p(w, el))
			rivet_va_cmd(w, SELECTION, CLEAR, el, 0);
		else
			rivet_va_cmd(w, SELECTION, "set", el, 0);
	} else {
		clear_set(w, el);
		rivet_va_cmd(w, SELECTION, ANCHOR, el, 0);
		if(tkPriv->listboxSelection)
			free(tkPriv->listboxSelection);
		tkPriv->listboxSelection = 0;
		strncpy(tkPriv->listboxPrev, el, sizeof(tkPriv->listboxPrev)-1);
	}
}

/*
 * tkListboxMotion --
 *
 * This procedure is called to process mouse motion events while
 * button 1 is down.  It may move or extend the selection, depending
 * on the listbox's selection mode.
 *
 * Arguments:
 * w -		The listbox widget.
 * el -		The element under the pointer (must be a number).
 */
static void
rivet_lb_motion(w, el_in)
	Rivetobj w;
	char *el_in;
{
#if 0
 proc tkListboxMotion {w el} {
    global tkPriv
    if {$el == $tkPriv(listboxPrev)} {
	return
    }
    set anchor [$w index anchor]
    switch [$w cget -selectmode] {
	browse {
	    $w selection clear 0 end
	    $w selection set $el
	    set tkPriv(listboxPrev) $el
	}
	extended {
	    set i $tkPriv(listboxPrev)
	    if [$w selection includes anchor] {
		$w selection clear $i $el
		$w selection set anchor $el
	    } else {
		$w selection clear $i $el
		$w selection clear anchor $el
	    }
	    while {($i < $el) && ($i < $anchor)} {
		if {[lsearch $tkPriv(listboxSelection) $i] >= 0} {
		    $w selection set $i
		}
		incr i
	    }
	    while {($i > $el) && ($i > $anchor)} {
		if {[lsearch $tkPriv(listboxSelection) $i] >= 0} {
		    $w selection set $i
		}
		incr i -1
	    }
	    set tkPriv(listboxPrev) $el
	}
    }
 }
#endif
	char *i_str;
	char anchor[20];
	int i;
	int el_int, anchor_int;
	char el[20];

	strncpy(el, el_in, sizeof(el));
	if(streq(el, tkPriv->listboxPrev))
		return;
	strncpy(anchor, rivet_va_cmd(w, INDEX, ANCHOR, 0), sizeof(anchor));
	if(selectmode_p(w, "browse")) {
		clear_set(w, el);
		strncpy(tkPriv->listboxPrev, el, sizeof(tkPriv->listboxPrev)-1);
	} else
	if(selectmode_p(w, "extended")) {
		i_str = tkPriv->listboxPrev;
		i = atoi(i_str);
		if(includes_p(w, ANCHOR)) {
			rivet_va_cmd(w, SELECTION, CLEAR, i_str, el, 0);
			rivet_va_cmd(w, SELECTION, "set", ANCHOR, el, 0);
		} else {
			rivet_va_cmd(w, SELECTION, CLEAR, i_str, el, 0);
			rivet_va_cmd(w, SELECTION, CLEAR, ANCHOR, el, 0);
		}
		anchor_int = atoi(anchor);
		el_int = atoi(el);
		while((i < el_int) && (i < anchor_int)) {
			if(int_in_list(i))
			    rivet_va_cmd(w, SELECTION, "set", rivet_itoa(i), 0);
			i++;
		}
		while((i > el_int) && (i > anchor_int)) {
			if(int_in_list(i))
			    rivet_va_cmd(w, SELECTION, "set", rivet_itoa(i), 0);
			i--;
		}
		strncpy(tkPriv->listboxPrev, el, sizeof(tkPriv->listboxPrev)-1);
	}
}

/*
 * tkListboxBeginExtend --
 *
 * This procedure is typically invoked on shift-button-1 presses.  It
 * begins the process of extending a selection in the listbox.  Its
 * exact behavior depends on the selection mode currently in effect
 * for the listbox;  see the Motif documentation for details.
 *
 * Arguments:
 * w -		The listbox widget.
 * el -		The element for the selection operation (typically the
 *		one under the pointer).  Must be in numerical form.
 */
static void
rivet_lb_begin_extend(w, el_in)
	Rivetobj w;
	char *el_in;
{
#if 0
 proc tkListboxBeginExtend {w el} {
    if {([$w cget -selectmode] == "extended")
	    && [$w selection includes anchor]} {
	tkListboxMotion $w $el
    }
 }
#endif
	char el[20];
	strncpy(el, el_in, sizeof(el));
	if(selectmode_p(w, "extended") && includes_p(w, ANCHOR))
		rivet_lb_motion(w, el);
}

/*
 * tkListboxBeginToggle --
 *
 * This procedure is typically invoked on control-button-1 presses.  It
 * begins the process of toggling a selection in the listbox.  Its
 * exact behavior depends on the selection mode currently in effect
 * for the listbox;  see the Motif documentation for details.
 *
 * Arguments:
 * w -		The listbox widget.
 * el -		The element for the selection operation (typically the
 *		one under the pointer).  Must be in numerical form.
 */
static void
rivet_lb_begin_toggle(w, el_in)
	Rivetobj w;
	char *el_in;
{
#if 0
 proc tkListboxBeginToggle {w el} {
    global tkPriv
    if {[$w cget -selectmode] == "extended"} {
	set tkPriv(listboxSelection) [$w curselection]
	set tkPriv(listboxPrev) $el
	$w selection anchor $el
	if [$w selection includes $el] {
	    $w selection clear $el
	} else {
	    $w selection set $el
	}
    }
 }
#endif
	char el[20];
	strncpy(el, el_in, sizeof(el));
	if(selectmode_p(w, "extended")) {
		if(tkPriv->listboxSelection)
			free(tkPriv->listboxSelection);
		tkPriv->listboxSelection = rivet_listbox_curselection(w);
		strncpy(tkPriv->listboxPrev, el, sizeof(tkPriv->listboxPrev)-1);
		rivet_va_cmd(w, SELECTION, ANCHOR, el, 0);
		if(includes_p(w, el))
			rivet_va_cmd(w, SELECTION, CLEAR, el, 0);
		else
			rivet_va_cmd(w, SELECTION, "set", el, 0);
	}
}

/*
 * tkListboxAutoScan --
 * This procedure is invoked when the mouse leaves an entry window
 * with button 1 down.  It scrolls the window up, down, left, or
 * right, depending on where the mouse left the window, and reschedules
 * itself as an "after" command so that the window continues to scroll until
 * the mouse moves back into the window or the mouse button is released.
 *
 * Arguments:
 * w -		The entry window.
 */
static void
rivet_lb_timer(cd)
	ClientData cd;
{
	Rivetobj w = (Rivetobj) cd;
	rivet_lb_autoscan(w);
}

static void
rivet_lb_autoscan(w)
	Rivetobj w;
{
#if 0
 proc tkListboxAutoScan {w} {
    global tkPriv
    set x $tkPriv(x)
    set y $tkPriv(y)
    if {$y >= [winfo height $w]} {
	$w yview scroll 1 units
    } elseif {$y < 0} {
	$w yview scroll -1 units
    } elseif {$x >= [winfo width $w]} {
	$w xview scroll 2 units
    } elseif {$x < 0} {
	$w xview scroll -2 units
    } else {
	return
    }
    tkListboxMotion $w [$w index @$x,$y]
    set tkPriv(afterId) [after 50 tkListboxAutoScan $w $x $y]
 }
#endif
	int x, y;
	XEvent phony_xevent;

	x = tkPriv->x;
	y = tkPriv->y;

	if(y >= Tk_Height(w->tkwin))
		rivet_va_cmd(w, YVIEW, SCROLL, POS1, UNITS, 0);
	else
	if (y < 0)
		rivet_va_cmd(w, YVIEW, SCROLL, NEG1, UNITS, 0);
	else
	if(x >= Tk_Width(w->tkwin))
		rivet_va_cmd(w, XVIEW, SCROLL, "2", UNITS, 0);
	else
	if(x < 0)
		rivet_va_cmd(w, XVIEW, SCROLL, "-2", UNITS, 0);
	else
		return;
	phony_xevent.xmotion.x = x;
	phony_xevent.xmotion.y = y;
	rivet_lb_motion(w, index_at(w, &phony_xevent));

	tkPriv->afterId = Tk_CreateTimerHandler(40,
			rivet_lb_timer, (ClientData)w);

}

/*
 * tkListboxUpDown --
 *
 * Moves the location cursor (active element) up or down by one element,
 * and changes the selection if we're in browse or extended selection
 * mode.
 *
 * Arguments:
 * w -		The listbox widget.
 * amount -	+1 to move down one item, -1 to move back one item.
 */
static void
rivet_lb_updown(w, amount)
	Rivetobj w;
	int amount;
{
#if 0
 proc tkListboxUpDown {w amount} {
    global tkPriv
    $w activate [expr [$w index active] + $amount]
    $w see active
    switch [$w cget -selectmode] {
	browse {
	    $w selection clear 0 end
	    $w selection set active
	}
	extended {
	    $w selection clear 0 end
	    $w selection set active
	    $w selection anchor active
	    set tkPriv(listboxPrev) [$w index active]
	    set tkPriv(listboxSelection) {}
	}
    }
 }
#endif
	int value;

	value = atoi(index_active(w)) + amount;
	rivet_va_cmd(w, "activate", rivet_itoa(value), 0);
	rivet_va_cmd(w, "see", ACTIVE, 0);
	if(selectmode_p(w, "browse"))
		clear_set(w, ACTIVE);
	else if(selectmode_p(w, "extended")) {
		clear_set(w, ACTIVE);
		rivet_va_cmd(w, SELECTION, ANCHOR, ACTIVE, 0);
		strncpy(tkPriv->listboxPrev, index_active(w), sizeof(tkPriv->listboxPrev)-1);
		if(tkPriv->listboxSelection)
			free(tkPriv->listboxSelection);
		tkPriv->listboxSelection = 0;
	}
}

/*
 * tkListboxExtendUpDown --
 *
 * Does nothing unless we're in extended selection mode;  in this
 * case it moves the location cursor (active element) up or down by
 * one element, and extends the selection to that point.
 *
 * Arguments:
 * w -		The listbox widget.
 * amount -	+1 to move down one item, -1 to move back one item.
 */
static void
rivet_lb_extend_updown(w, amount)
	Rivetobj w;
	int amount;
{
#if 0
 proc tkListboxExtendUpDown {w amount} {
    if {[$w cget -selectmode] != "extended"} {
	return
    }
    $w activate [expr [$w index active] + $amount]
    $w see active
    tkListboxMotion $w [$w index active]
 }
#endif
	int value;

	if(!selectmode_p(w, "extended")) 
		return;
	value = atoi(index_active(w)) + amount;
	rivet_va_cmd(w, "activate", rivet_itoa(value), 0);
	rivet_va_cmd(w, "see", ACTIVE, 0);
	rivet_lb_motion(w, index_active(w));
}

/*
 * tkListboxDataExtend
 *
 * This procedure is called for key-presses such as Shift-KEndData.
 * If the selection mode isn't multiple or extend then it does nothing.
 * Otherwise it moves the active element to el and, if we're in
 * extended mode, extends the selection to that point.
 *
 * Arguments:
 * w -		The listbox widget.
 * el -		An integer element number.
 */
static void
rivet_lb_data_extend(w, el)
	Rivetobj w;
	char * el;
{
#if 0
 proc tkListboxDataExtend {w el} {
    set mode [$w cget -selectmode]
    if {$mode == "extended"} {
	$w activate $el
	$w see $el
        if [$w selection includes anchor] {
	    tkListboxMotion $w $el
	}
    } elseif {$mode == "multiple"} {
	$w activate $el
	$w see $el
    }
 }
#endif

	if(selectmode_p(w, "extended")) {
		rivet_va_cmd(w, "activate", el, 0);
		rivet_va_cmd(w, "see", el, 0);
		if(includes_p(w, ANCHOR))
			rivet_lb_motion(w, el);
	} else
	if(selectmode_p(w, "multiple")) {
		rivet_va_cmd(w, "activate", el, 0);
		rivet_va_cmd(w, "see", el, 0);
	}
}

/*
 * tkListboxCancel
 *
 * This procedure is invoked to cancel an extended selection in
 * progress.  If there is an extended selection in progress, it
 * restores all of the items between the active one and the anchor
 * to their previous selection state.
 *
 * Arguments:
 * w -		The listbox widget.
 */

static void
rivet_lb_cancel(w)
	Rivetobj w;
{
#if 0
 proc tkListboxCancel w {
    global tkPriv
    if {[$w cget -selectmode] != "extended"} {
	return
    }
    set first [$w index anchor]
    set last $tkPriv(listboxPrev)
    if {$first > $last} {
	set tmp $first
	set first $last
	set last $tmp
    }
    $w selection clear $first $last
    while {$first <= $last} {
	if {[lsearch $tkPriv(listboxSelection) $first] >= 0} {
	    $w selection set $first
	}
	incr first
    }
 }
#endif
	int first, last;

	if(!selectmode_p(w, "extended"))
		return;
	first = rivet_va_cmd_int(w, INDEX, ANCHOR, 0);
	last = atoi(tkPriv->listboxPrev);
	if(first <= last) {
		int tmp;
		tmp = first;
		first = last;
		last = tmp;
	}
	rivet_va_cmd(w, SELECTION, CLEAR,
			rivet_itoa(first), rivet_itoa(last), 0);
	while(first <= last) {
		if(int_in_list(first))
			rivet_va_cmd(w, SELECTION, "set", rivet_itoa(first), 0);
		first++;
	}
}

/*
 * tkListboxSelectAll
 *
 * This procedure is invoked to handle the "select all" operation.
 * For single and browse mode, it just selects the active element.
 * Otherwise it selects everything in the widget.
 *
 * Arguments:
 * w -		The listbox widget.
 */
static void
rivet_lb_select_all(w)
	Rivetobj w;
{
#if 0
 proc tkListboxSelectAll w {
    set mode [$w cget -selectmode]
    if {($mode == "single") || ($mode == "browse")} {
	$w selection clear 0 end
	$w selection set active
    } else {
	$w selection set 0 end
    }
 }
#endif
	if(selectmode_p(w, "single") || selectmode_p(w, "browse"))
		clear_set(w, ACTIVE);
	else
		rivet_va_cmd(w, SELECTION, "set", "0", "end", 0);
}

 
/*
 * tkListboxBind --
 * This procedure is invoked the first time the mouse enters a listbox
 * widget or a listbox widget receives the input focus.  It creates
 * all of the class bindings for listboxes.
 *
 * Arguments:
 * event -	Indicates which event caused the procedure to be invoked
 *		(Enter or FocusIn).  It is used so that we can carry out
 *		the functions of that event in addition to setting up
 *		bindings.
 * ---------------------------------------------------------------------- 
 * This function is odd because it does both the Binding and the callbacks.
 * The goal is to keep the short sections of code close to the string that
 * it binds to.  Case numbers are intentionally sequential.
 */

#define binding(_pat_, _n_)						\
	case _n_:							\
	if(!rivet_is_class_initialized(w->rivet_class)) {		\
	   return rivet_create_binding(0, w->rivet_class->name, 	\
				_pat_, listbox_double_duty_event_handler,\
				(ClientData) assoc, 0, 0);		\
	}


static int
listbox_double_duty_event_handler(w, xevent, assoc)
	Rivetobj w;
	XEvent *xevent;
	ClientData assoc;
{
	
	switch( (int) assoc) {
	/* Standard Motif Bindings: */
	binding("<1>", 1)
		/*
		tkListboxBeginSelect %W [%W index @%x,%y]
		*/
		rivet_lb_begin_select(w, index_at(w, xevent));
		break;

	binding("<B1-Motion>", 2)
		/*
		set tkPriv(x) %x
		set tkPriv(y) %y
		tkListboxMotion %W [%W index @%x,%y]
		*/
		tkPriv->x = xevent->xmotion.x;
		tkPriv->y = xevent->xmotion.y;
		rivet_lb_motion(w, index_at(w, xevent));
		break;

	binding("<ButtonRelease-1>", 3)
		/*
		tkCancelRepeat
		%W activate @%x,%y
		*/
		rivet_cancel_repeat();
		rivet_va_cmd(w, "activate", index_at(w, xevent), 0);
		break;

	binding("<Shift-1>", 4)
		/*
		tkListboxBeginExtend %W [%W index @%x,%y]
		*/
		rivet_lb_begin_extend(w, index_at(w, xevent));
		break;

	binding("<Control-1>", 5)
		/*
		tkListboxBeginToggle %W [%W index @%x,%y]
		*/
		rivet_lb_begin_toggle(w, index_at(w, xevent));
		break;

	binding("<B1-Leave>", 6)
		/*
		set tkPriv(x) %x
		set tkPriv(y) %y
		tkListboxAutoScan %W
		*/
		tkPriv->x = xevent->xmotion.x;
		tkPriv->y = xevent->xmotion.y;
		rivet_lb_autoscan(w);
		break;

	binding("<B1-Enter>", 7)
		/*
		tkCancelRepeat
		*/
		rivet_cancel_repeat();
		break;

	binding("<Up>", 8)
		/*
		tkListboxUpDown %W -1
		*/
		rivet_lb_updown(w, -1);
		break;

	binding("<Shift-Up>", 9)
		/*
		tkListboxExtendUpDown %W -1
		*/
		rivet_lb_extend_updown(w, -1);
		break;

	binding("<Down>", 10)
		/*
		tkListboxUpDown %W 1
		*/
		rivet_lb_updown(w, 1);
		break;

	binding("<Shift-Down>", 11)
		/*
		tkListboxExtendUpDown %W 1
		*/
		rivet_lb_extend_updown(w, 1);
		break;

	binding("<Left>", 12)
		/*
		%W xview scroll -1 units
		*/
		rivet_va_cmd(w, XVIEW, SCROLL, NEG1, UNITS, 0);
		break;

	binding("<Control-Left>", 13)
		/*
		%W xview scroll -1 pages
		*/
		rivet_va_cmd(w, XVIEW, SCROLL, NEG1, PAGES, 0);
		break;

	binding("<Right>", 14)
		/*
		%W xview scroll 1 units
		*/
		rivet_va_cmd(w, XVIEW, SCROLL, POS1, UNITS, 0);
		break;

	binding("<Control-Right>", 15)
		/*
		%W xview scroll 1 pages
		*/
		rivet_va_cmd(w, XVIEW, SCROLL, POS1, PAGES, 0);
		break;

	binding("<Prior>", 16)
		/*
		%W yview scroll -1 pages
		%W activate @0,0
		*/
		rivet_va_cmd(w, YVIEW, SCROLL, NEG1, PAGES, 0);
		rivet_va_cmd(w, "activate", "@0,0", 0);
		break;

	binding("<Next>", 17)
		/*
		%W yview scroll 1 pages
		%W activate @0,0
		*/
		rivet_va_cmd(w, YVIEW, SCROLL, POS1, PAGES, 0);
		rivet_va_cmd(w, "activate", "@0,0", 0);
		break;

	binding("<Control-Prior>", 18)
		/*
		%W xview scroll -1 pages
		*/
		rivet_va_cmd(w, YVIEW, SCROLL, NEG1, PAGES, 0);
		break;

	binding("<Control-Next>", 19)
		/*
		%W xview scroll 1 pages
		*/
		rivet_va_cmd(w, YVIEW, SCROLL, POS1, PAGES, 0);
		break;

	binding("<Home>", 20)
		/*
		%W xview moveto 0
		*/
		rivet_va_cmd(w, XVIEW, "moveto", "0", 0);
		break;

	binding("<End>", 21)
		/*
		%W xview moveto 1
		*/
		rivet_va_cmd(w, XVIEW, "moveto", POS1, 0);
		break;

	binding("<Control-Home>", 22)
		/*
		%W activate 0
		%W see 0
		%W selection clear 0 end
		%W selection set 0
		*/
		rivet_va_cmd(w, "activate", "0", 0);
		rivet_va_cmd(w, "see", "0", 0);
		rivet_va_cmd(w, SELECTION, CLEAR, "0", "end", 0);
		rivet_va_cmd(w, SELECTION, "set", "0", 0);
		break;

	binding("<Shift-Control-Home>", 23)
		/*
		tkListboxDataExtend %W 0
		*/
		rivet_lb_data_extend(w, "0");
		break;

	binding("<Control-End>", 24)
		/*
		%W activate end
		%W see end
		%W selection clear 0 end
		%W selection set end
		*/
		rivet_va_cmd(w, "activate", "end", 0);
		rivet_va_cmd(w, "see", "end", 0);
		rivet_va_cmd(w, SELECTION, CLEAR, "0", "end", 0);
		rivet_va_cmd(w, SELECTION, "set", "end", 0);
		break;

	binding("<Shift-Control-End>", 25)
		/*
		tkListboxDataExtend %W end
		*/
		rivet_lb_data_extend(w, "end");
		break;

	binding("<F16>", 26)
		/*
		if {[selection own -displayof %W] == "%W"} {
		    clipboard clear -displayof %W
		    clipboard append -displayof %W [selection get -displayof %W]
		}
		*/
		if (streq(rivet_sel_displayof(w), rivet_path(w))) {
			rivet_va_func(w, Tk_ClipboardCmd,
				CLEAR, "-displayof", rivet_path(w), 0);
			rivet_va_func(w, Tk_ClipboardCmd,
				"append", "-displayof", rivet_path(w),
				rivet_sel_displayof(w), 0);
		}
		break;

	binding("<space>", 27)
		/*
		tkListboxBeginSelect %W [%W index active]
		*/
		/* FALL THROUGH */

	binding("<Select>", 28)
		/*
		tkListboxBeginSelect %W [%W index active]
		*/
		rivet_lb_begin_select(w, index_active(w));
		break;

	binding("<Control-Shift-space>", 29)
		/*
		tkListboxBeginExtend %W [%W index active]
		*/
		/* FALL THROUGH */

	binding("<Shift-Select>", 30)
		/*
		tkListboxBeginExtend %W [%W index active]
		*/
		rivet_lb_begin_extend(w, index_active(w));
		break;

	binding("<Escape>", 31)
		/*
		tkListboxCancel %W
		*/
		rivet_lb_cancel(w);
		break;

	binding("<Control-slash>", 32)
		/*
		tkListboxSelectAll %W
		*/
		rivet_lb_select_all(w);
		break;

	binding("<Control-backslash>", 33)
		/*
		if {[%W cget -selectmode] != "browse"} {
		    %W selection clear 0 end
		}
		*/
		if(!streq(rivet_get(w, "-selectmode"), "browse"))
			rivet_va_cmd(w, SELECTION, CLEAR, "0", "end", 0);
		break;

		/* Additional Tk bindings that aren't part of the Motif
		 * look and feel:
		 */

	binding("<2>", 34)
		/*
		%W scan mark %x %y
		*/
		rivet_va_cmd(w, "scan", "mark",
				rivet_itoa(((XMotionEvent*)xevent)->x),
				rivet_itoa(((XMotionEvent*)xevent)->y), 0);
		break;
	binding("<B2-Motion>", 35)
		/*
		%W scan dragto %x %y
		*/
		rivet_va_cmd(w, "scan", "mark",
				rivet_itoa(((XMotionEvent*)xevent)->x),
				rivet_itoa(((XMotionEvent*)xevent)->y), 0);
		break;

	case 36:
		return TCL_ERROR; /* should cause bind phase to complete */

	default:
		return 1;
	}

	return TCL_CONTINUE;
}

void
rivet_listbox_bind(widget)
	Rivetobj widget;
{
	int n;

	if(rivet_is_class_initialized(widget->rivet_class))
		return;

	rivet_class_initializing(widget->rivet_class);
	for(n=1;;n++) {
		if(listbox_double_duty_event_handler(widget, 0, n) == TCL_ERROR)
			break;
	}

	rivet_class_init_done(widget->rivet_class);
	return ;
}

