#include "rivet.h"

Rivetvalue rivetval_cget;
Tk_ConfigSpec **rivet_config_specs;
int rivet_config_specs_len;
int rivet_config_specs_size;
int rivet_error_occured;
char rivet_error_msg[80]; /* error message when status == TCL_ERROR */

static char nullstr[] ="";

/* 
 * A note on copying results:
 * The rules for copying the result are the same as with Tcl.
 * You will still have to copy the results or it might be scribbled
 * upon.  This scheme just gives a little ease of use for common
 * limited cases.
 *
 * We might also be able to remove the strncpy by refering to the
 * right buffer in Tcl_SetResult. (Save idea for tuning.)
 */
static char tmp_result[4][200];
static int which_result;
#define DO_RESULT {					\
	char *rv = tmp_result[which_result];		\
	strncpy(rv, widget->interp->result, 200);	\
	which_result++;					\
	if(which_result>3)				\
		which_result=0; 			\
	return rv;					\
	}

#define bad_widget() (!widget /* || interp->widget */ )

static void
do_error(widget)
	Rivetobj widget;
{
/* Error handling support is probably inadequate.  Also most callers into
 * these functions don't check the return values.
 */
	rivet_error_occured = 1;
	if(widget->interp->result)
		strncpy(rivet_error_msg, widget->interp->result,
			sizeof(rivet_error_msg));
}
#define reset_error() (rivet_error_occured = 0)

/*
 * rivet_set --
 * Synopsis:  a simple interface to set one value
 */
int
rivet_set(widget, resource, value)
	Rivetobj widget;
	char *resource;
	char *value;
{
	char *argv[4];
	int argc = 4;
	int rv;

	if(bad_widget())
		return 0;

	reset_error();
	Tcl_SetResult(widget->interp, 0, 0);
	argv[0] = rivet_path(widget); /* only used for errors */
	argv[1] = "conf"; /* abbrev for configure */
	argv[2] = resource;
	argv[3] = value;

	rv = (widget->rivet_class->widgetcmd)(widget, widget->interp,
			argc, argv);
	if(rv == TCL_ERROR)
		do_error(widget);
	return rv;
}

char *
rivet_get(widget, resource)
	Rivetobj widget;
	char *resource;
{
	char *argv[3];
	int argc = 3;
	int rv;

	if(bad_widget())
		return nullstr;

	reset_error();
	Tcl_SetResult(widget->interp, 0, 0);
	argv[0] = rivet_path(widget); /* only used for errors */
	argv[1] = "cget";
	argv[2] = resource;

	rv = (widget->rivet_class->widgetcmd)(widget, widget->interp,
			argc, argv);
	if(rv == TCL_ERROR)
		do_error(widget);
	else
	if(widget->interp->result)
		DO_RESULT
	return nullstr;
}

int
rivet_get_int(widget, resource)
	Rivetobj widget;
	char *resource;
{
	char *argv[3];
	int argc = 3;
	int rv;

	if(bad_widget())
		return 0;

	reset_error();
	Tcl_SetResult(widget->interp, 0, 0);
	argv[0] = rivet_path(widget); /* only used for errors */
	argv[1] = "cget";
	argv[2] = resource;

	rv = (widget->rivet_class->widgetcmd)(widget, widget->interp,
			argc, argv);
	if(rv == TCL_ERROR)
		do_error(widget);
	else
	if(widget->interp->result)
		return atoi(widget->interp->result);
	return 0;
}

int
rivet_state_p(widget, expectedvalue)
	Rivetobj widget;
	char *expectedvalue;
{
	int rv;
	char *argv[3];
	int argc = 3;

	if(bad_widget())
		return 0;

	if(!expectedvalue)
		return 0;

	reset_error();
	Tcl_SetResult(widget->interp, 0, 0);
	argv[0] = rivet_path(widget); /* only used for errors */
	argv[1] = "cget";
	argv[2] = "-state";

	rv = (widget->rivet_class->widgetcmd)(widget, widget->interp,
			argc, argv);

	if(rv == TCL_ERROR)
		do_error(widget);
	else
	if(widget->interp->result) {
		if(streq(widget->interp->result, expectedvalue))
			return 1;
	}
	return 0;
}

void
rivet_state_set(widget, newstate)
	Rivetobj widget;
	char *newstate;
{
	int rv;
	char *argv[4];
	int argc = 4;

	if(bad_widget())
		return;

	reset_error();
	Tcl_SetResult(widget->interp, 0, 0);
	argv[0] = rivet_path(widget); /* only used for errors */
	argv[1] = "config";
	argv[2] = "-state";
	argv[3] = newstate;

	rv = (widget->rivet_class->widgetcmd)(widget, widget->interp,
			argc, argv);
	if(rv == TCL_ERROR)
		do_error(widget);
}

int
rivet_update(widget, idletasks)
	Rivetobj widget;
	int idletasks;
{
	int argc;
	char *argv[2];
	int rv;

	if(bad_widget())
		return 0;

	reset_error();
	argv[0] = "update"; /* shouldn't matter */
	if (idletasks) {
		argc=2;
		argv[1] = "idletasks";
	} else {
		argc = 1;
	}

	rv = Tk_UpdateCmd(widget->tkwin, widget->interp, argc, argv);
	if(rv == TCL_ERROR)
		do_error(widget);
}

int
rivet_cmd(widget, argc, argv)
	Rivetobj widget;
	int argc;
	char *argv[];
{
	/* 
	 * Bare bones widget cmd routine, caller must handle errors
	 * building argv, and copying return value.
	 */
	int rv;

	Tcl_SetResult(widget->interp, 0, 0);
	argv[0] = rivet_path(widget);
	rv = (widget->rivet_class->widgetcmd)(widget, widget->interp,
			argc, argv);
	return rv;
}

char *
rivet_va_cmd(widget, va_alist)
	Rivetobj widget;
	va_dcl
{
	va_list arglist;
	char *argv[RIVET_MAX_ARGS];
	char *string;
	int argc;
	int rv;

	if(bad_widget())
		return nullstr;

	reset_error();
	Tcl_SetResult(widget->interp, 0, 0);
	argv[0] = rivet_path(widget);
	argc=1;

	va_start(arglist);
	while(1) {
		string = va_arg(arglist, char *);
		if(!string)
			break;
		if(argc>RIVET_MAX_ARGS-1)
			return nullstr; /* Be safe! Maybe it won't crash. */
		argv[argc] = string;
		argc++;
	}
	va_end(arglist);
	
	rv = (widget->rivet_class->widgetcmd)(widget, widget->interp,
			argc, argv);
	if(rv == TCL_ERROR)
		do_error(widget);
	else
	if(widget->interp->result)
		DO_RESULT
	return nullstr;
}

int
rivet_va_cmd_int(widget, va_alist)
	Rivetobj widget;
	va_dcl
{
	va_list arglist;
	char *argv[RIVET_MAX_ARGS];
	char *string;
	int argc;
	int rv;

	if(bad_widget())
		return 0;

	reset_error();
	Tcl_SetResult(widget->interp, 0, 0);
	argv[0] = rivet_path(widget);
	argc=1;

	va_start(arglist);
	while(1) {
		string = va_arg(arglist, char *);
		if(!string)
			break;
		if(argc>RIVET_MAX_ARGS-1)
			return 0; /* Be safe! Maybe it won't crash. */
		argv[argc] = string;
		argc++;
	}
	va_end(arglist);

	rv = (widget->rivet_class->widgetcmd)(widget, widget->interp,
			argc, argv);
	if(rv == TCL_ERROR)
		do_error(widget);
	else
	if(widget->interp->result)
		return atoi(widget->interp->result);
	return 0;
}

char *
rivet_va_func(widget, func, va_alist)
	Rivetobj widget;
	int	(*func)();
	va_dcl
{
	va_list arglist;
	char *argv[RIVET_MAX_ARGS];
	char *string;
	int argc;
	int rv;

	if(bad_widget())
		return nullstr;

	reset_error();
	Tcl_SetResult(widget->interp, 0, 0);
	argv[0] = "rivetcommand"; /* hopefully this will be invisible */
	argc=1;

	va_start(arglist);
	while(1) {
		string = va_arg(arglist, char *);
		if(!string)
			break;
		if(argc>RIVET_MAX_ARGS-1)
			return nullstr; /* Be safe! Maybe it won't crash. */
		argv[argc] = string;
		argc++;
	}
	va_end(arglist);
	
	rv = (func)(widget->tkwin, widget->interp, argc, argv);
	if(rv == TCL_ERROR)
		do_error(widget);
	else
	if(widget->interp->result)
		DO_RESULT
	return nullstr;
}


char *
rivet_itoa(i)
	int i;
{
	/* Typically you need two numbers at a time (for coordinates), so this
	 * will give you up to two strings.  Otherwise, you'll
	 * have to do some copying.
	 */
	static char a[20];
	static char b[20];
	static int which;

	which ^= 1;
	sprintf((which?a:b), "%d", i);
	return (which?a:b);
}

void
rivet_set_focus(widget)
	Rivetobj widget;
{
	if(bad_widget())
		return;

	rivet_va_func(widget, Tk_FocusCmd, rivet_path(widget), 0);
}

Rivetobj
rivet_get_focus(widget)
	Rivetobj widget; /* reference widget */
{
	TkWindow *focuswinptr;

	if(bad_widget())
		return 0;

	focuswinptr = TkGetFocus(widget->tkwin);
	if(focuswinptr)
		return focuswinptr->rivetobj;
	return 0;
}

Rivetobj
rivet_path_to_rivetobj(path)
	char *path;
{
	Tk_Window result;

	result = Tk_NameToWindow(rivet_main_window->interp, path,
			(Tk_Window) rivet_main_window->tkwin);
	if(result)
		return ((TkWindow*)result)->rivetobj;
	return 0;
}

void
rivet_rootcoords(widget, xevent, x, y)
	Rivetobj widget;
	XEvent *xevent;
	int *x;
	int *y;
{
	int width, height;
	int localx, localy;
	int numberx, numbery;

	if(bad_widget())
		return;

	numberx = xevent->xkey.x_root;
	numbery = xevent->xkey.y_root;
	Tk_GetVRootGeometry((Tk_Window) widget->tkwin,
			&localx, &localy, &width, &height);
	numberx -= localx;
	numbery -= localy;

	*x = numberx;
	*y = numbery;
}

Rivetobj
rivet_to_toplevel(rivetobj)
	Rivetobj        rivetobj;
{
	TkWindow       *winPtr;
	for (winPtr = rivetobj->tkwin; !(winPtr->flags & TK_TOP_LEVEL);
	     winPtr = winPtr->parentPtr) {
		/* Empty loop body. */
	}
	return winPtr->rivetobj;
}

int
rivet_dummy_event_proc(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent	*xevent;
	ClientData assoc;
{
	return TCL_CONTINUE;
}

char *
rivet_sel_displayof(rivetobj)
	Rivetobj rivetobj;
{
	/* frequently used fragment */
	return rivet_va_func(rivetobj,
		Tk_SelectionCmd, "get", "-displayof", rivet_path(rivetobj), 0);
}

static int
rivet_copy(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	/*
	if {[selection own -displayof %W] == "%W"} {
		clipboard clear -displayof %W
		clipboard append -displayof %W [selection get -displayof %W]
	}
	*/

	char           *path;
	path = rivet_path(rivetobj);
	if (streq(rivet_va_func(rivetobj, Tk_SelectionCmd,
				"own", "-displayof", path, 0), path)) {
		rivet_va_func(rivetobj, Tk_ClipboardCmd,
			      "clear", "-displayof", path, 0);
		rivet_va_func(rivetobj, Tk_ClipboardCmd,
			      "append", "-displayof", path,
			      rivet_sel_displayof(rivetobj),
			      0);
	}
	return TCL_CONTINUE;
}

static int
rivet_cut(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	/*
	if {[selection own -displayof %W] == "%W"} {
		clipboard clear -displayof %W
		clipboard append -displayof %W [selection get -displayof %W]
		%W delete sel.first sel.last
	}
	*/

	rivet_copy(rivetobj, xevent, assoc);
	rivet_va_cmd(rivetobj, "delete", "sel.first", "sel.last", 0);
	return TCL_CONTINUE;
}

static int
rivet_paste(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	/*
	%W insert insert [selection get -displayof %W \
		-selection CLIPBOARD]
	*/

	rivet_va_cmd(rivetobj, "insert", "insert",
		rivet_va_func(rivetobj, Tk_SelectionCmd,
			"get", "-displayof", rivet_path(rivetobj),
			"-selection", "CLIPBOARD", 0),
		0);
	return TCL_CONTINUE;
}

int
rivet_bind_clipboard(rivetobj, copy, cut, paste)
	Rivetobj rivetobj;
	char *copy;
	char *cut;
	char *paste;
{
	if(
	(rivet_create_binding(0,
		rivetobj->rivet_class->name,
		copy,
		rivet_copy,
		0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0,
		rivetobj->rivet_class->name,
		cut,
		rivet_cut,
		0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0,
		rivetobj->rivet_class->name,
		paste,
		rivet_paste,
		0, 0, 0) == TCL_ERROR))
			return TCL_ERROR;
	return TCL_OK;
}


int
rivet_tkerror(error)
	char *error;
{
	printf("tkerror message:\n%s\n", error);
	return TCL_CONTINUE; /* or TCL_BREAK */
}

/*
 * -- rivet_create_command
 *
 * By way of a macro, this takes the place of Tcl_CreateCommand
 * and allows us to get at the parameters (usually these are
 * not available through return parameters.)
 *
 */
Tcl_CmdProc *rivetcmd_proc;
ClientData rivetcmd_clientdata;
Tcl_CmdDeleteProc *rivetcmd_deleteProc;
static int in_rivet_create;

Tcl_Command
rivet_create_command(i, cmdName, proc, cd, deleteProc)
	Tcl_Interp *i;
	char *cmdName;
	Tcl_CmdProc *proc;
	ClientData cd;
	Tcl_CmdDeleteProc *deleteProc;
{
	/*
	 * Store these away so we can pick them up so we can scoop
	 * them up later.  Assume that globals are ok for this sitation.
	 */
	rivetcmd_proc = proc;
	rivetcmd_clientdata = cd;
	rivetcmd_deleteProc = deleteProc;
	if(in_rivet_create) {
		/* This is a hack to make sure that language_object is
		 * set to NULL.  Especially important when an error
		 * occured inside the create command function.
		 * This may get in the way if the macro for defining
		 * rivet classes is changed for other purposes.
		 */
		Rivetobj tmp;
		tmp = (Rivetobj)cd;
		tmp->language_object = 0;
	}

	return (Tcl_Command) 0;
}

int
rivet_delete_command(i, cmdName)
	Tcl_Interp *i;
	char *cmdName;
{
	/* This function replaces Tcl_DestroyCommand.  Don't bother
	 * calling the destroy proc, rivet destroys things through
	 * Tk_DestroyWindow.  This might be a problem if interesting
	 * code is used in the widget's destroy proc, but as it stands 
	 * now, these do no more than call Tk_DestroyWindow, which
	 * in our case is unnecessary.
	 * Images, however, have a different approach, and there
	 * is a seperate API to destroy them: Tk_DeleteImage.
	 */ 

	return 0;
}

Rivetobj
rivet_create(rclass, parent, argc, argv_in)
	Rivetclass rclass;
	Rivetobj parent;
	int argc;
	char **argv_in;
{
	Rivetobj widget;
	char **argv;
	char *argv_local[2];
	int rv;
	char newname[200];
	static int nameindex = 1;

	reset_error();
	if(argc<2) {
		argv = argv_local;
		argc = 2;
		argv[1] = 0;
	} else {
		argv = argv_in;
	}

	if(argv[1] && argv[1][0] == '.') {
		/* Caller has choosen a name for this object,
		 * suggested name must start with dot, and must not
		 * include the full path.
		 */
		strncpy(newname, rivet_path(parent), sizeof(newname)-1);
		strncat(newname, argv[1], sizeof(newname)-1);
	} else {
		/* Use default names. */
		if(rivet_path(parent)[1] == '\0')
			sprintf(newname, ".%x", nameindex);
		else
			sprintf(newname, "%s.%x",
				rivet_path(parent), nameindex);
		nameindex++;
	}

	argv[0] = rclass->name;
	argv[1] = newname; /* tk will copy it to malloc'd space. */

	Tcl_SetResult(parent->interp, 0, 0);
	in_rivet_create = 1;
	rv = (rclass->tk_cmd)(parent->tkwin, parent->interp, argc, argv);
	in_rivet_create = 0;
	if(rv != TCL_OK) {
		/* do by hand */
		do_error(parent);
		return 0;
	}
	widget = (Rivetobj) rivetcmd_clientdata;
	/*parent->interp->result = 0;*/
	widget->rivet_class = rclass;
	widget->tkwin->rivetobj = (ClientData)widget;
	if(rclass->bind)
		(rclass->bind)(widget);
	return widget;
}

void
rivet_destroy_window(widget)
	Rivetobj widget;
{
	if(bad_widget())
		return;

	Tcl_SetResult(widget->interp, 0, 0);
	Tk_DestroyWindow((Tk_Window)widget->tkwin);
}

void
rivet_destroy_notify(tkwindow)
	Tk_Window tkwindow;
{
	/* Called by Tk_DestroyWindow to inform of window destruction.
	 * When widget is NULL, it means the create hasn't completed
	 * yet.  In this case, we fake it up a little early and
	 * call the notify function.
	 */
	TkWindow *tkwin;
	tkwin = (TkWindow*) tkwindow;
	if(tkwin->rivetobj == NULL) {
		tkwin->rivetobj = (Rivetobj) rivetcmd_clientdata;
	}
	if(rivet_notify.rivetobj_destroy_notify)
		(rivet_notify.rivetobj_destroy_notify)(tkwin->rivetobj);
}

/*----------------------------------------------------------------------
 * Callback reference counting is done for the associated data, not
 * for the callback structures.
 *----------------------------------------------------------------------*/
void
rivet_decref_callback(callback)
	Callback callback;
{
	if(callback->type != 0 && rivet_notify.decref_assoc_data)
		(rivet_notify.decref_assoc_data)(callback);
}

void
rivet_incref_callback(callback)
	Callback callback;
{
	if(callback->type != 0 && rivet_notify.incref_assoc_data)
		(rivet_notify.incref_assoc_data)(callback);
}

void
rivet_ref_callback_list(callbacks, ncallbacks, inc_or_dec)
	Callback callbacks;
	int ncallbacks;
	int inc_or_dec;
{
	Callback cb;
	int i;

	for(i=0; i<ncallbacks; i++) {
		cb = callbacks+i;
		if(inc_or_dec == 0)
			rivet_decref_callback(cb);
		else
			rivet_incref_callback(cb);

	}
}

char *
rivet_new_callback(function, calldata, calldatatype)
	int (*function)();
	ClientData calldata;
	int calldatatype;
{
	/* Utility function for creating a callback inline, assumes
	 * someone else will cleanup callback memory.
	 */
	Callback cb;

	cb = (Callback) malloc(sizeof(CallbackStruct));
	cb->guard[0] = rivet_callback_guard[0];
	cb->guard[1] = rivet_callback_guard[1];
	cb->guard[2] = rivet_callback_guard[2];
	cb->guard[3] = rivet_callback_guard[3];
	cb->callback = function;
	cb->assoc = calldata;
	cb->type = calldatatype;
	return (char *)cb; /* return a char* to simplify casting in arglists */
}

static int
callback_parse_proc(clientData, interp, tkwin, value, widgRec, offset)
	ClientData clientData;
	Tcl_Interp *interp;
	Tk_Window tkwin;
	char *value;
	char *widgRec;
	int offset;
{
	/* Do the "set" operation on a callback inside a widget structure.
	 * All callbacks are assumed to be allocated by the caller, but 
	 * deallocation is done internally.
	 */
	Callback *callbackp;
	Callback old;
	Callback new;

	callbackp = (Callback*) (widgRec+offset);
	old = *callbackp;
	new = (Callback) value;

	if(old == new)
		return TCL_OK;
		
	if(new) {
		if(rivet_is_callback(new)) {
			*callbackp = new;
			rivet_incref_callback(new);
		} else {
			if(rivet_is_null_callback(new)) {
				*callbackp = 0;
			} else {
				interp->result = "argument is not a callback";
				return TCL_ERROR;
			}
		}
	}

	if(old) {
		/* decrement the callback that was previously there */
		rivet_decref_callback(old);
		free((char*)old);
	}

	return TCL_OK;
}


static char *
callback_print_proc(clientData, tkwin, widgRec, offset, freeProcPtr)
	ClientData clientData;
	Tk_Window tkwin;
	char *widgRec;
	int offset;
	Tcl_FreeProc **freeProcPtr;
{
	Callback *callbackp;
	callbackp = (Callback*) (widgRec+offset);

	if(*callbackp)
		return "rivet_callback";
	else
		return "";
} 

Tk_CustomOption rivet_custom_callback_option = {
	callback_parse_proc,
	callback_print_proc,
	0
};

int
rivet_call_callback(callback, obj, xevent, optarg, type)
	Callback callback;
	ClientData obj;
	XEvent *xevent;
	ClientData optarg;
	int type;
{
	int rv;
	if(!callback || !callback->callback)
		rv = TCL_ERROR;
	else
		rv = (callback->callback)(obj, xevent,
				callback->assoc, optarg, type);
	/* perhaps handle errors here (stacktrace?) */
	return rv;
}

