
#ifndef RIVET
#define RIVET

/*----------------------------------------------------------------------*/

/* forward references */
typedef struct _Rivetstruct *Rivetobj;
typedef struct _Rivetclassstruct *Rivetclass;

/*----------------------------------------------------------------------*/

#ifdef RIVETONLY
/* This define is set by rivet/Makefile but not tk4.x/Makefile */
/* redefinitions */
#define Tcl_CreateCommand	rivet_create_command
#define Tcl_DeleteCommand	rivet_delete_command
#define Tcl_GetCommandName(a,b)	""
#include "tkPort.h"
#endif

/*----------------------------------------------------------------------*/

#include "tk.h"
#include "tkInt.h"
#ifdef HAVE_STDARG_H	/* for python compatibility */
#include <stdarg.h>
#else
#include <varargs.h>
#endif

/*----------------------------------------------------------------------*/

/* utilities */
#define rivet_path(_ro)		(Tk_PathName(_ro->tkwin))
#define rivet_parent(_ro)	((_ro)->tkwin->parentPtr->rivetobj)
#define streq(A, B)		(!strcmp((A), (B)))
#define strtrue( _arg_ )	(!strcmp( (_arg_), "1"))
#define strfalse( _arg_ )	(!strcmp( (_arg_), "0"))
#define strnull(A)		((A)[0] == 0)

#define RIVET_MAX_ARGS		30

/*----------------------------------------------------------------------*/

#ifndef RIVET_CLASS_DECL
/*  Add fields for a "rivet object" with this macro.
 *  The 'language_object field is unused by rivet C, so this may be all that
 *  is needed to support a language.
 */
#define RIVET_CLASS_DECL	Rivetclass rivet_class; \
				ClientData language_object;
#endif

/*----------------------------------------------------------------------*/

#define RIVET_CLASS_SEAL1 0xbeefdead /* differentiates an initializing class */
#define RIVET_CLASS_SEAL2 0xdeadbeef /* differentiates a class */
#define is_rivet_class(obj) 	\
		((((Rivet_class_struct*)obj)->init_seal == RIVET_CLASS_SEAL1) || \
		 (((Rivet_class_struct*)obj)->init_seal == RIVET_CLASS_SEAL2))
#define rivet_class_initializing(obj) ((obj)->init_seal = RIVET_CLASS_SEAL1)
#define rivet_is_class_initialized(obj) ((obj)->init_seal == RIVET_CLASS_SEAL2)
#define rivet_class_init_done(obj) ((obj)->init_seal = RIVET_CLASS_SEAL2)

/*
 * This structure represents a "class", and collects the generic tk widgets
 * operations.  Apps call through these vectors.
 */
typedef struct _Rivetclassstruct {
	int		init_seal;
	char		*name;	/* for create command */
	int		(*tk_cmd) ();
	int             (*widgetcmd) ();
	void		(*deletecmd)();
	void		(*bind)();
	Tk_ConfigSpec	*config_specs;
	int		config_flags;
	ClientData	language_class_object;
} Rivet_class_struct;


/*
 * This structure reflects a convention in the widget structures for all the
 * tk widgets (see Button in tkButton.c, Canvas in tkCanvas.h, etc).  This is
 * probably only a moderately risky thing to do, and it yields uniform access
 * to the window object.
 */
typedef struct _Rivetstruct {
RIVET_CLASS_DECL
	TkWindow	*tkwin;
	Display		*display;
	Tcl_Interp	*interp;
} Rivetstruct;

/*----------------------------------------------------------------------*/

/* Structure as known to bind, returned from get binding calls. */
typedef struct {
	char guard[4]; /* for a legit callback, this discriminates from "" */
	short type; /* by convention, 0 == assoc is defined by C, no gc. */
	int (*callback)();
	ClientData assoc;
} CallbackStruct;
typedef CallbackStruct *Callback;

EXTERN char	*rivet_new_callback	_ANSI_ARGS_((int(*)(), ClientData,int));
EXTERN void	rivet_incref_callback	_ANSI_ARGS_((Callback));
EXTERN void	rivet_decref_callback	_ANSI_ARGS_((Callback));
EXTERN void	rivet_ref_callback_list	_ANSI_ARGS_((Callback, int, int));
EXTERN Tk_CustomOption rivet_custom_callback_option;
EXTERN int	rivet_call_callback	_ANSI_ARGS_((Callback, ClientData obj,
						XEvent*, ClientData optarg,
						int type));
EXTERN char	rivet_callback_guard[4];
#define rivet_is_callback(cb) (memcmp(rivet_callback_guard, (char*)(cb), \
					sizeof(rivet_callback_guard)) == 0)
#define rivet_is_null_callback(cb) (((char*)cb == 0) || (*(char*)cb == 0))


/* "Types" of callbacks generated from the guts of tk.  This is useful
 * for distinguishing the type of the "optarg" in a generalized callback
 * interface. 
 */
#define RIVET_CB_BIND					1
#define RIVET_CB_BUTTON_INVOKE				2
#define RIVET_CB_SCALE_INVOKE				3
#define RIVET_CB_MENUENTRY				4
#define RIVET_CB_MENUPOST				5
#define RIVET_CB_PROTOCOL				6
#define RIVET_CB_SBAR_CHANGE_CLIENT_XVIEW		7
#define RIVET_CB_SBAR_CHANGE_CLIENT_YVIEW		8
#define RIVET_CB_SBAR_CLIENT_CHANGED			9
#define RIVET_CB_TIMER					10
#define RIVET_CB_FILEEVENT				11
#define RIVET_CB_TRACE					12
#define RIVET_CB_SELHANDLE				13
#define RIVET_CB_SELOWN					14

/*----------------------------------------------------------------------*/

/* A common place to store some language implementation dependent function.  
 * Just one of these per process,
 */
typedef struct {
	void	(*rivetobj_destroy_notify) _ANSI_ARGS_((Rivetobj));
	void	(*incref_assoc_data) _ANSI_ARGS_((Callback));
	void	(*decref_assoc_data) _ANSI_ARGS_((Callback));
	void	(*begin_allow_threads) ();
	void	(*end_allow_threads) ();
} Rivetnotifystruct;
EXTERN Rivetnotifystruct rivet_notify;

/*----------------------------------------------------------------------*/
/* C-native types for return values. */
typedef enum {
	Rivet_str, Rivet_int, Rivet_double,
	Rivet_callback, Rivet_rivetobj
} Rivetvaluetype;
typedef struct {
	Rivetvaluetype type;
	char *strval;
	long intval;
	double doubleval;
	Callback callbackval;
	Rivetobj rivetobjval;
} Rivetvalue;
/* Assume that there is never more than one sequential call to
 * cget or configure that is not intercepted by an opportunity to
 * look at the results as stored in these globals.
 * 
 * rivetval_cget holds stuff after both cget and configure, except
 * for configure (in info mode) you have to call ...calc_result first.
 * 
 * If this is confusing, too bad.  We never know the call path or args
 * to cget and configure, this stuff is useful by the calls that wraps
 * those tk calls, so at least the result is available.
 */
EXTERN Rivetvalue rivetval_cget;
EXTERN Tk_ConfigSpec **rivet_config_specs;
EXTERN int rivet_config_specs_len; /* number in use */
EXTERN int rivet_config_specs_size; /* number allocated */
/* after a configure, call this and then get value from rivetval_cget */
EXTERN void	rivet_config_calc_result _ANSI_ARGS_((Tk_ConfigSpec*));

/*----------------------------------------------------------------------*/

EXTERN int	tk_strictMotif;
EXTERN Rivetobj	rivet_main_window;

/*----------------------------------------------------------------------*/

EXTERN Rivetclass MainWindowClass;
EXTERN Rivetclass ButtonClass;
EXTERN Rivetclass LabelClass;
EXTERN Rivetclass CheckButtonClass;
EXTERN Rivetclass RadioButtonClass;
EXTERN Rivetclass FrameClass;
EXTERN Rivetclass ToplevelClass;
EXTERN Rivetclass ScaleClass;
EXTERN Rivetclass CanvasClass;
EXTERN Rivetclass EntryClass;
EXTERN Rivetclass ListboxClass;
EXTERN Rivetclass MenuClass;
EXTERN Rivetclass MenubuttonClass;
EXTERN Rivetclass MessageClass;
EXTERN Rivetclass ScrollbarClass;
EXTERN Rivetclass TextClass;

/*----------------------------------------------------------------------*/

EXTERN Rivetobj rivet_init		_ANSI_ARGS_((int argc, char **argv));
EXTERN Rivetobj rivet_create		_ANSI_ARGS_((Rivetclass rc,
						Rivetobj parent,
						int argc, char **argv));
EXTERN int	rivet_set		_ANSI_ARGS_((Rivetobj ro,
						char *r, char *v));
EXTERN char	*rivet_get		_ANSI_ARGS_((Rivetobj ro, char *r));
EXTERN int	rivet_get_int		_ANSI_ARGS_((Rivetobj ro, char *r));
EXTERN int	rivet_cmd		_ANSI_ARGS_((Rivetobj, int, char *argv[]));
EXTERN char	*rivet_va_cmd();
EXTERN int	rivet_va_cmd_int();
EXTERN char	*rivet_va_func();
EXTERN int	rivet_state_p		_ANSI_ARGS_((Rivetobj ro,
						char *expected_value));
EXTERN int	rivet_set_state		_ANSI_ARGS_((Rivetobj ro,
						char *new_value));

EXTERN void	rivet_set_fileevent	_ANSI_ARGS_((int fd,int mode,Callback));
EXTERN Callback	rivet_get_fileevent	_ANSI_ARGS_((int fd,int mode));

EXTERN int	rivet_create_binding _ANSI_ARGS_((
				ClientData obj, ClientData tag,
				char *eventstr,
				int (*cb)(), ClientData assoc, int append,
				int type));
EXTERN int	rivet_delete_binding _ANSI_ARGS_((
				ClientData obj, ClientData tag,
				char *eventstr));
EXTERN Callback rivet_get_binding _ANSI_ARGS_((
				ClientData obj, ClientData tag,
				char *eventstr, int *ncallbacks));
EXTERN char	*rivet_get_all_bindings _ANSI_ARGS_((
				ClientData obj, ClientData tag));

EXTERN char	*rivet_expand_event	_ANSI_ARGS_((Rivetobj, char*, XEvent*, KeySym, int*));
EXTERN void	rivet_destroy_window	_ANSI_ARGS_((Rivetobj ro));
EXTERN void	rivet_destroy_notify	_ANSI_ARGS_((Tk_Window));
EXTERN char	*rivet_itoa		_ANSI_ARGS_((int i));
EXTERN char	*rivet_string_index	_ANSI_ARGS_((char*, int));
EXTERN char	*rivet_input_str	_ANSI_ARGS_((Rivetobj, XEvent*));
EXTERN void	rivet_set_focus		_ANSI_ARGS_((Rivetobj));
EXTERN Rivetobj	rivet_get_focus		_ANSI_ARGS_((Rivetobj));
EXTERN Rivetobj	rivet_path_to_rivetobj	_ANSI_ARGS_((char*));
EXTERN void	rivet_rootcoords	_ANSI_ARGS_((Rivetobj, XEvent*, int*, int*));
EXTERN Rivetobj	rivet_to_toplevel	_ANSI_ARGS_((Rivetobj));
EXTERN int	rivet_dummy_event_proc	_ANSI_ARGS_((Rivetobj, XEvent*, ClientData));
EXTERN char	*rivet_sel_displayof	_ANSI_ARGS_((Rivetobj));
EXTERN int	rivet_bind_clipboard	_ANSI_ARGS_((Rivetobj, char *, char *, char *));
EXTERN Rivetobj	rivet_focus_next	_ANSI_ARGS_((Rivetobj));
EXTERN Rivetobj	rivet_focus_prev	_ANSI_ARGS_((Rivetobj));

EXTERN Tcl_Command rivet_create_command	_ANSI_ARGS_((Tcl_Interp *interp,
					char *cmdName, Tcl_CmdProc *proc,
					ClientData clientData,
					Tcl_CmdDeleteProc *deleteProc));
EXTERN int	rivet_delete_command	_ANSI_ARGS_((Tcl_Interp *interp,
					char *cmdName));

/*----------------------------------------------------------------------*/

/* listbox related data structures and functions */
EXTERN int *	rivet_listbox_curselection _ANSI_ARGS_((Rivetobj));

/*----------------------------------------------------------------------*/

/* This calls the "-command" associated with a scrollbar.  The callback
 * is typically used to update a client's xview or yview. */
EXTERN void	rivet_scrollbar_notify_client _ANSI_ARGS_((Rivetobj scrollbar,
				int op, double pos));

/* This calls the clients "-yscrollcommand" or "-xscrollcommand", usually
 * called internally by widget to update scrollbar.  The scrollcommand
 * typically does a "set" on the scrollbar, the callback gets an array
 * of 2 doubles (the first and last) as the "optarg". */
EXTERN void	rivet_scrollbar_update _ANSI_ARGS_((Rivetobj client,
						Callback callback,
						double first, double last));
typedef struct {
	/* Scrollbar command callbacks get a pointer to this type of
	 * structure when the scrollbar is updated.
	 */
	int scrollbar_is_vertical;
	int scroll_op;
	double newpos;
} Scrollbar_notify_data;

/* A utility for applications to use with scrollbar "-command" functions. */
EXTERN void	rivet_scrollbar_set_client_view _ANSI_ARGS_((
				Rivetobj client, Scrollbar_notify_data*));

/* A utility for applications to use with client "-?scrollcommand" functions. */
EXTERN void	rivet_scrollbar_set _ANSI_ARGS_((
				Rivetobj scrollbar, double first, double last));

/*----------------------------------------------------------------------*/

/* error handling */
EXTERN int rivet_error_occured;
EXTERN char rivet_error_msg[];
EXTERN int (*rivet_tkerror_proc) _ANSI_ARGS_((char*));
EXTERN int rivet_tkerror _ANSI_ARGS_((char*)); /* the default handler */

/*----------------------------------------------------------------------*/

/* Internal rivet functions and data structures. */
EXTERN void 	rivet_button_bind	_ANSI_ARGS_((Rivetobj ro));
EXTERN void	rivet_scale_bind	_ANSI_ARGS_((Rivetobj ro));
EXTERN void	rivet_entry_bind	_ANSI_ARGS_((Rivetobj ro));
EXTERN void	rivet_listbox_bind	_ANSI_ARGS_((Rivetobj));
EXTERN void	rivet_menu_bind		_ANSI_ARGS_((Rivetobj));
EXTERN void	rivet_menubutton_bind	_ANSI_ARGS_((Rivetobj));
EXTERN void	rivet_text_bind		_ANSI_ARGS_((Rivetobj));
EXTERN void	rivet_screen_changed	_ANSI_ARGS_((Tcl_Interp *interp,
					char *name, int index));
EXTERN void	rivet_cancel_repeat	_ANSI_ARGS_((void));

typedef struct {
	Rivetobj        buttonWindow;
	Rivetobj        window;
	char           *buttons;
	char           *focusContinue;
	char           *focusDir;
	char           *focusOrigin;
	char           *grab;
	char           *selectMode;

	char		*grabStatus;
	char		*oldGrab;

	int		dragging;
	char		initValue[20];
	char		relief[20];

	int		mouseMoved;
	int		x;
	int		y;

	int		deltaX;
	int		deltaY;
	char 		activeBg[20];

	int		pressX;
	int		pressY;

	Tk_TimerToken	afterId;

	char		cursor[20];
	Rivetobj	inMenubutton;
	Rivetobj	focus;
	Rivetobj	postedMb;
	Rivetobj	popup;
	char		prevPos[20];
	char		chr[20];

	/* scrollbar stuff */
	double		initPos;
	double		initMouse;
	int		scrollbar_dragging;
	double		scrollbar_first;
	double		scrollbar_last;

	/* listbox stuff */
	char listboxPrev[20]; 
	int *listboxSelection;

} tkPrivType;
EXTERN tkPrivType *tkPriv;

/* Three data objects set by rivet_create_command. */
EXTERN Tcl_CmdProc *rivetcmd_proc;
EXTERN ClientData rivetcmd_clientdata;
EXTERN Tcl_CmdDeleteProc *rivetcmd_deleteProc;

/*----------------------------------------------------------------------*/

#endif

