
/* 
 * rivet_main.c --
 */

/* derived from ... */
static char sccsid[] = "@(#) tkMain.c 1.112 95/03/10 11:30:37";

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "rivet.h"
#include <tk.h>
#ifdef NO_STDLIB_H
#   include "compat/stdlib.h"
#else
#   include <stdlib.h>
#endif


Rivetnotifystruct rivet_notify;
Rivetobj rivet_main_window;
static Tk_Window mainWindow;	/* The main window for the application.  If
				 * NULL then the application no longer
				 * exists. */

static int synchronize = 0;
static char *name = NULL;
static char *display = NULL;
static char *geometry = NULL;

static Tk_ArgvInfo argTable[] = {
    {"-display", TK_ARGV_STRING, (char *) NULL, (char *) &display,
	"Display to use"},
    {"-geometry", TK_ARGV_STRING, (char *) NULL, (char *) &geometry,
	"Initial geometry for window"},
    {"-name", TK_ARGV_STRING, (char *) NULL, (char *) &name,
	"Name to use for application"},
    {"-sync", TK_ARGV_CONSTANT, (char *) 1, (char *) &synchronize,
	"Use synchronous mode for display server"},
    {(char *) NULL, TK_ARGV_END, (char *) NULL, (char *) NULL,
	(char *) NULL}
};


/*
 *----------------------------------------------------------------------
 *
 * rivet_init --
 *
 * Results:
 *
 *----------------------------------------------------------------------
 */

int (*rivet_tkerror_proc) _ANSI_ARGS_((char*));
char rivet_callback_guard[4];

Rivetobj
rivet_init(argc, argv)
	int argc;	/* Number of arguments. */
	char *argv[];	/* Array of argument strings. */
{
	char *p, *class;
	Tcl_Interp *interp;


	rivet_tkerror_proc = rivet_tkerror; /* set the default */

	/* This is used to distinguish from "", or "no callback" from a
	 * callback created by rivet.  Usually, a null string is used to
	 * indicate deletion of a callback. */
	rivet_callback_guard[0] = 0x20; /* bs */
	rivet_callback_guard[1] = 0; /* nul, to thwart potential tk strcmps. */
	rivet_callback_guard[2] = 0x7f; /* del */
	rivet_callback_guard[3] = 0xfe; /* no ascii equivalent */

	interp = (Tcl_Interp*) Tcl_CreateInterp();

	rivet_notify.rivetobj_destroy_notify = 0;
	rivet_notify.decref_assoc_data = 0;
	rivet_notify.incref_assoc_data = 0;
	rivet_notify.begin_allow_threads = 0;
	rivet_notify.end_allow_threads = 0;
	rivet_config_specs = (Tk_ConfigSpec**)malloc(20 * sizeof(Tk_ConfigSpec**));
	rivet_config_specs_size = 20;

	/*
	 * Parse command-line arguments.  A leading "-file" argument is
	 * ignored (a historical relic from the distant past).  If the next
	 * argument doesn't start with a "-" then strip it off and use it as
	 * the name of a script file to process.  Also check for other
	 * standard arguments, such as "-geometry", anywhere in the argument
	 * list.
	 */

	if (Tk_ParseArgv(interp, (Tk_Window) NULL, &argc, argv, argTable, 0)
	    != TCL_OK)
		return NULL; /* error message in interp->result */

	/*
	 * If a display was specified, put it into the DISPLAY environment
	 * variable so that it will be available for any sub-processes
	 * created by us.
	 */
	if (display != NULL) {
		char buf[1024];
		sprintf(buf, "DISPLAY=%s", display);
		putenv(buf);
	} else {
		display = getenv("DISPLAY");
	}

	/*
	 * Initialize the Tk application.  If a -name option was provided,
	 * use it;  otherwise, if a file name was provided, use the last
	 * element of its path as the name of the application; otherwise use
	 * the last element of the program name.  For the application's
	 * class, capitalize the first letter of the name.
	 */

	if (name == NULL) {
		p = argv[0];
		name = strrchr(p, '/');
		if (name != NULL) {
			name++;
		} else {
			name = p;
		}
	}

	class = (char *) ckalloc((unsigned) (strlen(name) + 1));
	strcpy(class, name);
	class[0] = toupper((unsigned char) class[0]);
	mainWindow = Tk_CreateMainWindow(interp, display, name, class);
	ckfree(class);
	if (mainWindow == NULL)
		return NULL; /* error message in interp->result */

	/*
	 * The main window is actually a frame, which has been
	 * created behind our backs in FrameInit, use same strategy
	 * as rivet_create.
	 */
	rivet_main_window = (Rivetobj) rivetcmd_clientdata;
	rivet_main_window->tkwin->rivetobj = rivet_main_window;
	rivet_main_window->rivet_class = MainWindowClass;

	if (synchronize) {
		XSynchronize(Tk_Display(mainWindow), True);
	}

	/*
	 * Set the geometry of the main window, if requested.  Unlike tcl/tk,
	 * don't keep the requested geometry into a variable.
	 */

	if (geometry != NULL) {
		if (rivet_va_func(rivet_main_window, Tk_WmCmd,
				  "geometry", ".", geometry, 0) != TCL_OK)
			return NULL; /* error in interp->result */
			fprintf(stderr, "%s\n", interp->result);
	}

	rivet_init_lib();
	return rivet_main_window;
}

Tk_Window
Tk_MainWindow(interp)
	Tcl_Interp *interp;
{
	return (Tk_Window)rivet_main_window->tkwin;
}

