#include "rivet.h"


/*
 *--------------------------------------------------------------
 * Following code from scale.tcl.
 *--------------------------------------------------------------
 */

#define BIG 1
#define LITTLE 2
#define INITIAL 11
#define NOREPEAT 12
#define AGAIN 13
#define UP 21
#define DOWN 22

static char *
identify(widget, xevent)
	Rivetobj        widget;
	XEvent         *xevent;
{
	return rivet_va_cmd(widget, "identify",
		rivet_itoa( ((XMotionEvent*)xevent)->x ),
		rivet_itoa( ((XMotionEvent*)xevent)->y ),
		0);
}

/*
 * tkScaleActivate --
 * This procedure is invoked to check a given x-y position in the
 * scale and activate the slider if the x-y position falls within
 * the slider.
 *
 * Arguments:
 * w -		The scale widget.
 * x, y -	Mouse coordinates.
 */
 
static int
rivet_scale_activate(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	if(rivet_state_p(rivetobj, "disabled"))
		return TCL_CONTINUE;
	if(streq(identify(rivetobj, xevent), "slider"))  {
		rivet_state_set(rivetobj, "active");
	} else {
		rivet_state_set(rivetobj, "normal");
	}
	return TCL_CONTINUE;
}

/* 
 * tkScaleButtonDown --
 * This procedure is invoked when a button is pressed in a scale.  It
 * takes different actions depending on where the button was pressed.
 *
 * Arguments:
 * w -		The scale widget.
 * x, y -	Mouse coordinates of button press.
 */
static int
rivet_scale_button_down(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	char *el;
	char *coords;
	long x, y;

	tkPriv->dragging = 0;
	el = identify(rivetobj, xevent);
	if (streq(el, "trough1"))
		rivet_scale_increment(rivetobj, UP, LITTLE, INITIAL);
	else if (streq(el, "trough2"))
	 	rivet_scale_increment(rivetobj, DOWN, LITTLE, INITIAL);
	else if (streq(el, "slider")) {
		tkPriv->dragging = 1;
		strcpy(tkPriv->initValue, rivet_va_cmd(rivetobj, "get", 0));
		coords = rivet_va_cmd(rivetobj, "coords", 0);
		x = strtol(coords, &coords, 10);
		y = strtol(coords, 0, 10);
		tkPriv->deltaX = ((XMotionEvent*)xevent)->x - x;
		tkPriv->deltaY = ((XMotionEvent*)xevent)->y - y;
	}
	return TCL_CONTINUE;
}
 
/*
 * tkScaleDrag --
 * This procedure is called when the mouse is dragged with
 * mouse button 1 down.  If the drag started inside the slider
 * (i.e. the scale is active) then the scale's value is adjusted
 * to reflect the mouse's position.
 *
 * Arguments:
 * w -		The scale widget.
 * x, y -	Mouse coordinates.
 */
static int
rivet_scale_drag(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	int x1, y1;

	if(tkPriv->dragging) {
 
	/*
	 *	$w set [$w get [expr $x - $tkPriv(deltaX)] \
	 *		[expr $y - $tkPriv(deltaY)]]
	 */

		x1 = ((XMotionEvent*)xevent)->x - tkPriv->deltaX;
		y1 = ((XMotionEvent*)xevent)->y - tkPriv->deltaY;

		rivet_va_cmd(rivetobj, "set",
			rivet_va_cmd(rivetobj, "get",
				rivet_itoa(x1), rivet_itoa(y1), 0), 0);
	}
	return TCL_CONTINUE;
}
 
/*
 * tkScaleEndDrag --
 * This procedure is called to end an interactive drag of the
 * slider.  It just marks the drag as over.
#ifdef BETA2
 * This procedure is called to end an interactive drag of the
 * slider.  If the mouse is far from the window then it just
 * restores the scale's value to what it was at the start of
 * the drag.
#endif
 *
 * Arguments:
 * w -		The scale widget.
 * x, y -	The mouse position at the end of the drag.
 */
static int
rivet_scale_end_drag(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	tkPriv->dragging = 0;
	return TCL_CONTINUE;
#ifdef BETA2
	int x, y;

	if(!tkPriv->dragging)
		return TCL_CONTINUE;
	tkPriv->dragging = 0;

	x = ((XMotionEvent*)xevent)->x;
	y = ((XMotionEvent*)xevent)->y;

	if((x < -10) || (y<-10) || 
		(x > Tk_Width(rivetobj->tkwin) + 10) ||
		(y > Tk_Height(rivetobj->tkwin) + 10)) {
			rivet_va_cmd(rivetobj, "set", tkPriv->initValue, 0);
	}
	return TCL_CONTINUE;
#endif
}
 
/*
 * tkScaleIncrement --
 * This procedure is invoked to increment the value of a scale and
 * to set up auto-repeating of the action if that is desired.  The
 * way the value is incremented depends on the "dir" and "big"
 * arguments.
 *
 * Arguments:
 * w -		The scale widget.
 * dir -		"up" means move value towards -from, "down" means
 *		move towards -to.
 * big -		Size of increments: "big" or "little".
 * repeat -	Whether and how to auto-repeat the action:  "noRepeat"
 *		means don't auto-repeat, "initial" means this is the
 *		first action in an auto-repeat sequence, and "again"
 *		means this is the second repetition or later.
 */
typedef struct {
	Rivetobj rivetobj;
	int dir;
	int big;
	int repeat;
} Increment_args;
static Increment_args ia;

static void
rivet_scale_timer(clientdata)
	ClientData clientdata;
{
	Increment_args *td = (Increment_args*)clientdata;
	rivet_scale_increment(td->rivetobj, td->dir, td->big, AGAIN);
}

static int 
rivet_scale_increment_cb(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	Increment_args *incarg = (Increment_args*)assoc;
	rivet_scale_increment(rivetobj,
			incarg->dir, incarg->big, incarg->repeat);
	return TCL_CONTINUE;
}

static int
rivet_scale_increment(rivetobj, dir, big, repeat)
	Rivetobj rivetobj;
	int	dir;
	int	big;
	int	repeat;
{
	int inc;
	int to, from;
	int tmp;
	int repeatn;
	Increment_args *td;

	to = rivet_get_int(rivetobj, "-to");
	from = rivet_get_int(rivetobj, "-from");

	if(big == BIG) {
		inc = rivet_get_int(rivetobj, "-bigincrement");
		if (inc == 0) {
			inc = abs(to-from)/10;
		}
		if (inc < rivet_get_int(rivetobj, "-resolution")) {
			inc = rivet_get_int(rivetobj, "-resolution");
		}
	} else {
		inc = rivet_get_int(rivetobj, "-resolution");
	}

	if((from > to) ^ (dir == UP))
		inc = -inc;

	tmp = rivet_va_cmd_int(rivetobj, "get", 0) + inc;
	(void) rivet_va_cmd(rivetobj, "set", rivet_itoa(tmp), 0);


	if (repeat == AGAIN) {
		repeatn = rivet_get_int(rivetobj, "-repeatinterval");
		td = &ia;
		if(td) {
			td->rivetobj = rivetobj;
			td->dir = dir;
			td->big = big;
			tkPriv->afterId = Tk_CreateTimerHandler(repeatn,
				rivet_scale_timer, td);
		}
	} else
	if (repeat == INITIAL) {
		repeatn = rivet_get_int(rivetobj, "-repeatdelay");
		td = &ia;
		if(td) {
			td->rivetobj = rivetobj;
			td->dir = dir;
			td->big = big;
			tkPriv->afterId = Tk_CreateTimerHandler(repeatn,
				rivet_scale_timer, td);
		}
	}
	return TCL_CONTINUE;
}
 
/*
 * tkScaleControlPress --
 * This procedure handles button presses that are made with the Control
 * key down.  Depending on the mouse position, it adjusts the scale
 * value to one end of the range or the other.
 *
 * Arguments:
 * w -		The scale widget.
 * x, y -	Mouse coordinates where the button was pressed.
 */
static int
rivet_scale_control_press(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	char *el;

	el = identify(rivetobj, xevent);
	if(streq(el, "trough1")) {
		rivet_va_cmd(rivetobj, "set", rivet_get(rivetobj, "-from"), 0);
	} else if(streq(el, "trough2")) {
		rivet_va_cmd(rivetobj, "set", rivet_get(rivetobj, "-to"), 0);
	}
	return TCL_CONTINUE;
}

/*
 * Miscellaneous
 */

static int
rivet_scale_enter(rivetobj, xevent, assoc)
        Rivetobj rivetobj;
        XEvent *xevent;
        ClientData assoc;
{
	if(tk_strictMotif) {
		strncpy(tkPriv->activeBg,
			rivet_get(rivetobj, "-activebackground"),
			sizeof(tkPriv->activeBg));
		rivet_set(rivetobj, "-activebackground",
			rivet_get(rivetobj, "-background"));
	}
	rivet_scale_activate(rivetobj, xevent, assoc);
	return TCL_CONTINUE;
}

static int
rivet_scale_leave(rivetobj, xevent, assoc)
        Rivetobj rivetobj;
        XEvent *xevent;
        ClientData assoc;
{
 	if(tk_strictMotif)
		rivet_set(rivetobj, "-activebackground", tkPriv->activeBg);
	if(rivet_state_p(rivetobj, "active"))
		rivet_state_set(rivetobj, "normal");
	return TCL_CONTINUE;
}

static int
rivet_scale_button_release(rivetobj, xevent, assoc)
        Rivetobj rivetobj;
        XEvent *xevent;
        ClientData assoc;
{
	rivet_cancel_repeat();
	rivet_scale_end_drag(rivetobj, xevent, assoc);
	rivet_scale_activate(rivetobj, xevent, assoc);
	return TCL_CONTINUE;
}

static int
rivet_scale_home_end(rivetobj, xevent, assoc)
        Rivetobj rivetobj;
        XEvent *xevent;
        ClientData assoc;
{
	char *val;

	if(assoc == 0)
		val = rivet_get(rivetobj, "-from");
	else /*if(assoc == 1)*/
		val = rivet_get(rivetobj, "-to");
	rivet_va_cmd(rivetobj, "set", val, 0);
	return TCL_CONTINUE;
}


/* 
 * tkScaleBind --
 * This procedure below invoked the first time the mouse enters a
 * scale widget or a scale widget receives the input focus.  It creates
 * all of the class bindings for scales.
 *
 * Arguments:
 * event -	Indicates which event caused the procedure to be invoked
 *		(Enter or FocusIn).  It is used so that we can carry out
 *		the functions of that event in addition to setting up
 *		bindings.
 */
void
rivet_scale_bind(widget)
	Rivetobj	widget;
{
	static Increment_args upargs = { 0, UP, LITTLE, NOREPEAT };
	static Increment_args downargs = { 0, DOWN, LITTLE, NOREPEAT };
	static Increment_args leftargs = { 0, UP, LITTLE, NOREPEAT };
	static Increment_args rightargs = { 0, DOWN, LITTLE, NOREPEAT };
	static Increment_args cupargs = { 0, UP, BIG, NOREPEAT };
	static Increment_args cdownargs = { 0, DOWN, BIG, NOREPEAT };
	static Increment_args cleftargs = { 0, UP, BIG, NOREPEAT };
	static Increment_args crightargs = { 0, DOWN, BIG, NOREPEAT };

	if(rivet_is_class_initialized(widget->rivet_class))
		return;
	rivet_class_init_done(widget->rivet_class);

	if(
	/* bind Scale <FocusIn> {}  /* no binding should exist */
	(rivet_create_binding(0, widget->rivet_class->name, "<Up>",
		rivet_scale_increment_cb, &upargs, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Down>",
		rivet_scale_increment_cb, &downargs, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Left>",
		rivet_scale_increment_cb, &leftargs, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Right>",
		rivet_scale_increment_cb, &rightargs, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Control-Up>",
		rivet_scale_increment_cb, &cupargs, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Control-Down>",
		rivet_scale_increment_cb, &cdownargs, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Control-Left>",
		rivet_scale_increment_cb, &cleftargs, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Control-Right>",
		rivet_scale_increment_cb, &crightargs, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Enter>",
		rivet_scale_enter, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Motion>",
		rivet_scale_activate, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Leave>",
		rivet_scale_leave, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<1>",
		rivet_scale_button_down, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<B1-Motion>",
		rivet_scale_drag, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<B1-Leave>",
		rivet_dummy_event_proc, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<B1-Enter>",
		rivet_dummy_event_proc, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<ButtonRelease-1>",
		rivet_scale_button_release, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<2>",
		rivet_scale_button_down, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<B2-Motion>",
		rivet_scale_drag, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<B1-Enter>",
		rivet_dummy_event_proc, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<B2-Enter>",
		rivet_dummy_event_proc, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<ButtonRelease-2>",
		rivet_scale_button_release, 0, 0, 0) == TCL_ERROR) ||

	(rivet_create_binding(0, widget->rivet_class->name, "<Control-1>",
		rivet_scale_control_press, 0, 0, 0) == TCL_ERROR) ||

	(rivet_create_binding(0, widget->rivet_class->name, "<Home>",
		rivet_scale_home_end, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<End>",
		rivet_scale_home_end, (ClientData)1, 0, 0) == TCL_ERROR))
			;

	return;
}
