#include "rivet.h"


/*
 *--------------------------------------------------------------
 * Following code from scrollbar.tcl.
 *--------------------------------------------------------------
 */


static double d_zero = 0.0;
static double d_one = 1.0;

#define INITIAL 0
#define AGAIN 1

#define HORIZ		1
#define VERT		2
#define NEGATIVE	4
#define UNIT_HV1	(ClientData) (HORIZ|VERT)
#define UNIT_HVNEG1	(ClientData) (HORIZ|VERT|NEGATIVE)
#define UNIT_V1		(ClientData) (VERT)
#define UNIT_VNEG1	(ClientData) (VERT|NEGATIVE)
#define UNIT_H1		(ClientData) (HORIZ)
#define UNIT_HNEG1	(ClientData) (HORIZ|NEGATIVE)

static void rivet_scroll_start_drag _ANSI_ARGS_((Rivetobj, XEvent*));
static void rivet_scroll_select _ANSI_ARGS_((Rivetobj, char*, int));
static void rivet_scroll_end_drag _ANSI_ARGS_((Rivetobj, XEvent*));
static int  rivet_scroll_by_units _ANSI_ARGS_((Rivetobj, XEvent*, ClientData));
static int  rivet_scroll_by_pages _ANSI_ARGS_((Rivetobj, XEvent*, ClientData));

static double
scrollbar_fraction(rivetobj, xevent)
	Rivetobj rivetobj;
	XEvent *xevent;
{
	/* utility function */
	char *res;
	res = rivet_va_cmd(rivetobj, "fraction",
		rivet_itoa(((XMotionEvent*)xevent)->x),
		rivet_itoa(((XMotionEvent*)xevent)->y),
		0);
	return atof(res);
}

static int
scrollbar_jump_p(rivetobj)
	Rivetobj rivetobj;
{
	/* utility */
	return (strtrue(rivet_get(rivetobj, "-jump")));
}

/* 
 * tkScrollButtonDown --
 * This procedure is invoked when a button is pressed in a scrollbar.
 * It changes the way the scrollbar is displayed and takes actions
 * depending on where the mouse is.
 *
 * Arguments:
 * w -		The scrollbar widget.
 * x, y -	Mouse coordinates.
 */ 
static int
rivet_scroll_button_down(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
#if 0
 proc tkScrollButtonDown {w x y} {
     global tkPriv
     set tkPriv(relief) [$w cget -activerelief]
     $w configure -activerelief sunken
     set element [$w identify $x $y]
     if {$element == "slider"} {
 	tkScrollStartDrag $w $x $y
     } else {
 	tkScrollSelect $w $element initial
     }
 }
#endif
	char *element;

	strncpy(tkPriv->relief, rivet_va_cmd(rivetobj, "cget", "-activerelief", 0),
		sizeof(tkPriv->relief));
	rivet_set(rivetobj, "-activerelief", "sunken");
	element = rivet_va_cmd(rivetobj, "identify",
		rivet_itoa(((XButtonEvent*)xevent)->x),
		rivet_itoa(((XButtonEvent*)xevent)->y),
		0);
	if(streq(element, "slider"))
		rivet_scroll_start_drag(rivetobj, xevent);
	else
		rivet_scroll_select(rivetobj, element, INITIAL);
	return TCL_CONTINUE;
}

/* 
 * tkScrollSelect --
 * This procedure is invoked when button 1 is pressed over the scrollbar.
 * It invokes one of several scrolling actions depending on where in
 * the scrollbar the button was pressed.
 *
 * Arguments:
 * w -		The scrollbar widget.
 * element -	The element of the scrollbar that was selected, such
 *		as "arrow1" or "trough2".  Shouldn't be "slider".
 * repeat -	Whether and how to auto-repeat the action:  "noRepeat"
 *		means don't auto-repeat, "initial" means this is the
 *		first action in an auto-repeat sequence, and "again"
 *		means this is the second repetition or later.
 */ 
static Rivetobj scroll_rivetobj;
static char scroll_element[20];

static void
rivet_scroll_timer(clientdata)
	ClientData clientdata;
{
	rivet_scroll_select(scroll_rivetobj, scroll_element, AGAIN);
}

static void
rivet_scroll_select(rivetobj, element, repeat)
	Rivetobj rivetobj;
	char *element;
	int repeat;
{
#if 0
 proc tkScrollSelect {w element repeat} {
     global tkPriv
     if {$element == "arrow1"} {
 	tkScrollByUnits $w hv -1
     } elseif {$element == "trough1"} {
 	tkScrollByPages $w hv -1
     } elseif {$element == "trough2"} {
 	tkScrollByPages $w hv 1
     } elseif {$element == "arrow2"} {
 	tkScrollByUnits $w hv 1
     } else {
 	return
     }
     if {$repeat == "again"} {
 	set tkPriv(afterId) [after [$w cget -repeatinterval] \
 		tkScrollSelect $w $element again]
     } elseif {$repeat == "initial"} {
 	set tkPriv(afterId) [after [$w cget -repeatdelay] \
 		tkScrollSelect $w $element again]
     }
 }
#endif
	int repeatn;

	if(streq(element, "arrow1"))
		rivet_scroll_by_units(rivetobj, (XEvent*)0, UNIT_HVNEG1);
	else
	if(streq(element, "trough1"))
		rivet_scroll_by_pages(rivetobj, (XEvent*)0, UNIT_HVNEG1);
	else
	if(streq(element, "trough2"))
		rivet_scroll_by_pages(rivetobj, (XEvent*)0, UNIT_HV1);
	else
	if(streq(element, "arrow2"))
		rivet_scroll_by_units(rivetobj, (XEvent*)0, UNIT_HV1);
	else
		return;

	repeatn = rivet_get_int(rivetobj, 
		(repeat == AGAIN)? "-repeatinterval" : "-repeatdelay");

	scroll_rivetobj = rivetobj;
	strncpy(scroll_element, element, sizeof(scroll_element));
	tkPriv->afterId = Tk_CreateTimerHandler(repeatn, rivet_scroll_timer, 0);
}

/* 
 * tkScrollButtonUp --
 * This procedure is invoked when a button is released in a scrollbar.
 * It cancels scans and auto-repeats that were in progress, and restores
 * the way the active element is displayed.
 *
 * Arguments:
 * w -		The scrollbar widget.
 * x, y -	Mouse coordinates.
 */ 
static int
rivet_scroll_button_up(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
#if 0
 proc tkScrollButtonUp {w x y} {
     global tkPriv
     tkCancelRepeat
     $w configure -activerelief $tkPriv(relief)
     tkScrollEndDrag $w $x $y
     $w activate [$w identify $x $y]
 }
#endif
	rivet_cancel_repeat();
	rivet_set(rivetobj, "-activerelief", tkPriv->relief);
	rivet_scroll_end_drag(rivetobj, xevent);
	rivet_va_cmd(rivetobj, "activate",
		rivet_va_cmd(rivetobj, "identify",
			rivet_itoa(((XButtonEvent*)xevent)->x),
			rivet_itoa(((XButtonEvent*)xevent)->y),
			0),
		0);
	return TCL_CONTINUE;
}

/* 
 * tkScrollStartDrag --
 * This procedure is called to initiate a drag of the slider.  It just
 * remembers the starting position of the slider.
 *
 * Arguments:
 * w -		The scrollbar widget.
 * x, y -	The mouse position at the start of the drag operation.
 */ 
static void
rivet_scroll_start_drag(rivetobj, xevent)
	Rivetobj rivetobj;
	XEvent *xevent;
{
#if 0
 proc tkScrollStartDrag {w x y} {
    global tkPriv

    if {[$w cget -command] == ""} {
        return
    }
    set tkPriv(pressX) $x
    set tkPriv(pressY) $y
    set tkPriv(initValues) [$w get]
    set iv0 [lindex $tkPriv(initValues) 0]
    if {[llength $tkPriv(initValues)] == 2} {
        set tkPriv(initPos) $iv0
    } else {
        if {$iv0 == 0} {
            set tkPriv(initPos) 0.0
        } else {
            set tkPriv(initPos) [expr (double([lindex $tkPriv(initValues) 2])) \
                    / [lindex $tkPriv(initValues) 0]]
        }
    }
 }
#endif
	char *positions;
	char *argv[3];
	char *next;

	/* The interface to cget of callbacks isn't great.s
	 * Since it's not a string, we have to be careful.
	 */
	argv[0] = "";
	argv[1] = "cget";
	argv[2] = "-command";
	Tcl_SetResult(rivetobj->interp, 0, 0);
	(rivetobj->rivet_class->widgetcmd)(rivetobj, rivetobj->interp,
		4, argv);
	if(rivetobj->interp->result == 0)
		return;

	tkPriv->pressX = xevent->xbutton.x;
	tkPriv->pressY = xevent->xbutton.y;
	positions = rivet_va_cmd(rivetobj, "get", 0);
	/* assume there are two numbers, not using older style args. */
	tkPriv->scrollbar_first = strtod(positions, &next);
	tkPriv->scrollbar_last = strtod(next, (char**)0);
	tkPriv->initPos = tkPriv->scrollbar_first;
	tkPriv->scrollbar_dragging = 1;
}


/* 
 * tkScrollDrag --
 * This procedure is called for each mouse motion even when the slider
 * is being dragged.  It notifies the associated widget if we're not
 * jump scrolling, and it just updates the scrollbar if we are jump
 * scrolling.
 *
 * Arguments:
 * w -		The scrollbar widget.
 * x, y -	The current mouse position.
 */ 
static int
rivet_scroll_drag(rivetobj, xevent, clientdata)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData clientdata;
{
#if 0
 proc tkScrollDrag {w x y} {
     global tkPriv
 
     if {$tkPriv(initPos) == ""} {
 	return
     }
     set delta [$w delta [expr $x - $tkPriv(pressX)] [expr $y - $tkPriv(pressY)]]
     if [$w cget -jump] {
 	if {[llength $tkPriv(initValues)] == 2} {
 	    $w set [expr [lindex $tkPriv(initValues) 0] + $delta] \
 		    [expr [lindex $tkPriv(initValues) 1] + $delta]
 	} else {
 	    set delta [expr round($delta * [lindex $tkPriv(initValues) 0])]
 	    eval $w set [lreplace $tkPriv(initValues) 2 3 \
 		    [expr [lindex $tkPriv(initValues) 2] + $delta] \
 		    [expr [lindex $tkPriv(initValues) 3] + $delta]]
 	}
     } else {
 	tkScrollToPos $w [expr $tkPriv(initPos) + $delta]
     }
 }
#endif
	double f, delta;

	if(!tkPriv->scrollbar_dragging)
		return TCL_CONTINUE;
	delta = strtod(rivet_va_cmd(rivetobj, "delta",
			rivet_itoa(xevent->xmotion.x - tkPriv->pressX),
			rivet_itoa(xevent->xmotion.y - tkPriv->pressY), 0), 0);
	if(scrollbar_jump_p(rivetobj)) {
		rivet_scrollbar_set(rivetobj, 
				tkPriv->scrollbar_first + delta,
				tkPriv->scrollbar_last + delta);
	} else {
		double tmp;
		tmp = tkPriv->initPos+delta;
		rivet_scroll_to_pos(rivetobj, xevent, &tmp);
	}
	return TCL_CONTINUE;
}

/* 
 * tkScrollEndDrag --
 * This procedure is called to end an interactive drag of the slider.
 * It scrolls the window if we're in jump mode, otherwise it does nothing.
 *
 * Arguments:
 * w -		The scrollbar widget.
 * x, y -	The mouse position at the end of the drag operation.
 */ 
static void
rivet_scroll_end_drag(rivetobj, xevent)
	Rivetobj rivetobj;
	XEvent *xevent;
{
#if 0
 proc tkScrollEndDrag {w x y} {
     global tkPriv
 
     if {$tkPriv(initPos) == ""} {
 	return
     }
     if [$w cget -jump] {
        set delta [$w delta [expr $x - $tkPriv(pressX)] \
                [expr $y - $tkPriv(pressY)]]
        tkScrollToPos $w [expr $tkPriv(initPos) + $delta]
     }
     set tkPriv(initMouse) ""
 }
#endif
	double val;
	char *delta;

	if(!tkPriv->scrollbar_dragging)
		return;
	if(scrollbar_jump_p(rivetobj)) {
		delta = rivet_va_cmd(rivetobj, "delta",
			rivet_itoa(xevent->xbutton.x - tkPriv->pressX),
			rivet_itoa(xevent->xbutton.y - tkPriv->pressY), 0);
		val = tkPriv->initPos + strtod(delta, 0);
		rivet_scroll_to_pos(rivetobj, xevent, &val);
	}
	tkPriv->scrollbar_dragging = 0;
}

/* 
 * tkScrollByUnits --
 * This procedure tells the scrollbar's associated widget to scroll up
 * or down by a given number of units.  It notifies the associated widget
 * in different ways for old and new command syntaxes.
 *
 * Arguments:
 * w -		The scrollbar widget.
 * orient -	Which kinds of scrollbars this applies to:  "h" for
 *		horizontal, "v" for vertical, "hv" for both.
 * amount -	How many units to scroll:  typically 1 or -1.
 */ 
static int
rivet_scroll_by_units(rivetobj, xevent, orientamount)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData orientamount;
{
#if 0
 proc tkScrollByUnits {w orient amount} {
     set cmd [$w cget -command]
     if {($cmd == "") || ([string first \
 	    [string index [$w cget -orient] 0] $orient] < 0)} {
 	return
     }
     set info [$w get]
     if {[llength $info] == 2} {
 	uplevel #0 $cmd scroll $amount units
     } else {
 	uplevel #0 $cmd [expr [lindex $info 2] + $amount]
     }
 }
#endif
	char *res;

	res = rivet_get(rivetobj, "-orient");
	if(res[0] == 'h')
	    if(!((int)orientamount & HORIZ))
		return TCL_CONTINUE;
	if(res[0] == 'v')
	    if(!((int)orientamount & VERT))
		return TCL_CONTINUE;

	rivet_scrollbar_notify_client(rivetobj, TK_SCROLL_UNITS,
		((int)orientamount & NEGATIVE) ? -1.0:1.0);
	return TCL_CONTINUE;
}


/* 
 * tkScrollByPages --
 * This procedure tells the scrollbar's associated widget to scroll up
 * or down by a given number of screenfuls.  It notifies the associated
 * widget in different ways for old and new command syntaxes.
 *
 * Arguments:
 * w -		The scrollbar widget.
 * orient -	Which kinds of scrollbars this applies to:  "h" for
 *		horizontal, "v" for vertical, "hv" for both.
 * amount -	How many screens to scroll:  typically 1 or -1.
 */ 
static int
rivet_scroll_by_pages(rivetobj, xevent, orientamount)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData orientamount;
{
#if 0
 proc tkScrollByPages {w orient amount} {
     set cmd [$w cget -command]
     if {($cmd == "") || ([string first \
 	    [string index [$w cget -orient] 0] $orient] < 0)} {
 	return
     }
     set info [$w get]
     if {[llength $info] == 2} {
 	uplevel #0 $cmd scroll $amount pages
     } else {
 	uplevel #0 $cmd [expr [lindex $info 2] + $amount*([lindex $info 1] - 1)]
     }
 }
#endif
	char *res;

	res = rivet_get(rivetobj, "-orient");
	if(res[0] == 'h')
	    if(!((int)orientamount & HORIZ))
		return TCL_CONTINUE;
	if(res[0] == 'v')
	    if(!((int)orientamount & VERT))
		return TCL_CONTINUE;

	rivet_scrollbar_notify_client(rivetobj, TK_SCROLL_PAGES,
		((int)orientamount & NEGATIVE) ? -1.0:1.0);
	return TCL_CONTINUE;
}

/* 
 * tkScrollToPos --
 * This procedure tells the scrollbar's associated widget to scroll to
 * a particular location, given by a fraction between 0 and 1.  It notifies
 * the associated widget in different ways for old and new command syntaxes.
 *
 * Arguments:
 * w -		The scrollbar widget.
 * pos -		A fraction between 0 and 1 indicating a desired position
 *		in the document.
 */ 
static int
rivet_scroll_to_pos(rivetobj, xevent, pos)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData pos;
{
#if 0
 proc tkScrollToPos {w pos} {
     set cmd [$w cget -command]
     if {($cmd == "")} {
 	return
     }
     set info [$w get]
     if {[llength $info] == 2} {
 	uplevel #0 $cmd moveto $pos
     } else {
 	uplevel #0 $cmd [expr round([lindex $info 0]*$pos)]
     }
 }
#endif
	double dpos = *(double*)pos;
	rivet_scrollbar_notify_client(rivetobj, TK_SCROLL_MOVETO, dpos);
	return TCL_CONTINUE;
}

/* 
 * tkScrollTopBottom
 * Scroll to the top or bottom of the document, depending on the mouse
 * position.
 *
 * Arguments:
 * w -		The scrollbar widget.
 * x, y -	Mouse coordinates within the widget.
 */ 
static int
rivet_scroll_top_bottom(rivetobj, xevent, unused)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData unused;
{
#if 0
 proc tkScrollTopBottom {w x y} {
     set element [$w identify $x $y]
     if [string match *1 $element] {
 	tkScrollToPos $w 0
     } elseif [string match *2 $element] {
 	tkScrollToPos $w 1
     }
 }
#endif

	char *element;
	int len;

	element = rivet_va_cmd(rivetobj, "identify",
			rivet_itoa(((XButtonEvent*)xevent)->x),
			rivet_itoa(((XButtonEvent*)xevent)->y),
			0);
	if(!element)
		return TCL_CONTINUE;
	len = strlen(element);
	if(element[len] == '1')
		rivet_scroll_to_pos(rivetobj, xevent, &d_zero);
	else
	if(element[len] == '2')
		rivet_scroll_to_pos(rivetobj, xevent, &d_one);

	return TCL_CONTINUE;
}



static int
rivet_scroll_motion(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	/* %W activate [%W identify %x %y] */
	rivet_va_cmd(rivetobj, "activate",
		rivet_va_cmd(rivetobj, "identify",
			rivet_itoa(((XMotionEvent*)xevent)->x),
			rivet_itoa(((XMotionEvent*)xevent)->y),
			0),
		0);
	return TCL_CONTINUE;
}



static int
rivet_scroll_enter(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
#if 0
 	if $tk_strictMotif {
 	    set tkPriv(activeBg) [%W cget -activebackground]
 	    %W config -activebackground [%W cget -background]
 	}
 	%W activate [%W identify %x %y]
#endif
	if(tk_strictMotif) {
		strcpy(tkPriv->activeBg, rivet_get(rivetobj, "-activebackground"));
		rivet_set(rivetobj, "-activebackground",
			rivet_get(rivetobj, "-background"));
	}
	return TCL_CONTINUE;
}
 

static int
rivet_scroll_leave(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
#if 0
 	if $tk_strictMotif {
 	    %W config -activebackground $tkPriv(activeBg)
 	}
 	%W activate {}
#endif
	if(tk_strictMotif)
		rivet_set(rivetobj, "-activebackground", tkPriv->activeBg);
	rivet_va_cmd(rivetobj, "activate", "", 0);
	return TCL_CONTINUE;
}

/* 
 * tkScrollbarBind --
 * This procedure below invoked the first time the mouse enters a
 * scrollbar widget or an scrollbar widget receives the input focus.  It
 * creates all of the class bindings for scrollbars.
 *
 * Arguments:
 * w -		The event in which the window occurred.
 * x, y -	Coordinates of the mouse.
 * event -	Indicates which event caused the procedure to be invoked
 *		(Enter or FocusIn).  It is used so that we can carry out
 *		the functions of that event in addition to setting up
 *		bindings.
 */ 
void
rivet_scrollbar_bind(widget)
	Rivetobj	widget;
{

	if(rivet_is_class_initialized(widget->rivet_class))
		return;
	tkPriv->scrollbar_dragging = 0;
	rivet_class_init_done(widget->rivet_class);

	if(
	(rivet_create_binding(0, widget->rivet_class->name, "<Enter>",
		rivet_scroll_enter, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Motion>",
		rivet_scroll_motion, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Leave>",
		rivet_scroll_leave, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<1>",
		rivet_scroll_button_down, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<B1-Motion>",
		rivet_scroll_drag, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<ButtonRelease-1>",
		rivet_scroll_button_up, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<B1-Leave>",
		rivet_dummy_event_proc, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<B1-Enter>",
		rivet_dummy_event_proc, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<2>",
		rivet_scroll_button_down, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<B2-Motion>",
		rivet_scroll_drag, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<ButtonRelease-2>",
		rivet_scroll_button_up, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<B2-Leave>",
		rivet_dummy_event_proc, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<B2-Enter>",
		rivet_dummy_event_proc, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Control-1>",
		rivet_scroll_top_bottom, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Control-2>",
		rivet_scroll_top_bottom, 0, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Up>",
		rivet_scroll_by_units, UNIT_VNEG1, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Down>",
		rivet_scroll_by_units, UNIT_V1, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Control-Up>",
		rivet_scroll_by_pages, UNIT_VNEG1, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Control-Down>",
		rivet_scroll_by_pages, UNIT_V1, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Left>",
		rivet_scroll_by_units, UNIT_HNEG1, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Right>",
		rivet_scroll_by_units, UNIT_H1, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Control-Left>",
		rivet_scroll_by_pages, UNIT_HNEG1, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Control-Right>",
		rivet_scroll_by_pages, UNIT_H1, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Prior>",
		rivet_scroll_by_pages, UNIT_HVNEG1, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Next>",
		rivet_scroll_by_pages, UNIT_HVNEG1, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<Home>",
		rivet_scroll_to_pos, &d_zero, 0, 0) == TCL_ERROR) ||
	(rivet_create_binding(0, widget->rivet_class->name, "<End>",
		rivet_scroll_to_pos, &d_one, 0, 0) == TCL_ERROR))
			return;
	rivet_va_cmd(widget, "set", "0.0", "1.0", 0); /* make it use new style commands */
	
}


