#include "rivet.h"

/*
 *--------------------------------------------------------------
 * Following code from tearoff.tcl.
 *--------------------------------------------------------------
 */

/*
 * tkMenuDup --
 * Given a menu (hierarchy), create a duplicate menu (hierarchy)
 * in a given window.
 *
 * Arguments:
 * src -		Source window.  Must be a menu.  It and its
 *			menu descendants will be duplicated at dst.
 * dst -		Name to use for topmost menu in duplicate
 *			hierarchy.
 */
static Rivetobj
rivet_menu_dup(src, parent)
	Rivetobj src;
	Rivetobj parent;
{
#if 0
 proc tkMenuDup {src dst} {
    set cmd "menu $dst"
    foreach option [$src configure] {
	if {[llength $option] == 2} {
	    continue
	}
	lappend cmd [lindex $option 0] [lindex $option 4]
    }
    eval $cmd
    set last [$src index last]
    if {$last == "none"} {
	return
    }
    for {set i [$src cget -tearoff]} {$i <= $last} {incr i} {
	set cmd "$dst add [$src type $i]"
	foreach option [$src entryconfigure $i]  {
	    lappend cmd [lindex $option 0] [lindex $option 4]
	}
	eval $cmd
	if {[$src type $i] == "cascade"} {
	    tkMenuDup [$src entrycget $i -menu] $dst.m$i
	    $dst entryconfigure $i -menu $dst.m$i
	}
    }
    # Duplicate the binding tags and bindings from the source menu.

    regsub -all . $src {\\&} quotedSrc
    regsub -all . $dst {\\&} quotedDst
    regsub -all $quotedSrc [bindtags $src] $dst x
    bindtags $dst $x
    foreach event [bind $src] {
        regsub -all $quotedSrc [bind $src $event] $dst x
        bind $dst $event $x
    }
 }
#endif


/* gack */

	return 0;
}



static int
tearoffsetfocus(menu, xevent, assoc)
	Rivetobj menu;
	XEvent *xevent;
	ClientData assoc;
{
	rivet_set_focus(menu);
	return TCL_CONTINUE;
}

/*
 * tkTearoffMenu --
 * Given the name of a menu, this procedure creates a torn-off menu
 * that is identical to the given menu (including nested submenus).
 * The new torn-off menu exists as a toplevel window managed by the
 * window manager.  The return value is the name of the new menu.
 *
 * Arguments:
 * w -			The menu to be torn-off (duplicated).
 */
void
rivet_tear_off_menu(w)
	Rivetobj w;
{
#if 0
 proc tkTearOffMenu w {
    # Find a unique name to use for the torn-off menu.  Find the first
    # ancestor of w that is a toplevel but not a menu, and use this as
    # the parent of the new menu.  This guarantees that the torn off
    # menu will be on the same screen as the original menu.  By making
    # it a child of the ancestor, rather than a child of the menu, it
    # can continue to live even if the menu is deleted;  it will go
    # away when the toplevel goes away.

    set parent [winfo parent $w]
    while {([winfo toplevel $parent] != $parent)
	    || ([winfo class $parent] == "Menu")} {
	set parent [winfo parent $parent]
    }
    if {$parent == "."} {
	set parent ""
    }
    for {set i 1} 1 {incr i} {
	set menu $parent.tearoff$i
	if ![winfo exists $menu] {
	    break
	}
    }

    tkMenuDup $w $menu
    wm overrideredirect $menu 0

    # Pick a title for the new menu by looking at the parent of the
    # original: if the parent is a menu, then use the text of the active
    # entry.  If it's a menubutton then use its text.

    set parent [winfo parent $w]
    switch [winfo class $parent] {
	Menubutton {
	    wm title $menu [$parent cget -text]
	}
	Menu {
	    wm title $menu [$parent entrycget active -label]
	}
    }

    $menu configure -tearoff 0
    $menu post [winfo x $w] [winfo y $w]

    # Set tkPriv(focus) on entry:  otherwise the focus will get lost
    # after keyboard invocation of a sub-menu (it will stay on the
    # submenu).

    bind $menu <Enter> {
	set tkPriv(focus) %W
    }
 }
#endif
	Rivetobj parent;
	Rivetobj toplevel;
	Rivetobj menu;
	char *tmp;

	parent = rivet_parent(w);
	while((rivet_to_toplevel(parent) != parent) ||
	      (parent->rivet_class == MenuClass)) {
		parent = rivet_parent(parent);
	}

	menu = rivet_menu_dup(w, parent);
	if(!menu)
		return;
	rivet_va_func(w, Tk_WmCmd,
			"overrideredirect", rivet_path(menu), "0", 0);

	parent = rivet_parent(w);
	if(parent->rivet_class == MenubuttonClass)
		tmp = rivet_get(parent, "-text");
	else
	if(parent->rivet_class == MenuClass)
		tmp = rivet_va_cmd(parent, "entrycget", "active", "-label", 0);
	rivet_va_func(menu, Tk_WmCmd, "title", rivet_path(menu), tmp, 0);

	rivet_set(menu, "-tearoff", "0");
	rivet_va_cmd(menu, "post",
		rivet_itoa(Tk_X((Tk_Window)w->tkwin)),
		rivet_itoa(Tk_Y((Tk_Window)w->tkwin)),
		0);

	rivet_create_binding(menu, 0, "<Enter>", tearoffsetfocus, 0, 0, 0);
}
