#include "rivet.h"

/* gack.  check for "get" of text mixed with rivet_va_cmd.
 * the results may be truncated by strncpy in the rivet temp buffers.
 * May have to have seperate entry point to just do a command without
 * copy.
 */

/*
 *--------------------------------------------------------------
 * Following code from text.tcl.
 *--------------------------------------------------------------
 */
/*
 * Some common strings.  Unfortunately, we cannot make these const
 * because tk sometimes touches them (presumably nondestructively)
 * and const would put them in a read only segment.
 */
static char anchor_string[] = "anchor";
#define ANCHOR anchor_string
static char index_string[] = "index";
#define INDEX index_string
static char insert_string[] = "insert";
#define INSERT insert_string
static char onedotoh_string[] = "1.0";
#define _1_0_ onedotoh_string

static int
text_compare(rivetobj, arg1, op, arg2)
	Rivetobj rivetobj;
	char *arg1;
	char *op;
	char *arg2;
{
	/* do the compare op, but since compare returns either "1" or "0"
	 * use the _int version.
	 */

	return rivet_va_cmd_int(rivetobj, "compare", arg1, op, arg2, 0);
}

static char *
text_index(w, arg1, arg2)
	Rivetobj w;
	char *arg1;
	char *arg2;
{
	char buf[80];

	if(arg2)
		sprintf(buf, arg1, arg2);
	return rivet_va_cmd(w, INDEX, arg2 ? buf : arg1, 0);
}

 
/* 
 * tkTextButton1 --
 * This procedure is invoked to handle button-1 presses in text
 * widgets.  It moves the insertion cursor, sets the selection anchor,
 * and claims the input focus.
 *
 * Arguments:
 * w -		The text window in which the button was pressed.
 * x -		The x-coordinate of the button press.
 * y -		The x-coordinate of the button press.
 */
static void
rivet_text_button1(rivetobj, x, y)
	Rivetobj rivetobj;
	int x, y;
{
#if 0 
 proc tkTextButton1 {w x y} {
     global tkPriv
 
     set tkPriv(selectMode) char
     set tkPriv(mouseMoved) 0
     set tkPriv(pressX) $x
     $w mark set insert @$x,$y
     $w mark set anchor insert
     if {[$w cget -state] == "normal"} {focus $w}
 }
#endif
	char buf[20];
	tkPriv->selectMode = "char";
	tkPriv->mouseMoved = 0;
	tkPriv->pressX = x;
	sprintf(buf, "@%d,%d", x, y);
	rivet_va_cmd(rivetobj, "mark", "set", INSERT, buf, 0);
	rivet_va_cmd(rivetobj, "mark", "set", ANCHOR, INSERT, 0);
	if(rivet_state_p(rivetobj, "normal"))
		rivet_set_focus(rivetobj);
}
 
/*
 * tkTextSelectTo --
 * This procedure is invoked to extend the selection, typically when
 * dragging it with the mouse.  Depending on the selection mode (character,
 * word, line) it selects in different-sized units.  This procedure
 * ignores mouse motions initially until the mouse has moved from
 * one character to another or until there have been multiple clicks.
 *
 * Arguments:
 * w -		The text window in which the button was pressed.
 * x -		Mouse x position.
 * y - 		Mouse y position.
 */
static int
rivet_text_select_to(rivetobj, x, y)
	Rivetobj rivetobj;
	int x, y;
{
#if 0
 proc tkTextSelectTo {w x y} {
     global tkPriv
 
     set cur [$w index @$x,$y]
     if [catch {$w index anchor}] {
 	$w mark set anchor $cur
     }
     set anchor [$w index anchor]
     if {[$w compare $cur != $anchor] || (abs($tkPriv(pressX) - $x) >= 3)} {
 	set tkPriv(mouseMoved) 1
     }
     switch $tkPriv(selectMode) {
 	char {
 	    if [$w compare $cur < anchor] {
 		set first $cur
 		set last anchor
 	    } else {
 		set first anchor
 		set last [$w index "$cur + 1c"]
 	    }
 	}
 	word {
 	    if [$w compare $cur < anchor] {
 		set first [$w index "$cur wordstart"]
 		set last [$w index "anchor - 1c wordend"]
 	    } else {
 		set first [$w index "anchor wordstart"]
 		set last [$w index "$cur wordend"]
 	    }
 	}
 	line {
 	    if [$w compare $cur < anchor] {
 		set first [$w index "$cur linestart"]
 		set last [$w index "anchor - 1c lineend + 1c"]
 	    } else {
 		set first [$w index "anchor linestart"]
 		set last [$w index "$cur lineend + 1c"]
 	    }
 	}
     }
     if {$tkPriv(mouseMoved) || ($tkPriv(selectMode) != "char")} {
 	$w tag remove sel 0.0 $first
 	$w tag add sel $first $last
 	$w tag remove sel $last end
 	update idletasks
     }
 }
#endif
	char buf[80];
	char cur[40];
	char first[40];
	char last[40];
	char *anchor_var;
	int curlessthananchor;

	sprintf(buf, "@%d,%d", x, y);
	strcpy(cur, rivet_va_cmd(rivetobj, INDEX, buf, 0));
	text_index(rivetobj, ANCHOR, 0);
	if(rivet_error_occured)
		rivet_va_cmd(rivetobj, "mark", "set", ANCHOR, cur, 0);
	anchor_var = text_index(rivetobj, ANCHOR, 0);
	if(text_compare(rivetobj, cur, "!=", anchor_var) ||
	  (abs(tkPriv->pressX - x) >= 3))
		tkPriv->mouseMoved = 1;

	if(text_compare(rivetobj, cur, "<", ANCHOR))
		curlessthananchor = 1;
	else
		curlessthananchor = 0;


	if(streq(tkPriv->selectMode, "char")) {
		if(curlessthananchor) {
			strcpy(first, cur);
			strcpy(last, ANCHOR);
		} else {
			strcpy(first, ANCHOR);
			strcpy(last, text_index(rivetobj, "%s + 1c", cur));
		}
	} else
	if(streq(tkPriv->selectMode, "word")) {
		if(curlessthananchor) {
			strcpy(first, text_index(rivetobj,
						"%s wordstart", cur));
			strcpy(last, text_index(rivetobj, 
						"anchor - 1c wordend", 0));
		} else {
			strcpy(first, text_index(rivetobj,
						"anchor wordstart", 0));
			strcpy(last, text_index(rivetobj,
						"%s wordend", cur));
		}
 	} else
	if(streq(tkPriv->selectMode, "word")) {
		if(curlessthananchor) {
			strcpy(first, text_index(rivetobj, 
						"%s linestart", cur));
			strcpy(last, text_index(rivetobj,
						"anchor - 1c lineend + 1c", 0));
		} else {
			strcpy(first, text_index(rivetobj,
						"anchor linestart", 0));
			strcpy(last, text_index(rivetobj, 
						"%s lineend + 1c", cur));
		}
 	}

	if(tkPriv->mouseMoved || !streq(tkPriv->selectMode, "char")) {
		rivet_va_cmd(rivetobj, "tag", "remove", "sel", "0.0", first, 0);
		rivet_va_cmd(rivetobj, "tag", "add", "sel", first, last, 0);
		rivet_va_cmd(rivetobj, "tag", "remove", "sel", last, "end", 0);
		rivet_update(rivetobj, 1);
	}
}
 
/*
 * tkTextKeyExtend --
 * This procedure handles extending the selection from the keyboard,
 * where the point to extend to is really the boundary between two
 * characters rather than a particular character.
 *
 * Arguments:
 * w -		The text window.
 * index -	The point to which the selection is to be extended.
 */
static void
rivet_text_key_extend(rivetobj, index)
	Rivetobj rivetobj;
	char *index;
{
#if 0 
 proc tkTextKeyExtend {w index} {
     global tkPriv
 
     set cur [$w index $index]
     if [catch {$w index anchor}] {
 	$w mark set anchor $cur
     }
     set anchor [$w index anchor]
     if [$w compare $cur < anchor] {
 	set first $cur
 	set last anchor
     } else {
 	set first anchor
 	set last $cur
     }
     $w tag remove sel 0.0 $first
     $w tag add sel $first $last
     $w tag remove sel $last end
 }
#endif
	char *first;
	char *last;
	char cur[40];
	char *anchor_var; /* mysteriously set but not read. */

	strcpy(cur, text_index(rivetobj, index, 0));
	text_index(rivetobj, ANCHOR, 0);
	if(rivet_error_occured)
		rivet_va_cmd(rivetobj, "mark", "set", ANCHOR, cur, 0);
	anchor_var = text_index(rivetobj, ANCHOR, 0);
	if(text_compare(rivetobj, cur, "<", ANCHOR)) {
		first = cur;
		last = ANCHOR;
	} else {
		first = ANCHOR;
		last = cur;
	}
	rivet_va_cmd(rivetobj, "tag", "remove", "sel", "0.0", first, 0);
	rivet_va_cmd(rivetobj, "tag", "add", "sel", first, last, 0);
	rivet_va_cmd(rivetobj, "tag", "remove", "sel", last, "end", 0);
}
 
/*
 * tkTextAutoScan --
 * This procedure is invoked when the mouse leaves a text window
 * with button 1 down.  It scrolls the window up, down, left, or right,
 * depending on where the mouse is (this information was saved in
 * tkPriv(x) and tkPriv(y)), and reschedules itself as an "after"
 * command so that the window continues to scroll until the mouse
 * moves back into the window or the mouse button is released.
 *
 * Arguments:
 * w -		The text window.
 */
static void
rivet_text_auto_scan(cd)
	ClientData cd;
{ 
#if 0
 proc tkTextAutoScan {w} {
     global tkPriv
     if {$tkPriv(y) >= [winfo height $w]} {
 	$w yview scroll 2 units
     } elseif {$tkPriv(y) < 0} {
 	$w yview scroll -2 units
     } elseif {$tkPriv(x) >= [winfo width $w]} {
 	$w xview scroll 2 units
     } elseif {$tkPriv(x) < 0} {
 	$w xview scroll -2 units
     } else {
 	return
     }
     tkTextSelectTo $w $tkPriv(x) $tkPriv(y)
     set tkPriv(afterId) [after 50 tkTextAutoScan $w]
 }
#endif
	Rivetobj rivetobj = (ClientData)cd;
	
	if (tkPriv->y >= Tk_Height(rivetobj->tkwin)) {
		rivet_va_cmd(rivetobj, "yview", "scroll", "2", "units", 0);
	} else if (tkPriv->y < 0) {
		rivet_va_cmd(rivetobj, "yview", "scroll", "-2", "units", 0);
	} else if (tkPriv->x >= Tk_Width(rivetobj->tkwin)) {
		rivet_va_cmd(rivetobj, "xview", "scroll", "2", "units", 0);
	} else if (tkPriv->x < 0) {
		rivet_va_cmd(rivetobj, "xview", "scroll", "-2", "units", 0);
	} else
		return;

	rivet_text_select_to(rivetobj, tkPriv->x, tkPriv->y);
	tkPriv->afterId = Tk_CreateTimerHandler(50,
					rivet_text_auto_scan, rivetobj);
}
 
/*
 * tkTextSetCursor
 * Move the insertion cursor to a given position in a text.  Also
 * clears the selection, if there is one in the text, and makes sure
 * that the insertion cursor is visible.  Also, don't let the insertion
 * cursor appear on the dummy last line of the text.
 *
 * Arguments:
 * w -		The text window.
 * pos -	The desired new position for the cursor in the window.
 */
static void
rivet_text_set_cursor(rivetobj, pos)
	Rivetobj rivetobj;
	char *pos;
{ 
#if 0
 proc tkTextSetCursor {w pos} {
     global tkPriv
 
     if [$w compare $pos == end] {
 	set pos {end - 1 chars}
     }
     $w mark set insert $pos
     $w tag remove sel 1.0 end
     $w see insert
 }
#endif

	if(text_compare(rivetobj, pos, "==", "end", 0))
		pos = "end - 1 chars";

	rivet_va_cmd(rivetobj, "mark", "set", INSERT, pos, 0);
	rivet_va_cmd(rivetobj, "tag", "remove", "sel", _1_0_, "end", 0);
	rivet_va_cmd(rivetobj, "see", INSERT, 0);
}
 
/*
 * tkTextKeySelect
 * This procedure is invoked when stroking out selections using the
 * keyboard.  It moves the cursor to a new position, then extends
 * the selection to that position.
 *
 * Arguments:
 * w -		The text window.
 * new -	A new position for the insertion cursor (the cursor hasn't
 *		actually been moved to this position yet).
 */
static void
rivet_text_key_select(rivetobj, new)
	Rivetobj rivetobj;
	char *new;
{
#if 0
 proc tkTextKeySelect {w new} {
     global tkPriv
 
     if {[$w tag nextrange sel 1.0 end] == ""} {
 	if [$w compare $new < insert] {
 	    $w tag add sel $new insert
 	} else {
 	    $w tag add sel insert $new
 	}
 	$w mark set anchor insert
     } else {
 	if [$w compare $new < anchor] {
 	    set first $new
 	    set last anchor
 	} else {
 	    set first anchor
 	    set last $new
 	}
 	$w tag remove sel 1.0 $first
 	$w tag add sel $first $last
 	$w tag remove sel $last end
     }
     $w mark set insert $new
     $w see insert
     update idletasks
 }
#endif
	char *first;
	char *last;

	if(strnull(rivet_va_cmd(rivetobj,
			"tag", "nextrange", "sel", _1_0_, "end", 0))) {
	  if(text_compare(rivetobj, new, "<", INSERT))
		rivet_va_cmd(rivetobj, "tag", "add", "sel", new, INSERT, 0);
	  else
		rivet_va_cmd(rivetobj, "tag", "add", "sel", INSERT, new, 0);
	  rivet_va_cmd(rivetobj, "mark", "set", ANCHOR, INSERT, 0);
	} else {
	  if(text_compare(rivetobj, new, "<", ANCHOR)) {
		first = new;
		last = ANCHOR;
	  } else {
		first = ANCHOR;
		last = new;
	  }
	  rivet_va_cmd(rivetobj, "tag", "remove", "sel", _1_0_, first, 0);
	  rivet_va_cmd(rivetobj, "tag", "add", "sel", first, last, 0);
	  rivet_va_cmd(rivetobj, "tag", "remove", "sel", last, "end", 0);
	}
	rivet_va_cmd(rivetobj, "mark", "set", INSERT, new, 0);
	rivet_va_cmd(rivetobj, "see", INSERT, 0);
	rivet_update(rivetobj, 1);
}
 
/*
 * tkTextResetAnchor --
 * Set the selection anchor to whichever end is farthest from the
 * index argument.  One special trick: if the selection has two or
 * fewer characters, just leave the anchor where it is.  In this
 * case it doesn't matter which point gets chosen for the anchor,
 * and for the things like Shift-Left and Shift-Right this produces
 * better behavior when the cursor moves back and forth across the
 * anchor.
 *
 * Arguments:
 * w -		The text widget.
 * index -	Position at which mouse button was pressed, which determines
 *		which end of selection should be used as anchor point.
 */
static void
rivet_text_reset_anchor(rivetobj, index)
	Rivetobj rivetobj;
	char *index;
{
#if 0 
 proc tkTextResetAnchor {w index} {
     global tkPriv
 
     if {[$w tag ranges sel] == ""} {
 	$w mark set anchor $index
 	return
     }
     set a [$w index $index]
     set b [$w index sel.first]
     set c [$w index sel.last]
     if [$w compare $a < $b] {
 	$w mark set anchor sel.last
 	return
     }
     if [$w compare $a > $c] {
 	$w mark set anchor sel.first
 	return
     }
     scan $a "%d.%d" lineA chA
     scan $b "%d.%d" lineB chB
     scan $c "%d.%d" lineC chC
     if {$lineB < $lineC+2} {
 	set total [string length [$w get $b $c]]
 	if {$total <= 2} {
 	    return
 	}
 	if {[string length [$w get $b $a]] < ($total/2)} {
 	    $w mark set anchor sel.last
 	} else {
 	    $w mark set anchor sel.first
 	}
 	return
     }
     if {($lineA-$lineB) < ($lineC-$lineA)} {
 	$w mark set anchor sel.last
     } else {
 	$w mark set anchor sel.first
     }
 }
#endif
	char a[20];
	char b[20];
	char c[20];
	int lineA, lineB, lineC;
	int total;

	if(strnull(rivet_va_cmd(rivetobj, "tag", "ranges", "sel", 0))){
		rivet_va_cmd(rivetobj, "mark", "set", ANCHOR, index, 0);
		return;
	}
	strncpy(a, text_index(rivetobj, index, 0), sizeof(a));
	strncpy(b, text_index(rivetobj, "sel.first", 0), sizeof(a));
	strncpy(c, text_index(rivetobj, "sel.last", 0), sizeof(a));
	if(text_compare(rivetobj, a, "<", b)) {
		rivet_va_cmd(rivetobj, "mark", "set", ANCHOR, "sel.last", 0);
		return;
	}
	if(text_compare(rivetobj, a, ">", c)) {
		rivet_va_cmd(rivetobj, "mark", "set", ANCHOR, "sel.first", 0);
		return;
	}
	lineA = atoi(a);
	lineB = atoi(b);
	lineC = atoi(c);
	
	if(lineB < lineC+2) {
		total = strlen(rivet_va_cmd(rivetobj, "get", b, c, 0));
		if(total <= 2)
			return;
		if((int)strlen(rivet_va_cmd(rivetobj, "get", b, a, 0))<total/2)
			rivet_va_cmd(rivetobj,
				"mark", "set", ANCHOR, "sel.last", 0);
		else
			rivet_va_cmd(rivetobj,
				"mark", "set", ANCHOR, "sel.first", 0);
		return;
	}
	if(lineA-lineB < lineC-lineA)
		rivet_va_cmd(rivetobj, "mark", "set", ANCHOR, "sel.last", 0);
	else
		rivet_va_cmd(rivetobj, "mark", "set", ANCHOR, "sel.first", 0);
}

 
/*
 * tkTextInsert --
 * Insert a string into a text at the point of the insertion cursor.
 * If there is a selection in the text, and it covers the point of the
 * insertion cursor, then delete the selection before inserting.
 *
 * Arguments:
 * w -		The text window in which to insert the string
 * s -		The string to insert (usually just a single character)
 */
static void
rivet_text_insert(rivetobj, s)
	Rivetobj rivetobj;
	char *s;
{ 
#if 0
 proc tkTextInsert {w s} {
     if {($s == "") || ([$w cget -state] == "disabled")} {
 	return
     }
     catch {
 	if {[$w compare sel.first <= insert]
 		&& [$w compare sel.last >= insert]} {
 	    $w delete sel.first sel.last
 	}
     }
     $w insert insert $s
     $w see insert
 }
#endif
/* gack, if s is returned from a rivet function, may have to copy it to be safe,
 * also same problem in other functions here in this file.
 */
	if(strnull(s))
		return;
	if(rivet_state_p(rivetobj, "disabled"))
		return;
	if((text_compare(rivetobj, "sel.first", "<=", INSERT) &&
		!rivet_error_occured) &&
	   (text_compare(rivetobj, "sel.last", ">=", INSERT) &&
		!rivet_error_occured))
		rivet_va_cmd(rivetobj, "delete", "sel.first", "sel.last", 0);
	rivet_va_cmd(rivetobj, INSERT, INSERT, s, 0);
	rivet_va_cmd(rivetobj, "see", INSERT, 0);
}
 
/*
 * tkTextUpDownLine --
 * Returns the index of the character one line above or below the
 * insertion cursor.  There are two tricky things here.  First,
 * we want to maintain the original column across repeated operations,
 * even though some lines that will get passed through don't have
 * enough characters to cover the original column.  Second, don't
 * try to scroll past the beginning or end of the text.
 *
 * Arguments:
 * w -		The text window in which the cursor is to move.
 * n -		The number of lines to move: -1 for up one line,
 *		+1 for down one line.
 */
static char *
rivet_text_up_down_line(rivetobj, n)
	Rivetobj rivetobj;
	int n;
{
#if 0
 proc tkTextUpDownLine {w n} {
     global tkPriv
 
     set i [$w index insert]
     scan $i "%d.%d" line char
     if {[string compare $tkPriv(prevPos) $i] != 0} {
 	set tkPriv(char) $char
     }
     set new [$w index [expr $line + $n].$tkPriv(char)]
     if {[$w compare $new == end] || [$w compare $new == "insert linestart"]} {
 	set new $i
     }
     set tkPriv(prevPos) $new
     return $new
 }
#endif
	int line, ch;
	char i[20];
	char buf[20];
	static char new[20];

	strncpy(i, text_index(rivetobj, INSERT, 0), sizeof(i));
	if(sscanf(i, "%d.%d", &line, &ch) != 2)
		return "0.0";
	if(!streq(tkPriv->prevPos, i))
		strncpy(tkPriv->chr, rivet_itoa(ch), sizeof(tkPriv->chr));
	sprintf(buf, "%d.%d", line+n, tkPriv->chr);
	strncpy(new, text_index(rivetobj, buf, 0), sizeof(new));
	if(text_compare(rivetobj, new, "==", "end") ||
	  (text_compare(rivetobj, new, "==", "insert linestart")))
		strncpy(new, i, sizeof(new));
	strncpy(tkPriv->prevPos, new, sizeof(tkPriv->prevPos));
	return new; /* Be careful with return value. */
}
 
/*
 * tkTextPrevPara --
 * Returns the index of the beginning of the paragraph just before a given
 * position in the text (the beginning of a paragraph is the first non-blank
 * character after a blank line).
 *
 * Arguments:
 * w -		The text window in which the cursor is to move.
 * pos -	Position at which to start search.
 */
static char *
rivet_text_prev_para(rivetobj, pos_in)
	Rivetobj rivetobj;
	char *pos_in;
{
#if 0 
 proc tkTextPrevPara {w pos} {
     set pos [$w index "$pos linestart"]
     while 1 {
 	if {(([$w get "$pos - 1 line"] == "\n") && ([$w get $pos] != "\n"))
 		|| ($pos == _1_0_)} {
 	    if [regexp -indices {^[ 	]+(.)} [$w get $pos "$pos lineend"] \
 		    dummy index] {
 		set pos [$w index "$pos + [lindex $index 0] chars"]
 	    }
 	    if {[$w compare $pos != insert] || ($pos == _1_0_)} {
 		return $pos
 	    }
 	}
 	set pos [$w index "$pos - 1 line"]
     }
 }
#endif
	char buf[40];
	static char pos[20];

	strncpy(pos, text_index(rivetobj, "%s linestart", pos_in), sizeof(pos));
	while (1) {
	  sprintf(buf, "%s - 1 line", pos);
	  if(streq(rivet_va_cmd(rivetobj, "get", buf, 0), "\n") &&
	     !streq(rivet_va_cmd(rivetobj, "get", pos, 0), "\n") ||
	     streq(pos, _1_0_)) {
		char *tmp;
		int i;
		sprintf(buf, "%s lineend", pos);
		tmp = rivet_va_cmd(rivetobj, "get", pos, buf, 0);
		/* ... regexp ... */
		for (i=0; tmp[i] == '\t' || tmp[i] == ' '; i++)
			;
		sprintf(buf, "%s + %d chars", pos, i>0?i+1:i);
		strncpy(pos, text_index(rivetobj, buf, 0), sizeof(pos));

		if(text_compare(rivetobj, pos, "!=", INSERT) ||
		   streq(pos, _1_0_))
			return pos;
	  }
	  strncpy(pos, text_index(rivetobj, "%s - 1 line", pos), sizeof(pos));
	}
}
 
/*
 * tkTextNextPara --
 * Returns the index of the beginning of the paragraph just after a given
 * position in the text (the beginning of a paragraph is the first non-blank
 * character after a blank line).
 *
 * Arguments:
 * w -		The text window in which the cursor is to move.
 * start -	Position at which to start search.
 */
 
static char *
rivet_text_next_para(rivetobj, start)
	Rivetobj rivetobj;
	char *start;
{
#if 0 
 proc tkTextNextPara {w start} {
     set pos [$w index "$start linestart + 1 line"]
     while {[$w get $pos] != "\n"} {
 	if [$w compare $pos == end] {
 	    return [$w index "end - 1c"]
 	}
 	set pos [$w index "$pos + 1 line"]
     }
     while {[$w get $pos] == "\n"} {
 	set pos [$w index "$pos + 1 line"]
 	if [$w compare $pos == end] {
 	    return [$w index "end - 1c"]
 	}
     }
     if [regexp -indices {^[ 	]+(.)} [$w get $pos "$pos lineend"] \
 	    dummy index] {
 	return [$w index "$pos + [lindex $index 0] chars"]
     }
     return $pos
 }
#endif
	char buf[40];
	static char pos[20];
	char *tmp;
	int i;

	strncpy(pos, text_index(rivetobj, "%s linestart + 1 line", start),
		sizeof(pos));
	while (!streq(rivet_va_cmd(rivetobj, "get", pos, 0), "\n")) {
		if(text_compare(rivetobj, pos, "==", "end")) {
			return text_index(rivetobj, "end - 1c", 0);
		}
		strncpy(pos, text_index(rivetobj, "%s + 1 line", pos),
			sizeof(pos));
	}

	while(streq(rivet_va_cmd(rivetobj, "get", pos, 0), "\n")) {
		strncpy(pos, text_index(rivetobj, "%s + 1 line", pos),
			sizeof(pos));
		if(text_compare(rivetobj, pos, "==", "end"))
			return text_index(rivetobj, "end - 1c", 0);
	}

	sprintf(buf, "%s lineend", pos);
	tmp = rivet_va_cmd(rivetobj, "get", buf, 0);
	for(i=0;tmp[i] == '\t' || tmp[i] == ' ';i++)
		;

	if(i > 0) {
		sprintf(buf, "%s + %d chars", pos, i+1);
		return text_index(rivetobj, buf, 0);
	}
	return pos;
}
 
/*
 * tkTextScrollPages --
 * This is a utility procedure used in bindings for moving up and down
 * pages and possibly extending the selection along the way.  It scrolls
 * the view in the widget by the number of pages, and it returns the
 * index of the character that is at the same position in the new view
 * as the insertion cursor used to be in the old view.
 *
 * Arguments:
 * w -		The text window in which the cursor is to move.
 * count -	Number of pages forward to scroll;  may be negative
 *		to scroll backwards.
 */
static char *
rivet_text_scroll_pages(rivetobj, count)
	Rivetobj rivetobj;
	char *count;
{
#if 0 
 proc tkTextScrollPages {w count} {
     set bbox [$w bbox insert]
     $w yview scroll $count pages
     if {$bbox == ""} {
 	return [$w index @[expr [winfo height $w]/2],0]
     }
     return [$w index @[lindex $bbox 0],[lindex $bbox 1]]
 }
#endif
	char buf[40];
	char *bbox;
	long x, y;

	bbox = rivet_va_cmd(rivetobj, "bbox", INSERT, 0);
	rivet_va_cmd(rivetobj, "yview", "scroll", count, "pages", 0);
	if(strnull(bbox)) {
		return text_index(rivetobj,
				"@%d,0", Tk_Height(rivetobj->tkwin)/2);
	}
	x = strtol(bbox, &bbox, 10);
	y = strtol(bbox, 0, 10);
	sprintf(buf, "@%d,%d", x, y);
	return text_index(rivetobj, buf, 0);
}
 
/*
 * tkTextTranspose --
 * This procedure implements the "transpose" function for text widgets.
 * It tranposes the characters on either side of the insertion cursor,
 * unless the cursor is at the end of the line.  In this case it
 * transposes the two characters to the left of the cursor.  In either
 * case, the cursor ends up to the right of the transposed characters.
 *
 * Arguments:
 * w -		Text window in which to transpose.
 */
 
static void
rivet_text_transpose(rivetobj)
	Rivetobj rivetobj;
{
#if 0 
 proc tkTextTranspose w {
     set pos insert
     if [$w compare $pos != "$pos lineend"] {
 	set pos [$w index "$pos + 1 char"]
     }
     set new [$w get "$pos - 1 char"][$w get  "$pos - 2 char"]
     if [$w compare "$pos - 1 char" == 1.0] {
 	return
     }

     $w delete "$pos - 2 char" $pos
     $w insert insert $new
     $w see insert
 }
#endif
	char buf1[20];
	char buf2[20];
	char *r1;
	char *r2;
	char new[40];
	char pos[20];


	strcpy(pos, INSERT);
	if(text_compare(rivetobj, INSERT, "!=", "insert lineend"))
		strncpy(pos, text_index(rivetobj, "insert + 1 char", 0),
			sizeof(pos));
	sprintf(buf1, "%s - 1 char", pos);
	sprintf(buf2, "%s - 2 char", pos);
	r1 = rivet_va_cmd(rivetobj, "get", buf1, 0);
	r2 = rivet_va_cmd(rivetobj, "get", buf2, 0);
	strncpy(new, r1, sizeof(new));
	strncat(new, r2, sizeof(new));
	if(text_compare(rivetobj, buf1, "==", _1_0_))
		return;
	/* puts!  Looks like this is not really finished. */
	rivet_va_cmd(rivetobj, "delete", buf2, pos, 0);
	rivet_va_cmd(rivetobj, INSERT, INSERT, new, 0);
	rivet_va_cmd(rivetobj, "see", INSERT, 0);
}


/*
 * tkTextBind --
 * This procedure below invoked the first time the mouse enters a text
 * widget or a text widget receives the input focus.  It creates all of
 * the class Bindings for texts.
 *
 * Arguments:
 * event -	Indicates which event caused the procedure to be invoked
 *		(Enter or FocusIn).  It is used so that we can carry out
 *		the functions of that event in addition to setting up
 *		Bindings.
 * ---------------------------------------------------------------------- 
 * This function is odd because it does both the Binding and the callbacks.
 * The goal is to keep the short sections of code close to the string that
 * it binds to.  Case numbers are intentionally sequential.
 */

#define binding(_pat_, _n_)						\
	case _n_:							\
	if(!rivet_is_class_initialized(rivetobj->rivet_class)) {	\
	   return rivet_create_binding(0, rivetobj->rivet_class->name,	\
				_pat_, text_double_duty_event_handler,	\
				(ClientData) assoc, 0, 0);		\
	}


static int
text_double_duty_event_handler(rivetobj, xevent, assoc)
	Rivetobj rivetobj;
	XEvent *xevent;
	ClientData assoc;
{
	int x, y;
	char buf[20];

	/* Don't use these if it ain't the right event! */
	/* They won't crash, but they won't be any good either */
	if(xevent) {
		x = ((XMotionEvent*)xevent)->x;
		y = ((XMotionEvent*)xevent)->y;
	}
	
	switch( (int) assoc) {
	/* Standard Motif Bindings: */
	binding("<1>", 1)
		/*
		tkTextButton1 %W %x %y
		%W tag remove sel 0.0 end
    		*/
		rivet_text_button1(rivetobj, x, y);
		rivet_va_cmd(rivetobj, "tag", "remove", "sel", "0.0", "end", 0);
		break;
	binding("<B1-Motion>", 2)
		/*
		set tkPriv(x) %x
		set tkPriv(y) %y
		tkTextSelectTo %W %x %y
    		*/
		tkPriv->x = x;
		tkPriv->y = y;
		rivet_text_select_to(rivetobj, x, y);
		break;
	binding("<Double-1>", 3)
		/*
		set tkPriv(selectMode) word
		tkTextSelectTo %W %x %y
		catch {%W mark set insert sel.first}
		*/
		tkPriv->selectMode = "word";
		rivet_text_select_to(rivetobj, x, y);
		rivet_va_cmd(rivetobj, "mark", "set", INSERT, "sel.first", 0);
		break;
	binding("<Triple-1>", 4)
		/*
		set tkPriv(selectMode) line
		tkTextSelectTo %W %x %y
		catch {%W mark set insert sel.first}
    		*/
		tkPriv->selectMode = "line";
		rivet_text_select_to(rivetobj, x, y);
		rivet_va_cmd(rivetobj, "mark", "set", INSERT, "sel.first", 0);
		break;
	binding("<Shift-1>", 5)
		/*
		tkTextResetAnchor %W @%x,%y
		set tkPriv(selectMode) char
		tkTextSelectTo %W %x %y
		*/
		sprintf(buf, "@%d,%d", x, y);
		tkPriv->selectMode = "char";
		rivet_text_select_to(rivetobj, x, y);
		break;
	binding("<Double-Shift-1>", 6)
		/*
		set tkPriv(selectMode) word
		tkTextSelectTo %W %x %y
		*/
		tkPriv->selectMode = "word";
		rivet_text_select_to(rivetobj, x, y);
		break;
	binding("<Triple-Shift-1>", 7)
		/*
		set tkPriv(selectMode) line
		tkTextSelectTo %W %x %y
		*/
		tkPriv->selectMode = "line";
		rivet_text_select_to(rivetobj, x, y);
		break;
	binding("<B1-Leave>", 8)
		/*
		set tkPriv(x) %x
		set tkPriv(y) %y
		tkTextAutoScan %W
		*/
		tkPriv->x = x;
		tkPriv->y = y;
		rivet_text_auto_scan(rivetobj);
		break;
	binding("<B1-Enter>", 9)
		/*
		tkCancelRepeat
		*/
		/* FALL THROUGH */
	binding("<ButtonRelease-1>", 10)
		/*
		tkCancelRepeat
		*/
		rivet_cancel_repeat();
		break;
	binding("<Control-1>", 11)
		/*
		%W mark set insert @%x,%y
		*/
		sprintf(buf, "@%d,%d", x, y);
		rivet_va_cmd(rivetobj, "mark", "set", INSERT, buf, 0);
		break;
	binding("<Left>", 12)
		/*
		tkTextSetCursor %W [%W index {insert - 1c}]
		*/
		rivet_text_set_cursor(rivetobj,
			rivet_va_cmd(rivetobj, INDEX, "insert - 1c", 0));
		break;
	binding("<Right>", 13)
		/*
		tkTextSetCursor %W [%W index {insert + 1c}]
		*/
		rivet_text_set_cursor(rivetobj,
			rivet_va_cmd(rivetobj, INDEX, "insert + 1c", 0));
		break;
	binding("<Up>", 14)
		/*
		tkTextSetCursor %W [tkTextUpDownLine %W -1]
		*/
		rivet_text_set_cursor(rivetobj,
			rivet_text_up_down_line(rivetobj, -1));
		break;
	binding("<Down>", 15)
		/*
		tkTextSetCursor %W [tkTextUpDownLine %W 1]
		*/
		rivet_text_set_cursor(rivetobj,
			rivet_text_up_down_line(rivetobj, 1));
		break;
	binding("<Shift-Left>", 16)
		/*
		tkTextKeySelect %W [%W index {insert - 1c}]
		*/
		rivet_text_key_select(rivetobj,
			rivet_va_cmd(rivetobj, INDEX, "insert - 1c", 0));
		break;
	binding("<Shift-Right>", 17)
		/*
		tkTextKeySelect %W [%W index {insert + 1c}]
		*/
		rivet_text_key_select(rivetobj,
			rivet_va_cmd(rivetobj, INDEX, "insert + 1c", 0));
		break;
	binding("<Shift-Up>", 18)
		/*
		tkTextKeySelect %W [tkTextUpDownLine %W -1]
		*/
		rivet_text_key_select(rivetobj,
			rivet_text_up_down_line(rivetobj, -1));
		break;
	binding("<Shift-Down>", 19)
		/*
		tkTextKeySelect %W [tkTextUpDownLine %W 1]
		*/
		rivet_text_key_select(rivetobj,
			rivet_text_up_down_line(rivetobj, 1));
		break;
	binding("<Control-Left>", 20)
		/*
		tkTextSetCursor %W [%W index {insert - 1c wordstart}]
		*/
		rivet_text_set_cursor(rivetobj,
		   rivet_va_cmd(rivetobj, INDEX, "insert - 1c wordstart", 0));
		break;
	binding("<Control-Right>", 21)
		/*
		tkTextSetCursor %W [%W index {insert wordend}]
		*/
		rivet_text_set_cursor(rivetobj,
			rivet_va_cmd(rivetobj, INDEX, "insert wordend", 0));
		break;
	binding("<Control-Up>", 22)
		/*
		tkTextSetCursor %W [tkTextPrevPara %W insert]
		*/
		rivet_text_set_cursor(rivetobj,
			rivet_text_prev_para(rivetobj, INSERT));
		break;
	binding("<Control-Down>", 23)
		/*
		tkTextSetCursor %W [tkTextNextPara %W insert]
		*/
		rivet_text_set_cursor(rivetobj,
			rivet_text_next_para(rivetobj, INSERT));
		break;
	binding("<Shift-Control-Left>", 24)
		/*
		tkTextKeySelect %W [%W index {insert - 1c wordstart}]
		*/
		rivet_text_key_select(rivetobj,
		   rivet_va_cmd(rivetobj, INDEX, "insert - 1c wordstart", 0));
		break;
	binding("<Shift-Control-Right>", 25)
		/*
		tkTextKeySelect %W [%W index {insert wordend}]
		*/
		rivet_text_key_select(rivetobj,
			rivet_va_cmd(rivetobj, INDEX, "insert wordend", 0));
		break;
	binding("<Shift-Control-Up>", 26)
		/*
		tkTextKeySelect %W [tkTextPrevPara %W insert]
		*/
		rivet_text_key_select(rivetobj,
			rivet_text_prev_para(rivetobj, INSERT));
		break;
	binding("<Shift-Control-Down>", 27)
		/*
		tkTextKeySelect %W [tkTextNextPara %W insert]
		*/
		rivet_text_key_select(rivetobj,
			rivet_text_next_para(rivetobj, INSERT));
		break;
	binding("<Prior>", 28)
		/*
		tkTextSetCursor %W [tkTextScrollPages %W -1]
		*/
		rivet_text_set_cursor(rivetobj,
			rivet_text_scroll_pages(rivetobj, "-1"));
		break;
	binding("<Shift-Prior>", 29)
		/*
		tkTextKeySelect %W [tkTextScrollPages %W -1]
		*/
		rivet_text_key_select(rivetobj,
			rivet_text_scroll_pages(rivetobj, "-1"));
		break;
	binding("<Next>", 30)
		/*
		tkTextSetCursor %W [tkTextScrollPages %W 1]
		*/
		rivet_text_set_cursor(rivetobj,
			rivet_text_scroll_pages(rivetobj, "1"));
		break;
	binding("<Shift-Next>", 31)
		/*
		tkTextKeySelect %W [tkTextScrollPages %W 1]
		*/
		rivet_text_key_select(rivetobj,
			rivet_text_scroll_pages(rivetobj, "1"));
		break;
	binding("<Control-Prior>", 32)
		/*
		%W xview scroll -1 page
		*/
		rivet_va_cmd(rivetobj, "xview", "scroll", "-1", "page", 0);
		break;
	binding("<Control-Next>", 33)
		/*
		%W xview scroll 1 page
		*/
		rivet_va_cmd(rivetobj, "xview", "scroll", "1", "page", 0);
		break;
	binding("<Home>", 34)
		/*
		tkTextSetCursor %W {insert linestart}
		*/
		rivet_text_set_cursor(rivetobj, "insert linestart");
		break;
	binding("<Shift-Home>", 35)
		/*
		tkTextKeySelect %W {insert linestart}
		*/
		rivet_text_key_select(rivetobj, "insert linestart");
		break;
	binding("<End>", 36)
		/*
		tkTextSetCursor %W {insert lineend}
		*/
		rivet_text_set_cursor(rivetobj, "insert lineend");
		break;
	binding("<Shift-End>", 37)
		/*
		tkTextKeySelect %W {insert lineend}
		*/
		rivet_text_key_select(rivetobj, "insert lineend");
		break;
	binding("<Control-Home>", 38)
		/*
		tkTextSetCursor %W 1.0
		*/
		rivet_text_set_cursor(rivetobj, _1_0_);
		break;
	binding("<Control-Shift-Home>", 39)
		/*
		tkTextKeySelect %W 1.0
		*/
		rivet_text_key_select(rivetobj, _1_0_);
		break;
	binding("<Control-End>", 40)
		/*
		tkTextSetCursor %W {end - 1 char}
		*/
		rivet_text_set_cursor(rivetobj, "end - 1 char");
		break;
	binding("<Control-Shift-End>", 41)
		/*
		tkTextKeySelect %W {end - 1 char}
		*/
		rivet_text_key_select(rivetobj, "end - 1 char");
		break;

	binding("<Tab>", 42)
		/*
		tkTextInsert %W \t
		focus %W
		break
		*/
		rivet_text_insert(rivetobj, "\t");
		rivet_set_focus(rivetobj);
		return TCL_BREAK;

	binding("<Shift-Tab>", 43)
		/*
		 * Needed only to keep <Tab> from triggering;  doesn't
		 * have to actually do anything.
		 */
		break;
	binding("<Control-Tab>", 44)
		/*
		focus [tk_focusNext %W]
		*/
		rivet_set_focus(rivet_focus_next(rivetobj));
		break;
	binding("<Control-Shift-Tab>", 45)
		/*
		focus [tk_focusPrev %W]
		*/
		rivet_set_focus(rivet_focus_prev(rivetobj));
		break;
	binding("<Control-i>", 46)
		/*
		tkTextInsert %W \t
		*/
		rivet_text_insert(rivetobj, "\t");
		break;
	binding("<Return>", 47)
		/*
		tkTextInsert %W \n
		*/
		rivet_text_insert(rivetobj, "\n");
		break;
	binding("<Delete>", 48)
		/*
		if {[%W tag nextrange sel 1.0 end] != ""} {
		    %W delete sel.first sel.last
		} else {
		    %W delete insert
		    %W see insert
		}
		*/
		if(!strnull( rivet_va_cmd(rivetobj,
				"tag", "nextrange", "sel", _1_0_, "end", 0)))
		  rivet_va_cmd(rivetobj, "delete", "sel.first", "sel.last", 0);
		else {
		  rivet_va_cmd(rivetobj, "delete", INSERT, 0);
		  rivet_va_cmd(rivetobj, "see", INSERT, 0);
		}
		break;
	binding("<BackSpace>", 49)
		/*
		if {[%W tag nextrange sel 1.0 end] != ""} {
		    %W delete sel.first sel.last
		} elseif [%W compare insert != 1.0] {
		    %W delete insert-1c
		    %W see insert
		}
		*/
		if(!strnull( rivet_va_cmd(rivetobj,
				"tag", "nextrange", "sel", _1_0_, "end", 0))) {
		  rivet_va_cmd(rivetobj, "delete", "sel.first", "sel.last", 0);
		} else {
		  if(text_compare(rivetobj, INSERT, "!=", _1_0_)) {
			rivet_va_cmd(rivetobj, "delete", "insert-1c", 0);
			rivet_va_cmd(rivetobj, "see", INSERT, 0);
		  }
		}
		break;
	binding("<Control-space>", 50)
		/*
		%W mark set anchor insert
		*/
		/* FALL THROUGH */
	binding("<Select>", 51)
		/*
		%W mark set anchor insert
		*/
		rivet_va_cmd(rivetobj, "mark", "set", ANCHOR, INSERT, 0);
		break;
	binding("<Control-Shift-space>", 52)
		/*
		set tkPriv(selectMode) char
		tkTextKeyExtend %W insert
		*/
		/* FALL THROUGH */
	binding("<Shift-Select>", 53)
		/*
		set tkPriv(selectMode) char
		tkTextKeyExtend %W insert
		*/
		tkPriv->selectMode = "char";
		rivet_text_key_extend(rivetobj, INSERT);
		break;
	binding("<Control-slash>", 54)
		/*
		%W tag add sel 1.0 end
		*/
		rivet_va_cmd(rivetobj, "tag", "add", "sel", _1_0_, "end", 0);
		break;
	binding("<Control-backslash>", 55)
		/*
		%W tag remove sel 1.0 end
		*/
		rivet_va_cmd(rivetobj, "tag", "remove", "sel", _1_0_, "end", 0);
		break;
	binding("<Insert>", 56)
		/*
		catch {tkTextInsert %W [selection get -displayof %W]}
		*/
		rivet_text_insert(rivetobj, rivet_sel_displayof(rivetobj));
		break;
	binding("<KeyPress>", 57)
		/*
		tkTextInsert %W %A
		*/
		rivet_text_insert(rivetobj, rivet_input_str(rivetobj, xevent));
		break;

	/*
	 * Ignore all Alt, Meta, and Control keypresses unless explicitly bound.
	 * Otherwise, if a widget Binding for one of these is defined, the
	 * <KeyPress> class Binding will also fire and insert the character,
	 * which is wrong.  Ditto for <Escape>.
	 */
	binding("<Alt-KeyPress>", 58)
		break;
	binding("<Meta-KeyPress>", 59)
		break;
	binding("<Control-KeyPress>", 60)
		break;
	binding("<Escape>", 61)
		break;
	binding("<KP_Enter>", 62)
		break;

	/* Additional emacs-like bindings: */

	case 63:
		if(tk_strictMotif)
			/* Returning 0 should prevent remaining functions from
			 * getting bound. */
			return 0;
		else
			return 1;

	binding("<Control-a>", 64)
		/*
		tkTextSetCursor %W {insert linestart}
		*/
		rivet_text_set_cursor(rivetobj, "insert linestart");
		break;
	binding("<Control-b>", 65)
		/*
		tkTextSetCursor %W insert-1c
		*/
		rivet_text_set_cursor(rivetobj, "insert-1c");
		break;
	binding("<Control-d>", 66)
		/*
		%W delete insert
		*/
		rivet_va_cmd(rivetobj, "delete", INSERT, 0);
		break;
	binding("<Control-e>", 67)
		/*
		tkTextSetCursor %W {insert lineend}
		*/
		rivet_text_set_cursor(rivetobj, "insert lineend");
		break;
	binding("<Control-f>", 68)
		/*
		tkTextSetCursor %W insert+1c
		*/
		rivet_text_set_cursor(rivetobj, "insert+1c");
		break;
	binding("<Control-k>", 69)
		/*
		if [%W compare insert == {insert lineend}] {
			%W delete insert
		} else {
			%W delete insert {insert lineend}
		}
		*/
		if(text_compare(rivetobj, INSERT, "==", "insert lineend"))
			rivet_va_cmd(rivetobj, "delete", INSERT, 0);
		else
			rivet_va_cmd(rivetobj,
				"delete", INSERT, "insert lineend", 0);
		break;
	binding("<Control-n>", 70)
		/*
		tkTextSetCursor %W [tkTextUpDownLine %W 1]
		*/
		rivet_text_set_cursor(rivetobj,
			rivet_text_up_down_line(rivetobj, 1));
		break;
	binding("<Control-o>", 71)
		/*
		%W insert insert \n
		%W mark set insert insert-1c
		*/
		rivet_va_cmd(rivetobj, INSERT, INSERT, "\n", 0);
		rivet_va_cmd(rivetobj, "mark", "set", INSERT, "insert-1c", 0);
		break;
	binding("<Control-p>", 72)
		/*
		tkTextSetCursor %W [tkTextUpDownLine %W -1]
		*/
		rivet_text_set_cursor(rivetobj,
			rivet_text_up_down_line(rivetobj, -1));
		break;
	binding("<Control-t>", 73)
		/*
		tkTextTranspose %W
		*/
		rivet_text_transpose(rivetobj);
		break;
	binding("<Meta-b>", 74)
		/*
		tkTextSetCursor %W {insert - 1c wordstart}
		*/
		rivet_text_set_cursor(rivetobj, "insert - 1c wordstart");
		break;
	binding("<Meta-d>", 75)
		/*
		%W delete insert {insert wordend}
		*/
		rivet_va_cmd(rivetobj, "delete", INSERT, "insert wordend", 0);
		break;
	binding("<Meta-f>", 76)
		/*
		tkTextSetCursor %W {insert wordend}
		*/
		rivet_text_set_cursor(rivetobj, "insert wordend");
		break;
	binding("<Meta-less>", 77)
		/*
		tkTextSetCursor %W 1.0
		*/
		rivet_text_set_cursor(rivetobj, _1_0_);
		break;
	binding("<Meta-greater>", 78)
		/*
		tkTextSetCursor %W end-1c
		*/
		rivet_text_set_cursor(rivetobj, "end-1c");
		break;
	binding("<Meta-BackSpace>", 79)
		/*
		%W delete {insert -1c wordstart} insert
		*/
		rivet_va_cmd(rivetobj,
			"delete", "insert -1c wordstart", INSERT, 0);
		break;

	binding("<Meta-Delete>", 80)
		/*
	        %W delete {insert -1c wordstart} insert
		*/
		rivet_va_cmd(rivetobj,
			"delete", "insert -1c wordstart", INSERT, 0);
		break;

	/* A few additional Bindings of my own. */
    
	binding("<Control-h>", 81)
		/*
		if [%W compare insert != 1.0] {
			%W delete insert-1c
			%W see insert
		}
		*/
		if(text_compare(rivetobj, INSERT, "!=", _1_0_)) {
			rivet_va_cmd(rivetobj, "delete", "insert-1c", 0);
			rivet_va_cmd(rivetobj, "see", INSERT, 0);
		}
		break;
	binding("<Control-v>", 82)
		/*
		tkTextScrollPages %W 1
		}
		*/
		rivet_text_scroll_pages(rivetobj, "1");
		break;
	binding("<2>", 83)
		/*
		%W scan mark %x %y
		set tkPriv(x) %x
		set tkPriv(y) %y
		set tkPriv(mouseMoved) 0
		*/
		rivet_va_cmd(rivetobj, "scan", "mark",
			rivet_itoa(x), rivet_itoa(y), 0);
		tkPriv->x = x;
		tkPriv->y = y;
		tkPriv->mouseMoved = 0;
		break;
	binding("<B2-Motion>", 84)
		/*
		if {(%x != $tkPriv(x)) || (%y != $tkPriv(y))} {
			set tkPriv(mouseMoved) 1
		}
		if $tkPriv(mouseMoved) {
			%W scan dragto %x %y
		}
		*/
		if((x != tkPriv->x) || (y != tkPriv->y))
			tkPriv->mouseMoved = 1;
		if(tkPriv->mouseMoved)
			rivet_va_cmd(rivetobj, "scan", "dragto",
			rivet_itoa(x), rivet_itoa(y), 0);
		break;
	binding("<ButtonRelease-2>", 85)
		/*
		if !$tkPriv(mouseMoved) {
			catch {
				%W insert @%x,%y [selection get -displayof %W]
			}
		}
		*/
		if(tkPriv->mouseMoved) {
			sprintf(buf, "@%d,%d", x, y);
			rivet_va_cmd(rivetobj, INSERT, buf,
				rivet_sel_displayof(rivetobj), 0);
		}
		break;

	case 86:
		return TCL_ERROR; /* should cause bind phase to complete */

	default:
		return 1;
	}

	return TCL_CONTINUE;
	error:
		return TCL_ERROR;
}

void
rivet_text_bind(widget)
	Rivetobj widget;
{
	int n;

	if(rivet_is_class_initialized(widget->rivet_class))
		return;
	rivet_class_initializing(widget->rivet_class); 

	tkPriv->prevPos[0] = 0;

	for(n=1;;n++) {
		if(text_double_duty_event_handler(widget, 0, n) == TCL_ERROR)
			break;
	}

	if(rivet_bind_clipboard(widget, "<F16>", "<F20>", "<F18>") == TCL_ERROR)
		return ;

	if(!tk_strictMotif &&
	  rivet_bind_clipboard(widget,
			"<Meta-w>", "<Control-w>", "<Control-y>") == TCL_ERROR)
		return ;

	rivet_class_init_done(widget->rivet_class);
	return ;
}

